(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('./shared.js')) :
  typeof define === 'function' && define.amd ? define(['./shared.js'], factory) :
  (global.Select = factory(global.svelteShared));
}(this, (function (shared_js) { 'use strict';

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template A
   */
  class MDCFoundation {
    /** @return enum{cssClasses} */
    static get cssClasses() {
      // Classes extending MDCFoundation should implement this method to return an object which exports every
      // CSS class the foundation class needs as a property. e.g. {ACTIVE: 'mdc-component--active'}
      return {};
    }

    /** @return enum{strings} */
    static get strings() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // semantic strings as constants. e.g. {ARIA_ROLE: 'tablist'}
      return {};
    }

    /** @return enum{numbers} */
    static get numbers() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // of its semantic numbers as constants. e.g. {ANIMATION_DELAY_MS: 350}
      return {};
    }

    /** @return {!Object} */
    static get defaultAdapter() {
      // Classes extending MDCFoundation may choose to implement this getter in order to provide a convenient
      // way of viewing the necessary methods of an adapter. In the future, this could also be used for adapter
      // validation.
      return {};
    }

    /**
     * @param {A=} adapter
     */
    constructor(adapter = {}) {
      /** @protected {!A} */
      this.adapter_ = adapter;
    }

    init() {
      // Subclasses should override this method to perform initialization routines (registering events, etc.)
    }

    destroy() {
      // Subclasses should override this method to perform de-initialization routines (de-registering events, etc.)
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template F
   */
  class MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCComponent}
     */
    static attachTo(root) {
      // Subclasses which extend MDCBase should provide an attachTo() method that takes a root element and
      // returns an instantiated component with its root set to that element. Also note that in the cases of
      // subclasses, an explicit foundation class will not have to be passed in; it will simply be initialized
      // from getDefaultFoundation().
      return new MDCComponent(root, new MDCFoundation());
    }

    /**
     * @param {!Element} root
     * @param {F=} foundation
     * @param {...?} args
     */
    constructor(root, foundation = undefined, ...args) {
      /** @protected {!Element} */
      this.root_ = root;
      this.initialize(...args);
      // Note that we initialize foundation here and not within the constructor's default param so that
      // this.root_ is defined and can be used within the foundation class.
      /** @protected {!F} */
      this.foundation_ = foundation === undefined ? this.getDefaultFoundation() : foundation;
      this.foundation_.init();
      this.initialSyncWithDOM();
    }

    initialize(/* ...args */) {
      // Subclasses can override this to do any additional setup work that would be considered part of a
      // "constructor". Essentially, it is a hook into the parent constructor before the foundation is
      // initialized. Any additional arguments besides root and foundation will be passed in here.
    }

    /**
     * @return {!F} foundation
     */
    getDefaultFoundation() {
      // Subclasses must override this method to return a properly configured foundation class for the
      // component.
      throw new Error('Subclasses must override getDefaultFoundation to return a properly configured ' +
        'foundation class');
    }

    initialSyncWithDOM() {
      // Subclasses should override this method if they need to perform work to synchronize with a host DOM
      // object. An example of this would be a form control wrapper that needs to synchronize its internal state
      // to some property or attribute of the host DOM. Please note: this is *not* the place to perform DOM
      // reads/writes that would cause layout / paint, as this is called synchronously from within the constructor.
    }

    destroy() {
      // Subclasses may implement this method to release any resources / deregister any listeners they have
      // attached. An example of this might be deregistering a resize event from the window object.
      this.foundation_.destroy();
    }

    /**
     * Wrapper method to add an event listener to the component's root element. This is most useful when
     * listening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    listen(evtType, handler) {
      this.root_.addEventListener(evtType, handler);
    }

    /**
     * Wrapper method to remove an event listener to the component's root element. This is most useful when
     * unlistening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    unlisten(evtType, handler) {
      this.root_.removeEventListener(evtType, handler);
    }

    /**
     * Fires a cross-browser-compatible custom event from the component root of the given type,
     * with the given data.
     * @param {string} evtType
     * @param {!Object} evtData
     * @param {boolean=} shouldBubble
     */
    emit(evtType, evtData, shouldBubble = false) {
      let evt;
      if (typeof CustomEvent === 'function') {
        evt = new CustomEvent(evtType, {
          detail: evtData,
          bubbles: shouldBubble,
        });
      } else {
        evt = document.createEvent('CustomEvent');
        evt.initCustomEvent(evtType, shouldBubble, false, evtData);
      }

      this.root_.dispatchEvent(evt);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const cssClasses = {
    LABEL_FLOAT_ABOVE: 'mdc-floating-label--float-above',
    LABEL_SHAKE: 'mdc-floating-label--shake',
  };

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCFoundation<!MDCFloatingLabelAdapter>}
   * @final
   */
  class MDCFloatingLabelFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses;
    }

    /**
     * {@see MDCFloatingLabelAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCFloatingLabelAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCFloatingLabelAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        getWidth: () => {},
        registerInteractionHandler: () => {},
        deregisterInteractionHandler: () => {},
      });
    }

    /**
     * @param {!MDCFloatingLabelAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCFloatingLabelFoundation.defaultAdapter, adapter));

      /** @private {function(!Event): undefined} */
      this.shakeAnimationEndHandler_ = () => this.handleShakeAnimationEnd_();
    }

    init() {
      this.adapter_.registerInteractionHandler('animationend', this.shakeAnimationEndHandler_);
    }

    destroy() {
      this.adapter_.deregisterInteractionHandler('animationend', this.shakeAnimationEndHandler_);
    }

    /**
     * Returns the width of the label element.
     * @return {number}
     */
    getWidth() {
      return this.adapter_.getWidth();
    }

    /**
     * Styles the label to produce the label shake for errors.
     * @param {boolean} shouldShake adds shake class if true,
     * otherwise removes shake class.
     */
    shake(shouldShake) {
      const {LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      if (shouldShake) {
        this.adapter_.addClass(LABEL_SHAKE);
      } else {
        this.adapter_.removeClass(LABEL_SHAKE);
      }
    }

    /**
     * Styles the label to float or dock.
     * @param {boolean} shouldFloat adds float class if true, otherwise remove
     * float and shake class to dock label.
     */
    float(shouldFloat) {
      const {LABEL_FLOAT_ABOVE, LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      if (shouldFloat) {
        this.adapter_.addClass(LABEL_FLOAT_ABOVE);
      } else {
        this.adapter_.removeClass(LABEL_FLOAT_ABOVE);
        this.adapter_.removeClass(LABEL_SHAKE);
      }
    }

    /**
     * Handles an interaction event on the root element.
     */
    handleShakeAnimationEnd_() {
      const {LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      this.adapter_.removeClass(LABEL_SHAKE);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCFloatingLabelFoundation>}
   * @final
   */
  class MDCFloatingLabel extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCFloatingLabel}
     */
    static attachTo(root) {
      return new MDCFloatingLabel(root);
    }

    /**
     * Styles the label to produce the label shake for errors.
     * @param {boolean} shouldShake styles the label to shake by adding shake class
     * if true, otherwise will stop shaking by removing shake class.
     */
    shake(shouldShake) {
      this.foundation_.shake(shouldShake);
    }

    /**
     * Styles label to float/dock.
     * @param {boolean} shouldFloat styles the label to float by adding float class
     * if true, otherwise docks the label by removing the float class.
     */
    float(shouldFloat) {
      this.foundation_.float(shouldFloat);
    }

    /**
     * @return {number}
     */
    getWidth() {
      return this.foundation_.getWidth();
    }

    /**
     * @return {!MDCFloatingLabelFoundation}
     */
    getDefaultFoundation() {
      return new MDCFloatingLabelFoundation({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        getWidth: () => this.root_.offsetWidth,
        registerInteractionHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterInteractionHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
      });
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const cssClasses$1 = {
    LINE_RIPPLE_ACTIVE: 'mdc-line-ripple--active',
    LINE_RIPPLE_DEACTIVATING: 'mdc-line-ripple--deactivating',
  };

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  /**
   * @extends {MDCFoundation<!MDCLineRippleAdapter>}
   * @final
   */
  class MDCLineRippleFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$1;
    }

    /**
     * {@see MDCLineRippleAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCLineRippleAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCLineRippleAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        hasClass: () => {},
        setStyle: () => {},
        registerEventHandler: () => {},
        deregisterEventHandler: () => {},
      });
    }

    /**
     * @param {!MDCLineRippleAdapter=} adapter
     */
    constructor(adapter = /** @type {!MDCLineRippleAdapter} */ ({})) {
      super(Object.assign(MDCLineRippleFoundation.defaultAdapter, adapter));

      /** @private {function(!Event): undefined} */
      this.transitionEndHandler_ = (evt) => this.handleTransitionEnd(evt);
    }

    init() {
      this.adapter_.registerEventHandler('transitionend', this.transitionEndHandler_);
    }

    destroy() {
      this.adapter_.deregisterEventHandler('transitionend', this.transitionEndHandler_);
    }

    /**
     * Activates the line ripple
     */
    activate() {
      this.adapter_.removeClass(cssClasses$1.LINE_RIPPLE_DEACTIVATING);
      this.adapter_.addClass(cssClasses$1.LINE_RIPPLE_ACTIVE);
    }

    /**
     * Sets the center of the ripple animation to the given X coordinate.
     * @param {number} xCoordinate
     */
    setRippleCenter(xCoordinate) {
      this.adapter_.setStyle('transform-origin', `${xCoordinate}px center`);
    }

    /**
     * Deactivates the line ripple
     */
    deactivate() {
      this.adapter_.addClass(cssClasses$1.LINE_RIPPLE_DEACTIVATING);
    }

    /**
     * Handles a transition end event
     * @param {!Event} evt
     */
    handleTransitionEnd(evt) {
      // Wait for the line ripple to be either transparent or opaque
      // before emitting the animation end event
      const isDeactivating = this.adapter_.hasClass(cssClasses$1.LINE_RIPPLE_DEACTIVATING);

      if (evt.propertyName === 'opacity') {
        if (isDeactivating) {
          this.adapter_.removeClass(cssClasses$1.LINE_RIPPLE_ACTIVE);
          this.adapter_.removeClass(cssClasses$1.LINE_RIPPLE_DEACTIVATING);
        }
      }
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCLineRippleFoundation>}
   * @final
   */
  class MDCLineRipple extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCLineRipple}
     */
    static attachTo(root) {
      return new MDCLineRipple(root);
    }

    /**
     * Activates the line ripple
     */
    activate() {
      this.foundation_.activate();
    }

    /**
     * Deactivates the line ripple
     */
    deactivate() {
      this.foundation_.deactivate();
    }

    /**
     * Sets the transform origin given a user's click location. The `rippleCenter` is the
     * x-coordinate of the middle of the ripple.
     * @param {number} xCoordinate
     */
    setRippleCenter(xCoordinate) {
      this.foundation_.setRippleCenter(xCoordinate);
    }

    /**
     * @return {!MDCLineRippleFoundation}
     */
    getDefaultFoundation() {
      return new MDCLineRippleFoundation(/** @type {!MDCLineRippleAdapter} */ (Object.assign({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        hasClass: (className) => this.root_.classList.contains(className),
        setStyle: (propertyName, value) => this.root_.style[propertyName] = value,
        registerEventHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterEventHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
      })));
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  const cssClasses$2 = {
    // Ripple is a special case where the "root" component is really a "mixin" of sorts,
    // given that it's an 'upgrade' to an existing component. That being said it is the root
    // CSS class that all other CSS classes derive from.
    ROOT: 'mdc-ripple-upgraded',
    UNBOUNDED: 'mdc-ripple-upgraded--unbounded',
    BG_FOCUSED: 'mdc-ripple-upgraded--background-focused',
    FG_ACTIVATION: 'mdc-ripple-upgraded--foreground-activation',
    FG_DEACTIVATION: 'mdc-ripple-upgraded--foreground-deactivation',
  };

  const strings = {
    VAR_LEFT: '--mdc-ripple-left',
    VAR_TOP: '--mdc-ripple-top',
    VAR_FG_SIZE: '--mdc-ripple-fg-size',
    VAR_FG_SCALE: '--mdc-ripple-fg-scale',
    VAR_FG_TRANSLATE_START: '--mdc-ripple-fg-translate-start',
    VAR_FG_TRANSLATE_END: '--mdc-ripple-fg-translate-end',
  };

  const numbers = {
    PADDING: 10,
    INITIAL_ORIGIN_SCALE: 0.6,
    DEACTIVATION_TIMEOUT_MS: 225, // Corresponds to $mdc-ripple-translate-duration (i.e. activation animation duration)
    FG_DEACTIVATION_MS: 150, // Corresponds to $mdc-ripple-fade-out-duration (i.e. deactivation animation duration)
    TAP_DELAY_MS: 300, // Delay between touch and simulated mouse events on touch devices
  };

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Stores result from supportsCssVariables to avoid redundant processing to detect CSS custom variable support.
   * @private {boolean|undefined}
   */
  let supportsCssVariables_;

  /**
   * Stores result from applyPassive to avoid redundant processing to detect passive event listener support.
   * @private {boolean|undefined}
   */
  let supportsPassive_;

  /**
   * @param {!Window} windowObj
   * @return {boolean}
   */
  function detectEdgePseudoVarBug(windowObj) {
    // Detect versions of Edge with buggy var() support
    // See: https://developer.microsoft.com/en-us/microsoft-edge/platform/issues/11495448/
    const document = windowObj.document;
    const node = document.createElement('div');
    node.className = 'mdc-ripple-surface--test-edge-var-bug';
    document.body.appendChild(node);

    // The bug exists if ::before style ends up propagating to the parent element.
    // Additionally, getComputedStyle returns null in iframes with display: "none" in Firefox,
    // but Firefox is known to support CSS custom properties correctly.
    // See: https://bugzilla.mozilla.org/show_bug.cgi?id=548397
    const computedStyle = windowObj.getComputedStyle(node);
    const hasPseudoVarBug = computedStyle !== null && computedStyle.borderTopStyle === 'solid';
    node.remove();
    return hasPseudoVarBug;
  }

  /**
   * @param {!Window} windowObj
   * @param {boolean=} forceRefresh
   * @return {boolean|undefined}
   */

  function supportsCssVariables(windowObj, forceRefresh = false) {
    let supportsCssVariables = supportsCssVariables_;
    if (typeof supportsCssVariables_ === 'boolean' && !forceRefresh) {
      return supportsCssVariables;
    }

    const supportsFunctionPresent = windowObj.CSS && typeof windowObj.CSS.supports === 'function';
    if (!supportsFunctionPresent) {
      return;
    }

    const explicitlySupportsCssVars = windowObj.CSS.supports('--css-vars', 'yes');
    // See: https://bugs.webkit.org/show_bug.cgi?id=154669
    // See: README section on Safari
    const weAreFeatureDetectingSafari10plus = (
      windowObj.CSS.supports('(--css-vars: yes)') &&
      windowObj.CSS.supports('color', '#00000000')
    );

    if (explicitlySupportsCssVars || weAreFeatureDetectingSafari10plus) {
      supportsCssVariables = !detectEdgePseudoVarBug(windowObj);
    } else {
      supportsCssVariables = false;
    }

    if (!forceRefresh) {
      supportsCssVariables_ = supportsCssVariables;
    }
    return supportsCssVariables;
  }

  //
  /**
   * Determine whether the current browser supports passive event listeners, and if so, use them.
   * @param {!Window=} globalObj
   * @param {boolean=} forceRefresh
   * @return {boolean|{passive: boolean}}
   */
  function applyPassive(globalObj = window, forceRefresh = false) {
    if (supportsPassive_ === undefined || forceRefresh) {
      let isSupported = false;
      try {
        globalObj.document.addEventListener('test', null, {get passive() {
          isSupported = true;
        }});
      } catch (e) { }

      supportsPassive_ = isSupported;
    }

    return supportsPassive_ ? {passive: true} : false;
  }

  /**
   * @param {!Object} HTMLElementPrototype
   * @return {!Array<string>}
   */
  function getMatchesProperty(HTMLElementPrototype) {
    return [
      'webkitMatchesSelector', 'msMatchesSelector', 'matches',
    ].filter((p) => p in HTMLElementPrototype).pop();
  }

  /**
   * @param {!Event} ev
   * @param {{x: number, y: number}} pageOffset
   * @param {!ClientRect} clientRect
   * @return {{x: number, y: number}}
   */
  function getNormalizedEventCoords(ev, pageOffset, clientRect) {
    const {x, y} = pageOffset;
    const documentX = x + clientRect.left;
    const documentY = y + clientRect.top;

    let normalizedX;
    let normalizedY;
    // Determine touch point relative to the ripple container.
    if (ev.type === 'touchstart') {
      normalizedX = ev.changedTouches[0].pageX - documentX;
      normalizedY = ev.changedTouches[0].pageY - documentY;
    } else {
      normalizedX = ev.pageX - documentX;
      normalizedY = ev.pageY - documentY;
    }

    return {x: normalizedX, y: normalizedY};
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  // Activation events registered on the root element of each instance for activation
  const ACTIVATION_EVENT_TYPES = ['touchstart', 'pointerdown', 'mousedown', 'keydown'];

  // Deactivation events registered on documentElement when a pointer-related down event occurs
  const POINTER_DEACTIVATION_EVENT_TYPES = ['touchend', 'pointerup', 'mouseup'];

  // Tracks activations that have occurred on the current frame, to avoid simultaneous nested activations
  /** @type {!Array<!EventTarget>} */
  let activatedTargets = [];

  /**
   * @extends {MDCFoundation<!MDCRippleAdapter>}
   */
  class MDCRippleFoundation extends MDCFoundation {
    static get cssClasses() {
      return cssClasses$2;
    }

    static get strings() {
      return strings;
    }

    static get numbers() {
      return numbers;
    }

    static get defaultAdapter() {
      return {
        browserSupportsCssVars: () => /* boolean - cached */ {},
        isUnbounded: () => /* boolean */ {},
        isSurfaceActive: () => /* boolean */ {},
        isSurfaceDisabled: () => /* boolean */ {},
        addClass: (/* className: string */) => {},
        removeClass: (/* className: string */) => {},
        containsEventTarget: (/* target: !EventTarget */) => {},
        registerInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerResizeHandler: (/* handler: EventListener */) => {},
        deregisterResizeHandler: (/* handler: EventListener */) => {},
        updateCssVariable: (/* varName: string, value: string */) => {},
        computeBoundingRect: () => /* ClientRect */ {},
        getWindowPageOffset: () => /* {x: number, y: number} */ {},
      };
    }

    constructor(adapter) {
      super(Object.assign(MDCRippleFoundation.defaultAdapter, adapter));

      /** @private {number} */
      this.layoutFrame_ = 0;

      /** @private {!ClientRect} */
      this.frame_ = /** @type {!ClientRect} */ ({width: 0, height: 0});

      /** @private {!ActivationStateType} */
      this.activationState_ = this.defaultActivationState_();

      /** @private {number} */
      this.initialSize_ = 0;

      /** @private {number} */
      this.maxRadius_ = 0;

      /** @private {function(!Event)} */
      this.activateHandler_ = (e) => this.activate_(e);

      /** @private {function(!Event)} */
      this.deactivateHandler_ = (e) => this.deactivate_(e);

      /** @private {function(?Event=)} */
      this.focusHandler_ = () => this.handleFocus();

      /** @private {function(?Event=)} */
      this.blurHandler_ = () => this.handleBlur();

      /** @private {!Function} */
      this.resizeHandler_ = () => this.layout();

      /** @private {{left: number, top:number}} */
      this.unboundedCoords_ = {
        left: 0,
        top: 0,
      };

      /** @private {number} */
      this.fgScale_ = 0;

      /** @private {number} */
      this.activationTimer_ = 0;

      /** @private {number} */
      this.fgDeactivationRemovalTimer_ = 0;

      /** @private {boolean} */
      this.activationAnimationHasEnded_ = false;

      /** @private {!Function} */
      this.activationTimerCallback_ = () => {
        this.activationAnimationHasEnded_ = true;
        this.runDeactivationUXLogicIfReady_();
      };

      /** @private {?Event} */
      this.previousActivationEvent_ = null;
    }

    /**
     * We compute this property so that we are not querying information about the client
     * until the point in time where the foundation requests it. This prevents scenarios where
     * client-side feature-detection may happen too early, such as when components are rendered on the server
     * and then initialized at mount time on the client.
     * @return {boolean}
     * @private
     */
    isSupported_() {
      return this.adapter_.browserSupportsCssVars();
    }

    /**
     * @return {!ActivationStateType}
     */
    defaultActivationState_() {
      return {
        isActivated: false,
        hasDeactivationUXRun: false,
        wasActivatedByPointer: false,
        wasElementMadeActive: false,
        activationEvent: null,
        isProgrammatic: false,
      };
    }

    /** @override */
    init() {
      if (!this.isSupported_()) {
        return;
      }
      this.registerRootHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.addClass(ROOT);
        if (this.adapter_.isUnbounded()) {
          this.adapter_.addClass(UNBOUNDED);
          // Unbounded ripples need layout logic applied immediately to set coordinates for both shade and ripple
          this.layoutInternal_();
        }
      });
    }

    /** @override */
    destroy() {
      if (!this.isSupported_()) {
        return;
      }

      if (this.activationTimer_) {
        clearTimeout(this.activationTimer_);
        this.activationTimer_ = 0;
        const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
        this.adapter_.removeClass(FG_ACTIVATION);
      }

      this.deregisterRootHandlers_();
      this.deregisterDeactivationHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.removeClass(ROOT);
        this.adapter_.removeClass(UNBOUNDED);
        this.removeCssVars_();
      });
    }

    /** @private */
    registerRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.registerInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.registerInteractionHandler('focus', this.focusHandler_);
      this.adapter_.registerInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.registerResizeHandler(this.resizeHandler_);
      }
    }

    /**
     * @param {!Event} e
     * @private
     */
    registerDeactivationHandlers_(e) {
      if (e.type === 'keydown') {
        this.adapter_.registerInteractionHandler('keyup', this.deactivateHandler_);
      } else {
        POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
          this.adapter_.registerDocumentInteractionHandler(type, this.deactivateHandler_);
        });
      }
    }

    /** @private */
    deregisterRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.deregisterInteractionHandler('focus', this.focusHandler_);
      this.adapter_.deregisterInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.deregisterResizeHandler(this.resizeHandler_);
      }
    }

    /** @private */
    deregisterDeactivationHandlers_() {
      this.adapter_.deregisterInteractionHandler('keyup', this.deactivateHandler_);
      POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterDocumentInteractionHandler(type, this.deactivateHandler_);
      });
    }

    /** @private */
    removeCssVars_() {
      const {strings: strings$$1} = MDCRippleFoundation;
      Object.keys(strings$$1).forEach((k) => {
        if (k.indexOf('VAR_') === 0) {
          this.adapter_.updateCssVariable(strings$$1[k], null);
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    activate_(e) {
      if (this.adapter_.isSurfaceDisabled()) {
        return;
      }

      const activationState = this.activationState_;
      if (activationState.isActivated) {
        return;
      }

      // Avoid reacting to follow-on events fired by touch device after an already-processed user interaction
      const previousActivationEvent = this.previousActivationEvent_;
      const isSameInteraction = previousActivationEvent && e && previousActivationEvent.type !== e.type;
      if (isSameInteraction) {
        return;
      }

      activationState.isActivated = true;
      activationState.isProgrammatic = e === null;
      activationState.activationEvent = e;
      activationState.wasActivatedByPointer = activationState.isProgrammatic ? false : (
        e.type === 'mousedown' || e.type === 'touchstart' || e.type === 'pointerdown'
      );

      const hasActivatedChild =
        e && activatedTargets.length > 0 && activatedTargets.some((target) => this.adapter_.containsEventTarget(target));
      if (hasActivatedChild) {
        // Immediately reset activation state, while preserving logic that prevents touch follow-on events
        this.resetActivationState_();
        return;
      }

      if (e) {
        activatedTargets.push(/** @type {!EventTarget} */ (e.target));
        this.registerDeactivationHandlers_(e);
      }

      activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
      if (activationState.wasElementMadeActive) {
        this.animateActivation_();
      }

      requestAnimationFrame(() => {
        // Reset array on next frame after the current event has had a chance to bubble to prevent ancestor ripples
        activatedTargets = [];

        if (!activationState.wasElementMadeActive && (e.key === ' ' || e.keyCode === 32)) {
          // If space was pressed, try again within an rAF call to detect :active, because different UAs report
          // active states inconsistently when they're called within event handling code:
          // - https://bugs.chromium.org/p/chromium/issues/detail?id=635971
          // - https://bugzilla.mozilla.org/show_bug.cgi?id=1293741
          // We try first outside rAF to support Edge, which does not exhibit this problem, but will crash if a CSS
          // variable is set within a rAF callback for a submit button interaction (#2241).
          activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
          if (activationState.wasElementMadeActive) {
            this.animateActivation_();
          }
        }

        if (!activationState.wasElementMadeActive) {
          // Reset activation state immediately if element was not made active.
          this.activationState_ = this.defaultActivationState_();
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    checkElementMadeActive_(e) {
      return (e && e.type === 'keydown') ? this.adapter_.isSurfaceActive() : true;
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    activate(event = null) {
      this.activate_(event);
    }

    /** @private */
    animateActivation_() {
      const {VAR_FG_TRANSLATE_START, VAR_FG_TRANSLATE_END} = MDCRippleFoundation.strings;
      const {FG_DEACTIVATION, FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      const {DEACTIVATION_TIMEOUT_MS} = MDCRippleFoundation.numbers;

      this.layoutInternal_();

      let translateStart = '';
      let translateEnd = '';

      if (!this.adapter_.isUnbounded()) {
        const {startPoint, endPoint} = this.getFgTranslationCoordinates_();
        translateStart = `${startPoint.x}px, ${startPoint.y}px`;
        translateEnd = `${endPoint.x}px, ${endPoint.y}px`;
      }

      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_START, translateStart);
      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_END, translateEnd);
      // Cancel any ongoing activation/deactivation animations
      clearTimeout(this.activationTimer_);
      clearTimeout(this.fgDeactivationRemovalTimer_);
      this.rmBoundedActivationClasses_();
      this.adapter_.removeClass(FG_DEACTIVATION);

      // Force layout in order to re-trigger the animation.
      this.adapter_.computeBoundingRect();
      this.adapter_.addClass(FG_ACTIVATION);
      this.activationTimer_ = setTimeout(() => this.activationTimerCallback_(), DEACTIVATION_TIMEOUT_MS);
    }

    /**
     * @private
     * @return {{startPoint: PointType, endPoint: PointType}}
     */
    getFgTranslationCoordinates_() {
      const {activationEvent, wasActivatedByPointer} = this.activationState_;

      let startPoint;
      if (wasActivatedByPointer) {
        startPoint = getNormalizedEventCoords(
          /** @type {!Event} */ (activationEvent),
          this.adapter_.getWindowPageOffset(), this.adapter_.computeBoundingRect()
        );
      } else {
        startPoint = {
          x: this.frame_.width / 2,
          y: this.frame_.height / 2,
        };
      }
      // Center the element around the start point.
      startPoint = {
        x: startPoint.x - (this.initialSize_ / 2),
        y: startPoint.y - (this.initialSize_ / 2),
      };

      const endPoint = {
        x: (this.frame_.width / 2) - (this.initialSize_ / 2),
        y: (this.frame_.height / 2) - (this.initialSize_ / 2),
      };

      return {startPoint, endPoint};
    }

    /** @private */
    runDeactivationUXLogicIfReady_() {
      // This method is called both when a pointing device is released, and when the activation animation ends.
      // The deactivation animation should only run after both of those occur.
      const {FG_DEACTIVATION} = MDCRippleFoundation.cssClasses;
      const {hasDeactivationUXRun, isActivated} = this.activationState_;
      const activationHasEnded = hasDeactivationUXRun || !isActivated;

      if (activationHasEnded && this.activationAnimationHasEnded_) {
        this.rmBoundedActivationClasses_();
        this.adapter_.addClass(FG_DEACTIVATION);
        this.fgDeactivationRemovalTimer_ = setTimeout(() => {
          this.adapter_.removeClass(FG_DEACTIVATION);
        }, numbers.FG_DEACTIVATION_MS);
      }
    }

    /** @private */
    rmBoundedActivationClasses_() {
      const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      this.adapter_.removeClass(FG_ACTIVATION);
      this.activationAnimationHasEnded_ = false;
      this.adapter_.computeBoundingRect();
    }

    resetActivationState_() {
      this.previousActivationEvent_ = this.activationState_.activationEvent;
      this.activationState_ = this.defaultActivationState_();
      // Touch devices may fire additional events for the same interaction within a short time.
      // Store the previous event until it's safe to assume that subsequent events are for new interactions.
      setTimeout(() => this.previousActivationEvent_ = null, MDCRippleFoundation.numbers.TAP_DELAY_MS);
    }

    /**
     * @param {?Event} e
     * @private
     */
    deactivate_(e) {
      const activationState = this.activationState_;
      // This can happen in scenarios such as when you have a keyup event that blurs the element.
      if (!activationState.isActivated) {
        return;
      }

      const state = /** @type {!ActivationStateType} */ (Object.assign({}, activationState));

      if (activationState.isProgrammatic) {
        const evtObject = null;
        requestAnimationFrame(() => this.animateDeactivation_(evtObject, state));
        this.resetActivationState_();
      } else {
        this.deregisterDeactivationHandlers_();
        requestAnimationFrame(() => {
          this.activationState_.hasDeactivationUXRun = true;
          this.animateDeactivation_(e, state);
          this.resetActivationState_();
        });
      }
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    deactivate(event = null) {
      this.deactivate_(event);
    }

    /**
     * @param {Event} e
     * @param {!ActivationStateType} options
     * @private
     */
    animateDeactivation_(e, {wasActivatedByPointer, wasElementMadeActive}) {
      if (wasActivatedByPointer || wasElementMadeActive) {
        this.runDeactivationUXLogicIfReady_();
      }
    }

    layout() {
      if (this.layoutFrame_) {
        cancelAnimationFrame(this.layoutFrame_);
      }
      this.layoutFrame_ = requestAnimationFrame(() => {
        this.layoutInternal_();
        this.layoutFrame_ = 0;
      });
    }

    /** @private */
    layoutInternal_() {
      this.frame_ = this.adapter_.computeBoundingRect();
      const maxDim = Math.max(this.frame_.height, this.frame_.width);

      // Surface diameter is treated differently for unbounded vs. bounded ripples.
      // Unbounded ripple diameter is calculated smaller since the surface is expected to already be padded appropriately
      // to extend the hitbox, and the ripple is expected to meet the edges of the padded hitbox (which is typically
      // square). Bounded ripples, on the other hand, are fully expected to expand beyond the surface's longest diameter
      // (calculated based on the diagonal plus a constant padding), and are clipped at the surface's border via
      // `overflow: hidden`.
      const getBoundedRadius = () => {
        const hypotenuse = Math.sqrt(Math.pow(this.frame_.width, 2) + Math.pow(this.frame_.height, 2));
        return hypotenuse + MDCRippleFoundation.numbers.PADDING;
      };

      this.maxRadius_ = this.adapter_.isUnbounded() ? maxDim : getBoundedRadius();

      // Ripple is sized as a fraction of the largest dimension of the surface, then scales up using a CSS scale transform
      this.initialSize_ = maxDim * MDCRippleFoundation.numbers.INITIAL_ORIGIN_SCALE;
      this.fgScale_ = this.maxRadius_ / this.initialSize_;

      this.updateLayoutCssVars_();
    }

    /** @private */
    updateLayoutCssVars_() {
      const {
        VAR_FG_SIZE, VAR_LEFT, VAR_TOP, VAR_FG_SCALE,
      } = MDCRippleFoundation.strings;

      this.adapter_.updateCssVariable(VAR_FG_SIZE, `${this.initialSize_}px`);
      this.adapter_.updateCssVariable(VAR_FG_SCALE, this.fgScale_);

      if (this.adapter_.isUnbounded()) {
        this.unboundedCoords_ = {
          left: Math.round((this.frame_.width / 2) - (this.initialSize_ / 2)),
          top: Math.round((this.frame_.height / 2) - (this.initialSize_ / 2)),
        };

        this.adapter_.updateCssVariable(VAR_LEFT, `${this.unboundedCoords_.left}px`);
        this.adapter_.updateCssVariable(VAR_TOP, `${this.unboundedCoords_.top}px`);
      }
    }

    /** @param {boolean} unbounded */
    setUnbounded(unbounded) {
      const {UNBOUNDED} = MDCRippleFoundation.cssClasses;
      if (unbounded) {
        this.adapter_.addClass(UNBOUNDED);
      } else {
        this.adapter_.removeClass(UNBOUNDED);
      }
    }

    handleFocus() {
      requestAnimationFrame(() =>
        this.adapter_.addClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }

    handleBlur() {
      requestAnimationFrame(() =>
        this.adapter_.removeClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends MDCComponent<!MDCRippleFoundation>
   */
  class MDCRipple extends MDCComponent {
    /** @param {...?} args */
    constructor(...args) {
      super(...args);

      /** @type {boolean} */
      this.disabled = false;

      /** @private {boolean} */
      this.unbounded_;
    }

    /**
     * @param {!Element} root
     * @param {{isUnbounded: (boolean|undefined)}=} options
     * @return {!MDCRipple}
     */
    static attachTo(root, {isUnbounded = undefined} = {}) {
      const ripple = new MDCRipple(root);
      // Only override unbounded behavior if option is explicitly specified
      if (isUnbounded !== undefined) {
        ripple.unbounded = /** @type {boolean} */ (isUnbounded);
      }
      return ripple;
    }

    /**
     * @param {!RippleCapableSurface} instance
     * @return {!MDCRippleAdapter}
     */
    static createAdapter(instance) {
      const MATCHES = getMatchesProperty(HTMLElement.prototype);

      return {
        browserSupportsCssVars: () => supportsCssVariables(window),
        isUnbounded: () => instance.unbounded,
        isSurfaceActive: () => instance.root_[MATCHES](':active'),
        isSurfaceDisabled: () => instance.disabled,
        addClass: (className) => instance.root_.classList.add(className),
        removeClass: (className) => instance.root_.classList.remove(className),
        containsEventTarget: (target) => instance.root_.contains(target),
        registerInteractionHandler: (evtType, handler) =>
          instance.root_.addEventListener(evtType, handler, applyPassive()),
        deregisterInteractionHandler: (evtType, handler) =>
          instance.root_.removeEventListener(evtType, handler, applyPassive()),
        registerDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.addEventListener(evtType, handler, applyPassive()),
        deregisterDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.removeEventListener(evtType, handler, applyPassive()),
        registerResizeHandler: (handler) => window.addEventListener('resize', handler),
        deregisterResizeHandler: (handler) => window.removeEventListener('resize', handler),
        updateCssVariable: (varName, value) => instance.root_.style.setProperty(varName, value),
        computeBoundingRect: () => instance.root_.getBoundingClientRect(),
        getWindowPageOffset: () => ({x: window.pageXOffset, y: window.pageYOffset}),
      };
    }

    /** @return {boolean} */
    get unbounded() {
      return this.unbounded_;
    }

    /** @param {boolean} unbounded */
    set unbounded(unbounded) {
      this.unbounded_ = Boolean(unbounded);
      this.setUnbounded_();
    }

    /**
     * Closure Compiler throws an access control error when directly accessing a
     * protected or private property inside a getter/setter, like unbounded above.
     * By accessing the protected property inside a method, we solve that problem.
     * That's why this function exists.
     * @private
     */
    setUnbounded_() {
      this.foundation_.setUnbounded(this.unbounded_);
    }

    activate() {
      this.foundation_.activate();
    }

    deactivate() {
      this.foundation_.deactivate();
    }

    layout() {
      this.foundation_.layout();
    }

    /**
     * @return {!MDCRippleFoundation}
     * @override
     */
    getDefaultFoundation() {
      return new MDCRippleFoundation(MDCRipple.createAdapter(this));
    }

    /** @override */
    initialSyncWithDOM() {
      this.unbounded = 'mdcRippleIsUnbounded' in this.root_.dataset;
    }
  }

  /**
   * See Material Design spec for more details on when to use ripples.
   * https://material.io/guidelines/motion/choreography.html#choreography-creation
   * @record
   */
  class RippleCapableSurface {}

  /** @protected {!Element} */
  RippleCapableSurface.prototype.root_;

  /**
   * Whether or not the ripple bleeds out of the bounds of the element.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.unbounded;

  /**
   * Whether or not the ripple is attached to a disabled component.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.disabled;

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const strings$1 = {
    PATH_SELECTOR: '.mdc-notched-outline__path',
    IDLE_OUTLINE_SELECTOR: '.mdc-notched-outline__idle',
  };

  /** @enum {string} */
  const cssClasses$3 = {
    OUTLINE_NOTCHED: 'mdc-notched-outline--notched',
  };

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCFoundation<!MDCNotchedOutlineAdapter>}
   * @final
   */
  class MDCNotchedOutlineFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get strings() {
      return strings$1;
    }

    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$3;
    }

    /**
     * {@see MDCNotchedOutlineAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCNotchedOutlineAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCNotchedOutlineAdapter} */ ({
        getWidth: () => {},
        getHeight: () => {},
        addClass: () => {},
        removeClass: () => {},
        setOutlinePathAttr: () => {},
        getIdleOutlineStyleValue: () => {},
      });
    }

    /**
     * @param {!MDCNotchedOutlineAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCNotchedOutlineFoundation.defaultAdapter, adapter));
    }

    /**
     * Adds the outline notched selector and updates the notch width
     * calculated based off of notchWidth and isRtl.
     * @param {number} notchWidth
     * @param {boolean=} isRtl
     */
    notch(notchWidth, isRtl = false) {
      const {OUTLINE_NOTCHED} = MDCNotchedOutlineFoundation.cssClasses;
      this.adapter_.addClass(OUTLINE_NOTCHED);
      this.updateSvgPath_(notchWidth, isRtl);
    }

    /**
     * Removes notched outline selector to close the notch in the outline.
     */
    closeNotch() {
      const {OUTLINE_NOTCHED} = MDCNotchedOutlineFoundation.cssClasses;
      this.adapter_.removeClass(OUTLINE_NOTCHED);
    }

    /**
     * Updates the SVG path of the focus outline element based on the notchWidth
     * and the RTL context.
     * @param {number} notchWidth
     * @param {boolean=} isRtl
     * @private
     */
    updateSvgPath_(notchWidth, isRtl) {
      // Fall back to reading a specific corner's style because Firefox doesn't report the style on border-radius.
      const radiusStyleValue = this.adapter_.getIdleOutlineStyleValue('border-radius') ||
          this.adapter_.getIdleOutlineStyleValue('border-top-left-radius');
      const radius = parseFloat(radiusStyleValue);
      const width = this.adapter_.getWidth();
      const height = this.adapter_.getHeight();
      const cornerWidth = radius + 1.2;
      const leadingStrokeLength = Math.abs(11 - cornerWidth);
      const paddedNotchWidth = notchWidth + 8;

      // The right, bottom, and left sides of the outline follow the same SVG path.
      const pathMiddle = 'a' + radius + ',' + radius + ' 0 0 1 ' + radius + ',' + radius
        + 'v' + (height - (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + -radius + ',' + radius
        + 'h' + (-width + (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + -radius + ',' + -radius
        + 'v' + (-height + (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + radius + ',' + -radius;

      let path;
      if (!isRtl) {
        path = 'M' + (cornerWidth + leadingStrokeLength + paddedNotchWidth) + ',' + 1
          + 'h' + (width - (2 * cornerWidth) - paddedNotchWidth - leadingStrokeLength)
          + pathMiddle
          + 'h' + leadingStrokeLength;
      } else {
        path = 'M' + (width - cornerWidth - leadingStrokeLength) + ',' + 1
          + 'h' + leadingStrokeLength
          + pathMiddle
          + 'h' + (width - (2 * cornerWidth) - paddedNotchWidth - leadingStrokeLength);
      }

      this.adapter_.setOutlinePathAttr(path);
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCNotchedOutlineFoundation>}
   * @final
   */
  class MDCNotchedOutline extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCNotchedOutline}
     */
    static attachTo(root) {
      return new MDCNotchedOutline(root);
    }

    /**
      * Updates outline selectors and SVG path to open notch.
      * @param {number} notchWidth The notch width in the outline.
      * @param {boolean=} isRtl Determines if outline is rtl. If rtl is true, notch
      * will be right justified in outline path, otherwise left justified.
      */
    notch(notchWidth, isRtl) {
      this.foundation_.notch(notchWidth, isRtl);
    }

    /**
     * Updates the outline selectors to close notch and return it to idle state.
     */
    closeNotch() {
      this.foundation_.closeNotch();
    }

    /**
     * @return {!MDCNotchedOutlineFoundation}
     */
    getDefaultFoundation() {
      return new MDCNotchedOutlineFoundation({
        getWidth: () => this.root_.offsetWidth,
        getHeight: () => this.root_.offsetHeight,
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        setOutlinePathAttr: (value) => {
          const path = this.root_.querySelector(strings$1.PATH_SELECTOR);
          path.setAttribute('d', value);
        },
        getIdleOutlineStyleValue: (propertyName) => {
          const idleOutlineElement = this.root_.parentNode.querySelector(strings$1.IDLE_OUTLINE_SELECTOR);
          return window.getComputedStyle(idleOutlineElement).getPropertyValue(propertyName);
        },
      });
    }
  }

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  const cssClasses$4 = {
    BOX: 'mdc-select--box',
    DISABLED: 'mdc-select--disabled',
    ROOT: 'mdc-select',
    OUTLINED: 'mdc-select--outlined',
  };

  const strings$2 = {
    CHANGE_EVENT: 'MDCSelect:change',
    LINE_RIPPLE_SELECTOR: '.mdc-line-ripple',
    LABEL_SELECTOR: '.mdc-floating-label',
    NATIVE_CONTROL_SELECTOR: '.mdc-select__native-control',
    OUTLINE_SELECTOR: '.mdc-notched-outline',
  };

  /** @enum {number} */
  const numbers$1 = {
    LABEL_SCALE: 0.75,
  };

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  class MDCSelectFoundation extends MDCFoundation {
    static get cssClasses() {
      return cssClasses$4;
    }

    static get numbers() {
      return numbers$1;
    }

    static get strings() {
      return strings$2;
    }

    static get defaultAdapter() {
      return {
        addClass: (/* className: string */) => {},
        removeClass: (/* className: string */) => {},
        hasClass: (/* className: string */) => false,
        floatLabel: (/* value: boolean */) => {},
        activateBottomLine: () => {},
        deactivateBottomLine: () => {},
        getValue: () => {},
        isRtl: () => false,
        hasLabel: () => {},
        getLabelWidth: () => {},
        hasOutline: () => {},
        notchOutline: () => {},
        closeOutline: () => {},
      };
    }

    constructor(adapter) {
      super(Object.assign(MDCSelectFoundation.defaultAdapter, adapter));

      this.focusHandler_ = (evt) => this.handleFocus_(evt);
      this.blurHandler_ = (evt) => this.handleBlur_(evt);
    }

    updateDisabledStyle(disabled) {
      const {DISABLED} = MDCSelectFoundation.cssClasses;
      if (disabled) {
        this.adapter_.addClass(DISABLED);
      } else {
        this.adapter_.removeClass(DISABLED);
      }
    }

    handleChange() {
      const optionHasValue = this.adapter_.getValue().length > 0;
      this.adapter_.floatLabel(optionHasValue);
      this.notchOutline(optionHasValue);
    }

    handleFocus() {
      this.adapter_.floatLabel(true);
      this.notchOutline(true);
      this.adapter_.activateBottomLine();
    }

    handleBlur() {
      this.handleChange();
      this.adapter_.deactivateBottomLine();
    }

    /**
     * Opens/closes the notched outline.
     * @param {boolean} openNotch
     */
    notchOutline(openNotch) {
      if (!this.adapter_.hasOutline() || !this.adapter_.hasLabel()) {
        return;
      }

      if (openNotch) {
        const labelScale = numbers$1.LABEL_SCALE;
        const labelWidth = this.adapter_.getLabelWidth() * labelScale;
        const isRtl = this.adapter_.isRtl();
        this.adapter_.notchOutline(labelWidth, isRtl);
      } else {
        this.adapter_.closeOutline();
      }
    }
  }

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  class MDCSelect extends MDCComponent {
    static attachTo(root) {
      return new MDCSelect(root);
    }

    get value() {
      return this.nativeControl_.value;
    }

    set value(value) {
      this.nativeControl_.value = value;
      this.foundation_.handleChange();
    }

    get selectedIndex() {
      return this.nativeControl_.selectedIndex;
    }

    set selectedIndex(selectedIndex) {
      this.nativeControl_.selectedIndex = selectedIndex;
      this.foundation_.handleChange();
    }

    get disabled() {
      return this.nativeControl_.disabled;
    }

    set disabled(disabled) {
      this.nativeControl_.disabled = disabled;
      this.foundation_.updateDisabledStyle(disabled);
    }

    /**
     * Recomputes the outline SVG path for the outline element.
     */
    layout() {
      const openNotch = this.nativeControl_.value.length > 0;
      this.foundation_.notchOutline(openNotch);
    }

    initialize(
      labelFactory = (el) => new MDCFloatingLabel(el),
      lineRippleFactory = (el) => new MDCLineRipple(el),
      outlineFactory = (el) => new MDCNotchedOutline(el)) {
      this.nativeControl_ = this.root_.querySelector(strings$2.NATIVE_CONTROL_SELECTOR);
      const labelElement = this.root_.querySelector(strings$2.LABEL_SELECTOR);
      if (labelElement) {
        this.label_ = labelFactory(labelElement);
      }
      const lineRippleElement = this.root_.querySelector(strings$2.LINE_RIPPLE_SELECTOR);
      if (lineRippleElement) {
        this.lineRipple_ = lineRippleFactory(lineRippleElement);
      }
      const outlineElement = this.root_.querySelector(strings$2.OUTLINE_SELECTOR);
      if (outlineElement) {
        this.outline_ = outlineFactory(outlineElement);
      }

      if (this.root_.classList.contains(cssClasses$4.BOX)) {
        this.ripple = this.initRipple_();
      }
    }

    initRipple_() {
      const adapter = Object.assign(MDCRipple.createAdapter(this), {
        registerInteractionHandler: (type, handler) => this.nativeControl_.addEventListener(type, handler),
        deregisterInteractionHandler: (type, handler) => this.nativeControl_.removeEventListener(type, handler),
      });
      const foundation = new MDCRippleFoundation(adapter);
      return new MDCRipple(this.root_, foundation);
    }

    initialSyncWithDOM() {
      this.handleChange_ = () => this.foundation_.handleChange();
      this.handleFocus_ = () => this.foundation_.handleFocus();
      this.handleBlur_ = () => this.foundation_.handleBlur();

      this.nativeControl_.addEventListener('change', this.handleChange_);
      this.nativeControl_.addEventListener('focus', this.handleFocus_);
      this.nativeControl_.addEventListener('blur', this.handleBlur_);

      // Initially sync floating label
      this.foundation_.handleChange();

      if (this.nativeControl_.disabled) {
        this.disabled = true;
      }
    }

    destroy() {
      this.nativeControl_.removeEventListener('change', this.handleChange_);
      this.nativeControl_.removeEventListener('focus', this.handleFocus_);
      this.nativeControl_.removeEventListener('blur', this.handleBlur_);

      if (this.ripple) {
        this.ripple.destroy();
      }
      if (this.outline_) {
        this.outline_.destroy();
      }

      super.destroy();
    }

    getDefaultFoundation() {
      return new MDCSelectFoundation((Object.assign({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        hasClass: (className) => this.root_.classList.contains(className),
        activateBottomLine: () => {
          if (this.lineRipple_) {
            this.lineRipple_.activate();
          }
        },
        deactivateBottomLine: () => {
          if (this.lineRipple_) {
            this.lineRipple_.deactivate();
          }
        },
        isRtl: () => window.getComputedStyle(this.root_).getPropertyValue('direction') === 'rtl',
        getValue: () => this.nativeControl_.value,
      },
      this.getOutlineAdapterMethods_(),
      this.getLabelAdapterMethods_()))
      );
    }

    /**
     * @return {!{
     *   notchOutline: function(number, boolean): undefined,
     *   hasOutline: function(): boolean,
     * }}
     */
    getOutlineAdapterMethods_() {
      return {
        notchOutline: (labelWidth, isRtl) => {
          if (this.outline_) {
            this.outline_.notch(labelWidth, isRtl);
          }
        },
        closeOutline: () => {
          if (this.outline_) {
            this.outline_.closeNotch();
          }
        },
        hasOutline: () => !!this.outline_,
      };
    }

    /**
     * @return {!{
     *   floatLabel: function(boolean): undefined,
     *   hasLabel: function(): boolean,
     *   getLabelWidth: function(): number,
     * }}
     */
    getLabelAdapterMethods_() {
      return {
        floatLabel: (shouldFloat) => {
          if (this.label_) {
            this.label_.float(shouldFloat);
          }
        },
        hasLabel: () => !!this.label_,
        getLabelWidth: () => {
          if (this.label_) {
            return this.label_.getWidth();
          }
        },
      };
    }
  }

  const regex = /%s/gi;

  function processClasses(classes, userClasses) {
      if (!userClasses) {
          return classes.join(" ");
      }
      if (Array.isArray(userClasses)) {
          userClasses = userClasses.join(" ");
      }
      if (!regex.test(userClasses)) {
          return userClasses+" "+classes.join(" ");
      }
      return userClasses.replace(regex, classes.join(" "));
  }

  /* src/Select.html generated by Svelte v2.9.11 */

  function attrs(attributes) {
      let result = Object.assign({}, attributes);
      let cls = "mdc-select";
      let classes = [cls];
      let key = "box";
      if (result[key]) {
          classes.push(cls+"--"+key);
      }
      delete result[key];
      for (let key of ["attrs", "id", "name", "value", "selectedIndex", "disabled"]) {
          delete result[key];
      }
      result["class"] = processClasses(classes, result["class"]);
      return result;
  }
  var methods = {
      getMDCField() {
          return this.mdcComponent;
      }
  };

  function oncreate() {
      this.mdcComponent = new MDCSelect(this.refs.select);
      this.mdcComponent.listen("change", () => {
          this.fire("change");
          this.set({
              "value": this.mdcComponent.value,
              "selectedIndex": this.mdcComponent.selectedIndex,
          });
      });
  }
  function ondestroy() {
      this.mdcComponent.destroy();
  }
  function onstate({current, changed}) {
      for (let key of ["value", "selectedIndex", "disabled"]) {
          if (changed[key]) {
              this.mdcComponent[key] = current[key];
          }
      }
  }
  function create_main_fragment(component, ctx) {
  	var div, select, slot_content_default = component._slotted.default, slot_content_default_after, text, label, text_1, text_2, div_1;

  	var div_levels = [
  		ctx.attrs
  	];

  	var div_data = {};
  	for (var i = 0; i < div_levels.length; i += 1) {
  		div_data = shared_js.assign(div_data, div_levels[i]);
  	}

  	return {
  		c() {
  			div = shared_js.createElement("div");
  			select = shared_js.createElement("select");
  			text = shared_js.createText("\n  ");
  			label = shared_js.createElement("label");
  			text_1 = shared_js.createText(ctx.label);
  			text_2 = shared_js.createText("\n  ");
  			div_1 = shared_js.createElement("div");
  			select.className = "mdc-select__native-control";
  			select.name = ctx.name;
  			select.id = ctx.id;
  			label.className = "mdc-floating-label";
  			div_1.className = "mdc-line-ripple";
  			shared_js.setAttributes(div, div_data);
  		},

  		m(target, anchor) {
  			shared_js.insert(target, div, anchor);
  			shared_js.append(div, select);

  			if (slot_content_default) {
  				shared_js.append(select, slot_content_default);
  				shared_js.append(select, slot_content_default_after || (slot_content_default_after = shared_js.createComment()));
  			}

  			shared_js.append(div, text);
  			shared_js.append(div, label);
  			shared_js.append(label, text_1);
  			shared_js.append(div, text_2);
  			shared_js.append(div, div_1);
  			component.refs.select = div;
  		},

  		p(changed, ctx) {
  			if (changed.name) {
  				select.name = ctx.name;
  			}

  			if (changed.id) {
  				select.id = ctx.id;
  			}

  			if (changed.label) {
  				shared_js.setData(text_1, ctx.label);
  			}

  			shared_js.setAttributes(div, shared_js.getSpreadUpdate(div_levels, [
  				(changed.attrs) && ctx.attrs
  			]));
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(div);
  			}

  			if (slot_content_default) {
  				shared_js.reinsertBefore(slot_content_default_after, slot_content_default);
  			}

  			if (component.refs.select === div) component.refs.select = null;
  		}
  	};
  }

  function Select(options) {
  	shared_js.init(this, options);
  	this.refs = {};
  	this._state = shared_js.assign({}, options.data);
  	this._recompute({  }, this._state);
  	this._intro = true;

  	this._handlers.state = [onstate];

  	this._handlers.destroy = [ondestroy];

  	this._slotted = options.slots || {};

  	if (!options.root) {
  		this._oncreate = [];
  	}

  	this.slots = {};

  	this._fragment = create_main_fragment(this, this._state);

  	this.root._oncreate.push(() => {
  		onstate.call(this, { changed: shared_js.assignTrue({}, this._state), current: this._state });
  		oncreate.call(this);
  		this.fire("update", { changed: shared_js.assignTrue({}, this._state), current: this._state });
  	});

  	if (options.target) {
  		this._fragment.c();
  		this._mount(options.target, options.anchor);

  		shared_js.callAll(this._oncreate);
  	}
  }

  shared_js.assign(Select.prototype, shared_js.proto);
  shared_js.assign(Select.prototype, methods);

  Select.prototype._recompute = function _recompute(changed, state) {
  	if (this._differs(state.attrs, (state.attrs = attrs(state)))) changed.attrs = true;
  };

  return Select;

})));
