(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('./shared.js')) :
  typeof define === 'function' && define.amd ? define(['./shared.js'], factory) :
  (global.Snackbar = factory(global.svelteShared));
}(this, (function (shared_js) { 'use strict';

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template A
   */
  class MDCFoundation {
    /** @return enum{cssClasses} */
    static get cssClasses() {
      // Classes extending MDCFoundation should implement this method to return an object which exports every
      // CSS class the foundation class needs as a property. e.g. {ACTIVE: 'mdc-component--active'}
      return {};
    }

    /** @return enum{strings} */
    static get strings() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // semantic strings as constants. e.g. {ARIA_ROLE: 'tablist'}
      return {};
    }

    /** @return enum{numbers} */
    static get numbers() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // of its semantic numbers as constants. e.g. {ANIMATION_DELAY_MS: 350}
      return {};
    }

    /** @return {!Object} */
    static get defaultAdapter() {
      // Classes extending MDCFoundation may choose to implement this getter in order to provide a convenient
      // way of viewing the necessary methods of an adapter. In the future, this could also be used for adapter
      // validation.
      return {};
    }

    /**
     * @param {A=} adapter
     */
    constructor(adapter = {}) {
      /** @protected {!A} */
      this.adapter_ = adapter;
    }

    init() {
      // Subclasses should override this method to perform initialization routines (registering events, etc.)
    }

    destroy() {
      // Subclasses should override this method to perform de-initialization routines (de-registering events, etc.)
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template F
   */
  class MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCComponent}
     */
    static attachTo(root) {
      // Subclasses which extend MDCBase should provide an attachTo() method that takes a root element and
      // returns an instantiated component with its root set to that element. Also note that in the cases of
      // subclasses, an explicit foundation class will not have to be passed in; it will simply be initialized
      // from getDefaultFoundation().
      return new MDCComponent(root, new MDCFoundation());
    }

    /**
     * @param {!Element} root
     * @param {F=} foundation
     * @param {...?} args
     */
    constructor(root, foundation = undefined, ...args) {
      /** @protected {!Element} */
      this.root_ = root;
      this.initialize(...args);
      // Note that we initialize foundation here and not within the constructor's default param so that
      // this.root_ is defined and can be used within the foundation class.
      /** @protected {!F} */
      this.foundation_ = foundation === undefined ? this.getDefaultFoundation() : foundation;
      this.foundation_.init();
      this.initialSyncWithDOM();
    }

    initialize(/* ...args */) {
      // Subclasses can override this to do any additional setup work that would be considered part of a
      // "constructor". Essentially, it is a hook into the parent constructor before the foundation is
      // initialized. Any additional arguments besides root and foundation will be passed in here.
    }

    /**
     * @return {!F} foundation
     */
    getDefaultFoundation() {
      // Subclasses must override this method to return a properly configured foundation class for the
      // component.
      throw new Error('Subclasses must override getDefaultFoundation to return a properly configured ' +
        'foundation class');
    }

    initialSyncWithDOM() {
      // Subclasses should override this method if they need to perform work to synchronize with a host DOM
      // object. An example of this would be a form control wrapper that needs to synchronize its internal state
      // to some property or attribute of the host DOM. Please note: this is *not* the place to perform DOM
      // reads/writes that would cause layout / paint, as this is called synchronously from within the constructor.
    }

    destroy() {
      // Subclasses may implement this method to release any resources / deregister any listeners they have
      // attached. An example of this might be deregistering a resize event from the window object.
      this.foundation_.destroy();
    }

    /**
     * Wrapper method to add an event listener to the component's root element. This is most useful when
     * listening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    listen(evtType, handler) {
      this.root_.addEventListener(evtType, handler);
    }

    /**
     * Wrapper method to remove an event listener to the component's root element. This is most useful when
     * unlistening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    unlisten(evtType, handler) {
      this.root_.removeEventListener(evtType, handler);
    }

    /**
     * Fires a cross-browser-compatible custom event from the component root of the given type,
     * with the given data.
     * @param {string} evtType
     * @param {!Object} evtData
     * @param {boolean=} shouldBubble
     */
    emit(evtType, evtData, shouldBubble = false) {
      let evt;
      if (typeof CustomEvent === 'function') {
        evt = new CustomEvent(evtType, {
          detail: evtData,
          bubbles: shouldBubble,
        });
      } else {
        evt = document.createEvent('CustomEvent');
        evt.initCustomEvent(evtType, shouldBubble, false, evtData);
      }

      this.root_.dispatchEvent(evt);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  const cssClasses = {
    ROOT: 'mdc-snackbar',
    TEXT: 'mdc-snackbar__text',
    ACTION_WRAPPER: 'mdc-snackbar__action-wrapper',
    ACTION_BUTTON: 'mdc-snackbar__action-button',
    ACTIVE: 'mdc-snackbar--active',
    MULTILINE: 'mdc-snackbar--multiline',
    ACTION_ON_BOTTOM: 'mdc-snackbar--action-on-bottom',
  };

  const strings = {
    TEXT_SELECTOR: '.mdc-snackbar__text',
    ACTION_WRAPPER_SELECTOR: '.mdc-snackbar__action-wrapper',
    ACTION_BUTTON_SELECTOR: '.mdc-snackbar__action-button',
    SHOW_EVENT: 'MDCSnackbar:show',
    HIDE_EVENT: 'MDCSnackbar:hide',
  };

  const numbers = {
    MESSAGE_TIMEOUT: 2750,
  };

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  class MDCSnackbarFoundation extends MDCFoundation {
    static get cssClasses() {
      return cssClasses;
    }

    static get strings() {
      return strings;
    }

    static get defaultAdapter() {
      return {
        addClass: (/* className: string */) => {},
        removeClass: (/* className: string */) => {},
        setAriaHidden: () => {},
        unsetAriaHidden: () => {},
        setActionAriaHidden: () => {},
        unsetActionAriaHidden: () => {},
        setActionText: (/* actionText: string */) => {},
        setMessageText: (/* message: string */) => {},
        setFocus: () => {},
        visibilityIsHidden: () => /* boolean */ false,
        registerCapturedBlurHandler: (/* handler: EventListener */) => {},
        deregisterCapturedBlurHandler: (/* handler: EventListener */) => {},
        registerVisibilityChangeHandler: (/* handler: EventListener */) => {},
        deregisterVisibilityChangeHandler: (/* handler: EventListener */) => {},
        registerCapturedInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterCapturedInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerActionClickHandler: (/* handler: EventListener */) => {},
        deregisterActionClickHandler: (/* handler: EventListener */) => {},
        registerTransitionEndHandler: (/* handler: EventListener */) => {},
        deregisterTransitionEndHandler: (/* handler: EventListener */) => {},
        notifyShow: () => {},
        notifyHide: () => {},
      };
    }

    get active() {
      return this.active_;
    }

    constructor(adapter) {
      super(Object.assign(MDCSnackbarFoundation.defaultAdapter, adapter));

      this.active_ = false;
      this.actionWasClicked_ = false;
      this.dismissOnAction_ = true;
      this.firstFocus_ = true;
      this.pointerDownRecognized_ = false;
      this.snackbarHasFocus_ = false;
      this.snackbarData_ = null;
      this.queue_ = [];
      this.actionClickHandler_ = () => {
        this.actionWasClicked_ = true;
        this.invokeAction_();
      };
      this.visibilitychangeHandler_ = () => {
        clearTimeout(this.timeoutId_);
        this.snackbarHasFocus_ = true;

        if (!this.adapter_.visibilityIsHidden()) {
          setTimeout(this.cleanup_.bind(this), this.snackbarData_.timeout || numbers.MESSAGE_TIMEOUT);
        }
      };
      this.interactionHandler_ = (evt) => {
        if (evt.type == 'touchstart' || evt.type == 'mousedown') {
          this.pointerDownRecognized_ = true;
        }
        this.handlePossibleTabKeyboardFocus_(evt);

        if (evt.type == 'focus') {
          this.pointerDownRecognized_ = false;
        }
      };
      this.blurHandler_ = () => {
        clearTimeout(this.timeoutId_);
        this.snackbarHasFocus_ = false;
        this.timeoutId_ = setTimeout(this.cleanup_.bind(this), this.snackbarData_.timeout || numbers.MESSAGE_TIMEOUT);
      };
    }

    init() {
      this.adapter_.registerActionClickHandler(this.actionClickHandler_);
      this.adapter_.setAriaHidden();
      this.adapter_.setActionAriaHidden();
    }

    destroy() {
      this.adapter_.deregisterActionClickHandler(this.actionClickHandler_);
      this.adapter_.deregisterCapturedBlurHandler(this.blurHandler_);
      this.adapter_.deregisterVisibilityChangeHandler(this.visibilitychangeHandler_);
      ['touchstart', 'mousedown', 'focus'].forEach((evtType) => {
        this.adapter_.deregisterCapturedInteractionHandler(evtType, this.interactionHandler_);
      });
    }

    dismissesOnAction() {
      return this.dismissOnAction_;
    }

    setDismissOnAction(dismissOnAction) {
      this.dismissOnAction_ = !!dismissOnAction;
    }

    show(data) {
      if (!data) {
        throw new Error(
          'Please provide a data object with at least a message to display.');
      }
      if (!data.message) {
        throw new Error('Please provide a message to be displayed.');
      }
      if (data.actionHandler && !data.actionText) {
        throw new Error('Please provide action text with the handler.');
      }
      if (this.active) {
        this.queue_.push(data);
        return;
      }
      clearTimeout(this.timeoutId_);
      this.snackbarData_ = data;
      this.firstFocus_ = true;
      this.adapter_.registerVisibilityChangeHandler(this.visibilitychangeHandler_);
      this.adapter_.registerCapturedBlurHandler(this.blurHandler_);
      ['touchstart', 'mousedown', 'focus'].forEach((evtType) => {
        this.adapter_.registerCapturedInteractionHandler(evtType, this.interactionHandler_);
      });

      const {ACTIVE, MULTILINE, ACTION_ON_BOTTOM} = cssClasses;

      this.adapter_.setMessageText(this.snackbarData_.message);

      if (this.snackbarData_.multiline) {
        this.adapter_.addClass(MULTILINE);
        if (this.snackbarData_.actionOnBottom) {
          this.adapter_.addClass(ACTION_ON_BOTTOM);
        }
      }

      if (this.snackbarData_.actionHandler) {
        this.adapter_.setActionText(this.snackbarData_.actionText);
        this.actionHandler_ = this.snackbarData_.actionHandler;
        this.setActionHidden_(false);
      } else {
        this.setActionHidden_(true);
        this.actionHandler_ = null;
        this.adapter_.setActionText(null);
      }

      this.active_ = true;
      this.adapter_.addClass(ACTIVE);
      this.adapter_.unsetAriaHidden();
      this.adapter_.notifyShow();

      this.timeoutId_ = setTimeout(this.cleanup_.bind(this), this.snackbarData_.timeout || numbers.MESSAGE_TIMEOUT);
    }

    handlePossibleTabKeyboardFocus_() {
      const hijackFocus =
        this.firstFocus_ && !this.pointerDownRecognized_;

      if (hijackFocus) {
        this.setFocusOnAction_();
      }

      this.firstFocus_ = false;
    }

    setFocusOnAction_() {
      this.adapter_.setFocus();
      this.snackbarHasFocus_ = true;
      this.firstFocus_ = false;
    }

    invokeAction_() {
      try {
        if (!this.actionHandler_) {
          return;
        }

        this.actionHandler_();
      } finally {
        if (this.dismissOnAction_) {
          this.cleanup_();
        }
      }
    }

    cleanup_() {
      const allowDismissal = !this.snackbarHasFocus_ || this.actionWasClicked_;

      if (allowDismissal) {
        const {ACTIVE, MULTILINE, ACTION_ON_BOTTOM} = cssClasses;

        this.adapter_.removeClass(ACTIVE);

        const handler = () => {
          clearTimeout(this.timeoutId_);
          this.adapter_.deregisterTransitionEndHandler(handler);
          this.adapter_.removeClass(MULTILINE);
          this.adapter_.removeClass(ACTION_ON_BOTTOM);
          this.setActionHidden_(true);
          this.adapter_.setAriaHidden();
          this.active_ = false;
          this.snackbarHasFocus_ = false;
          this.adapter_.notifyHide();
          this.showNext_();
        };

        this.adapter_.registerTransitionEndHandler(handler);
      }
    }

    showNext_() {
      if (!this.queue_.length) {
        return;
      }
      this.show(this.queue_.shift());
    }

    setActionHidden_(isHidden) {
      if (isHidden) {
        this.adapter_.setActionAriaHidden();
      } else {
        this.adapter_.unsetActionAriaHidden();
      }
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @const {Object<string, !VendorPropertyMapType>} */
  const eventTypeMap = {
    'animationstart': {
      noPrefix: 'animationstart',
      webkitPrefix: 'webkitAnimationStart',
      styleProperty: 'animation',
    },
    'animationend': {
      noPrefix: 'animationend',
      webkitPrefix: 'webkitAnimationEnd',
      styleProperty: 'animation',
    },
    'animationiteration': {
      noPrefix: 'animationiteration',
      webkitPrefix: 'webkitAnimationIteration',
      styleProperty: 'animation',
    },
    'transitionend': {
      noPrefix: 'transitionend',
      webkitPrefix: 'webkitTransitionEnd',
      styleProperty: 'transition',
    },
  };

  /** @const {Object<string, !VendorPropertyMapType>} */
  const cssPropertyMap = {
    'animation': {
      noPrefix: 'animation',
      webkitPrefix: '-webkit-animation',
    },
    'transform': {
      noPrefix: 'transform',
      webkitPrefix: '-webkit-transform',
    },
    'transition': {
      noPrefix: 'transition',
      webkitPrefix: '-webkit-transition',
    },
  };

  /**
   * @param {!Object} windowObj
   * @return {boolean}
   */
  function hasProperShape(windowObj) {
    return (windowObj['document'] !== undefined && typeof windowObj['document']['createElement'] === 'function');
  }

  /**
   * @param {string} eventType
   * @return {boolean}
   */
  function eventFoundInMaps(eventType) {
    return (eventType in eventTypeMap || eventType in cssPropertyMap);
  }

  /**
   * @param {string} eventType
   * @param {!Object<string, !VendorPropertyMapType>} map
   * @param {!Element} el
   * @return {string}
   */
  function getJavaScriptEventName(eventType, map, el) {
    return map[eventType].styleProperty in el.style ? map[eventType].noPrefix : map[eventType].webkitPrefix;
  }

  /**
   * Helper function to determine browser prefix for CSS3 animation events
   * and property names.
   * @param {!Object} windowObj
   * @param {string} eventType
   * @return {string}
   */
  function getAnimationName(windowObj, eventType) {
    if (!hasProperShape(windowObj) || !eventFoundInMaps(eventType)) {
      return eventType;
    }

    const map = /** @type {!Object<string, !VendorPropertyMapType>} */ (
      eventType in eventTypeMap ? eventTypeMap : cssPropertyMap
    );
    const el = windowObj['document']['createElement']('div');
    let eventName = '';

    if (map === eventTypeMap) {
      eventName = getJavaScriptEventName(eventType, map, el);
    } else {
      eventName = map[eventType].noPrefix in el.style ? map[eventType].noPrefix : map[eventType].webkitPrefix;
    }

    return eventName;
  }

  /**
   * @param {!Object} windowObj
   * @param {string} eventType
   * @return {string}
   */
  function getCorrectEventName(windowObj, eventType) {
    return getAnimationName(windowObj, eventType);
  }

  /**
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  class MDCSnackbar extends MDCComponent {
    static attachTo(root) {
      return new MDCSnackbar(root);
    }

    show(data) {
      this.foundation_.show(data);
    }

    getDefaultFoundation() {
      const {
        TEXT_SELECTOR,
        ACTION_BUTTON_SELECTOR,
      } = MDCSnackbarFoundation.strings;
      const getText = () => this.root_.querySelector(TEXT_SELECTOR);
      const getActionButton = () => this.root_.querySelector(ACTION_BUTTON_SELECTOR);

      /* eslint brace-style: "off" */
      return new MDCSnackbarFoundation({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        setAriaHidden: () => this.root_.setAttribute('aria-hidden', 'true'),
        unsetAriaHidden: () => this.root_.removeAttribute('aria-hidden'),
        setActionAriaHidden: () => getActionButton().setAttribute('aria-hidden', 'true'),
        unsetActionAriaHidden: () => getActionButton().removeAttribute('aria-hidden'),
        setActionText: (text) => {getActionButton().textContent = text;},
        setMessageText: (text) => {getText().textContent = text;},
        setFocus: () => getActionButton().focus(),
        visibilityIsHidden: () => document.hidden,
        registerCapturedBlurHandler: (handler) => getActionButton().addEventListener('blur', handler, true),
        deregisterCapturedBlurHandler: (handler) => getActionButton().removeEventListener('blur', handler, true),
        registerVisibilityChangeHandler: (handler) => document.addEventListener('visibilitychange', handler),
        deregisterVisibilityChangeHandler: (handler) => document.removeEventListener('visibilitychange', handler),
        registerCapturedInteractionHandler: (evt, handler) =>
          document.body.addEventListener(evt, handler, true),
        deregisterCapturedInteractionHandler: (evt, handler) =>
          document.body.removeEventListener(evt, handler, true),
        registerActionClickHandler: (handler) => getActionButton().addEventListener('click', handler),
        deregisterActionClickHandler: (handler) => getActionButton().removeEventListener('click', handler),
        registerTransitionEndHandler:
          (handler) => this.root_.addEventListener(getCorrectEventName(window, 'transitionend'), handler),
        deregisterTransitionEndHandler:
          (handler) => this.root_.removeEventListener(getCorrectEventName(window, 'transitionend'), handler),
        notifyShow: () => this.emit(MDCSnackbarFoundation.strings.SHOW_EVENT),
        notifyHide: () => this.emit(MDCSnackbarFoundation.strings.HIDE_EVENT),
      });
    }

    get dismissesOnAction() {
      return this.foundation_.dismissesOnAction();
    }

    set dismissesOnAction(dismissesOnAction) {
      this.foundation_.setDismissOnAction(dismissesOnAction);
    }
  }

  /* src/Snackbar.html generated by Svelte v2.9.11 */

  var methods = {
      getStatus() {
          return this.status;
      },
      show(args) {
          return this.mdcComponent.show(args);
      }
  };

  function oncreate() {
      this.mdcComponent = new MDCSnackbar(this.refs.snackbar);
      this.status = "undefined";
      this.mdcComponent.listen("MDCSnackbar:show", () => {
          this.status = "shown";
          this.fire("show");
          this.set({"open": false});
      });
      this.mdcComponent.listen("MDCSnackbar:hide", () => {
          this.status = "hidden";
          this.fire("hide");
          this.set({"open": false});
      });
  }
  function ondestroy() {
      this.mdcComponent.destroy();
  }
  function onstate({current, changed}) {
      if (changed.dismissesOnAction) {
          this.mdcComponent.dismissesOnAction = current.dismissesOnAction;
      }
  }
  function create_main_fragment(component, ctx) {
  	var div, div_class_value;

  	return {
  		c() {
  			div = shared_js.createElement("div");
  			div.innerHTML = "<div class=\"mdc-snackbar__text\"></div>\n  <div class=\"mdc-snackbar__action-wrapper\"><button type=\"button\" class=\"mdc-snackbar__action-button\"></button></div>";
  			div.className = div_class_value = "mdc-snackbar " + (ctx.alignStart?'mdc-snackbar--align-start': '');
  			shared_js.setAttribute(div, "aria-live", "assertive");
  			shared_js.setAttribute(div, "aria-atomic", "true");
  			shared_js.setAttribute(div, "aria-hidden", "true");
  		},

  		m(target, anchor) {
  			shared_js.insert(target, div, anchor);
  			component.refs.snackbar = div;
  		},

  		p(changed, ctx) {
  			if ((changed.alignStart) && div_class_value !== (div_class_value = "mdc-snackbar " + (ctx.alignStart?'mdc-snackbar--align-start': ''))) {
  				div.className = div_class_value;
  			}
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(div);
  			}

  			if (component.refs.snackbar === div) component.refs.snackbar = null;
  		}
  	};
  }

  function Snackbar(options) {
  	shared_js.init(this, options);
  	this.refs = {};
  	this._state = shared_js.assign({}, options.data);
  	this._intro = true;

  	this._handlers.state = [onstate];

  	this._handlers.destroy = [ondestroy];

  	if (!options.root) {
  		this._oncreate = [];
  	}

  	this._fragment = create_main_fragment(this, this._state);

  	this.root._oncreate.push(() => {
  		onstate.call(this, { changed: shared_js.assignTrue({}, this._state), current: this._state });
  		oncreate.call(this);
  		this.fire("update", { changed: shared_js.assignTrue({}, this._state), current: this._state });
  	});

  	if (options.target) {
  		this._fragment.c();
  		this._mount(options.target, options.anchor);

  		shared_js.callAll(this._oncreate);
  	}
  }

  shared_js.assign(Snackbar.prototype, shared_js.proto);
  shared_js.assign(Snackbar.prototype, methods);

  return Snackbar;

})));
