(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('./shared.js')) :
  typeof define === 'function' && define.amd ? define(['./shared.js'], factory) :
  (global.TabBar = factory(global.svelteShared));
}(this, (function (shared_js) { 'use strict';

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template A
   */
  class MDCFoundation {
    /** @return enum{cssClasses} */
    static get cssClasses() {
      // Classes extending MDCFoundation should implement this method to return an object which exports every
      // CSS class the foundation class needs as a property. e.g. {ACTIVE: 'mdc-component--active'}
      return {};
    }

    /** @return enum{strings} */
    static get strings() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // semantic strings as constants. e.g. {ARIA_ROLE: 'tablist'}
      return {};
    }

    /** @return enum{numbers} */
    static get numbers() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // of its semantic numbers as constants. e.g. {ANIMATION_DELAY_MS: 350}
      return {};
    }

    /** @return {!Object} */
    static get defaultAdapter() {
      // Classes extending MDCFoundation may choose to implement this getter in order to provide a convenient
      // way of viewing the necessary methods of an adapter. In the future, this could also be used for adapter
      // validation.
      return {};
    }

    /**
     * @param {A=} adapter
     */
    constructor(adapter = {}) {
      /** @protected {!A} */
      this.adapter_ = adapter;
    }

    init() {
      // Subclasses should override this method to perform initialization routines (registering events, etc.)
    }

    destroy() {
      // Subclasses should override this method to perform de-initialization routines (de-registering events, etc.)
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template F
   */
  class MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCComponent}
     */
    static attachTo(root) {
      // Subclasses which extend MDCBase should provide an attachTo() method that takes a root element and
      // returns an instantiated component with its root set to that element. Also note that in the cases of
      // subclasses, an explicit foundation class will not have to be passed in; it will simply be initialized
      // from getDefaultFoundation().
      return new MDCComponent(root, new MDCFoundation());
    }

    /**
     * @param {!Element} root
     * @param {F=} foundation
     * @param {...?} args
     */
    constructor(root, foundation = undefined, ...args) {
      /** @protected {!Element} */
      this.root_ = root;
      this.initialize(...args);
      // Note that we initialize foundation here and not within the constructor's default param so that
      // this.root_ is defined and can be used within the foundation class.
      /** @protected {!F} */
      this.foundation_ = foundation === undefined ? this.getDefaultFoundation() : foundation;
      this.foundation_.init();
      this.initialSyncWithDOM();
    }

    initialize(/* ...args */) {
      // Subclasses can override this to do any additional setup work that would be considered part of a
      // "constructor". Essentially, it is a hook into the parent constructor before the foundation is
      // initialized. Any additional arguments besides root and foundation will be passed in here.
    }

    /**
     * @return {!F} foundation
     */
    getDefaultFoundation() {
      // Subclasses must override this method to return a properly configured foundation class for the
      // component.
      throw new Error('Subclasses must override getDefaultFoundation to return a properly configured ' +
        'foundation class');
    }

    initialSyncWithDOM() {
      // Subclasses should override this method if they need to perform work to synchronize with a host DOM
      // object. An example of this would be a form control wrapper that needs to synchronize its internal state
      // to some property or attribute of the host DOM. Please note: this is *not* the place to perform DOM
      // reads/writes that would cause layout / paint, as this is called synchronously from within the constructor.
    }

    destroy() {
      // Subclasses may implement this method to release any resources / deregister any listeners they have
      // attached. An example of this might be deregistering a resize event from the window object.
      this.foundation_.destroy();
    }

    /**
     * Wrapper method to add an event listener to the component's root element. This is most useful when
     * listening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    listen(evtType, handler) {
      this.root_.addEventListener(evtType, handler);
    }

    /**
     * Wrapper method to remove an event listener to the component's root element. This is most useful when
     * unlistening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    unlisten(evtType, handler) {
      this.root_.removeEventListener(evtType, handler);
    }

    /**
     * Fires a cross-browser-compatible custom event from the component root of the given type,
     * with the given data.
     * @param {string} evtType
     * @param {!Object} evtData
     * @param {boolean=} shouldBubble
     */
    emit(evtType, evtData, shouldBubble = false) {
      let evt;
      if (typeof CustomEvent === 'function') {
        evt = new CustomEvent(evtType, {
          detail: evtData,
          bubbles: shouldBubble,
        });
      } else {
        evt = document.createEvent('CustomEvent');
        evt.initCustomEvent(evtType, shouldBubble, false, evtData);
      }

      this.root_.dispatchEvent(evt);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  const cssClasses = {
    // Ripple is a special case where the "root" component is really a "mixin" of sorts,
    // given that it's an 'upgrade' to an existing component. That being said it is the root
    // CSS class that all other CSS classes derive from.
    ROOT: 'mdc-ripple-upgraded',
    UNBOUNDED: 'mdc-ripple-upgraded--unbounded',
    BG_FOCUSED: 'mdc-ripple-upgraded--background-focused',
    FG_ACTIVATION: 'mdc-ripple-upgraded--foreground-activation',
    FG_DEACTIVATION: 'mdc-ripple-upgraded--foreground-deactivation',
  };

  const strings = {
    VAR_LEFT: '--mdc-ripple-left',
    VAR_TOP: '--mdc-ripple-top',
    VAR_FG_SIZE: '--mdc-ripple-fg-size',
    VAR_FG_SCALE: '--mdc-ripple-fg-scale',
    VAR_FG_TRANSLATE_START: '--mdc-ripple-fg-translate-start',
    VAR_FG_TRANSLATE_END: '--mdc-ripple-fg-translate-end',
  };

  const numbers = {
    PADDING: 10,
    INITIAL_ORIGIN_SCALE: 0.6,
    DEACTIVATION_TIMEOUT_MS: 225, // Corresponds to $mdc-ripple-translate-duration (i.e. activation animation duration)
    FG_DEACTIVATION_MS: 150, // Corresponds to $mdc-ripple-fade-out-duration (i.e. deactivation animation duration)
    TAP_DELAY_MS: 300, // Delay between touch and simulated mouse events on touch devices
  };

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Stores result from supportsCssVariables to avoid redundant processing to detect CSS custom variable support.
   * @private {boolean|undefined}
   */
  let supportsCssVariables_;

  /**
   * Stores result from applyPassive to avoid redundant processing to detect passive event listener support.
   * @private {boolean|undefined}
   */
  let supportsPassive_;

  /**
   * @param {!Window} windowObj
   * @return {boolean}
   */
  function detectEdgePseudoVarBug(windowObj) {
    // Detect versions of Edge with buggy var() support
    // See: https://developer.microsoft.com/en-us/microsoft-edge/platform/issues/11495448/
    const document = windowObj.document;
    const node = document.createElement('div');
    node.className = 'mdc-ripple-surface--test-edge-var-bug';
    document.body.appendChild(node);

    // The bug exists if ::before style ends up propagating to the parent element.
    // Additionally, getComputedStyle returns null in iframes with display: "none" in Firefox,
    // but Firefox is known to support CSS custom properties correctly.
    // See: https://bugzilla.mozilla.org/show_bug.cgi?id=548397
    const computedStyle = windowObj.getComputedStyle(node);
    const hasPseudoVarBug = computedStyle !== null && computedStyle.borderTopStyle === 'solid';
    node.remove();
    return hasPseudoVarBug;
  }

  /**
   * @param {!Window} windowObj
   * @param {boolean=} forceRefresh
   * @return {boolean|undefined}
   */

  function supportsCssVariables(windowObj, forceRefresh = false) {
    let supportsCssVariables = supportsCssVariables_;
    if (typeof supportsCssVariables_ === 'boolean' && !forceRefresh) {
      return supportsCssVariables;
    }

    const supportsFunctionPresent = windowObj.CSS && typeof windowObj.CSS.supports === 'function';
    if (!supportsFunctionPresent) {
      return;
    }

    const explicitlySupportsCssVars = windowObj.CSS.supports('--css-vars', 'yes');
    // See: https://bugs.webkit.org/show_bug.cgi?id=154669
    // See: README section on Safari
    const weAreFeatureDetectingSafari10plus = (
      windowObj.CSS.supports('(--css-vars: yes)') &&
      windowObj.CSS.supports('color', '#00000000')
    );

    if (explicitlySupportsCssVars || weAreFeatureDetectingSafari10plus) {
      supportsCssVariables = !detectEdgePseudoVarBug(windowObj);
    } else {
      supportsCssVariables = false;
    }

    if (!forceRefresh) {
      supportsCssVariables_ = supportsCssVariables;
    }
    return supportsCssVariables;
  }

  //
  /**
   * Determine whether the current browser supports passive event listeners, and if so, use them.
   * @param {!Window=} globalObj
   * @param {boolean=} forceRefresh
   * @return {boolean|{passive: boolean}}
   */
  function applyPassive(globalObj = window, forceRefresh = false) {
    if (supportsPassive_ === undefined || forceRefresh) {
      let isSupported = false;
      try {
        globalObj.document.addEventListener('test', null, {get passive() {
          isSupported = true;
        }});
      } catch (e) { }

      supportsPassive_ = isSupported;
    }

    return supportsPassive_ ? {passive: true} : false;
  }

  /**
   * @param {!Object} HTMLElementPrototype
   * @return {!Array<string>}
   */
  function getMatchesProperty(HTMLElementPrototype) {
    return [
      'webkitMatchesSelector', 'msMatchesSelector', 'matches',
    ].filter((p) => p in HTMLElementPrototype).pop();
  }

  /**
   * @param {!Event} ev
   * @param {{x: number, y: number}} pageOffset
   * @param {!ClientRect} clientRect
   * @return {{x: number, y: number}}
   */
  function getNormalizedEventCoords(ev, pageOffset, clientRect) {
    const {x, y} = pageOffset;
    const documentX = x + clientRect.left;
    const documentY = y + clientRect.top;

    let normalizedX;
    let normalizedY;
    // Determine touch point relative to the ripple container.
    if (ev.type === 'touchstart') {
      normalizedX = ev.changedTouches[0].pageX - documentX;
      normalizedY = ev.changedTouches[0].pageY - documentY;
    } else {
      normalizedX = ev.pageX - documentX;
      normalizedY = ev.pageY - documentY;
    }

    return {x: normalizedX, y: normalizedY};
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  // Activation events registered on the root element of each instance for activation
  const ACTIVATION_EVENT_TYPES = ['touchstart', 'pointerdown', 'mousedown', 'keydown'];

  // Deactivation events registered on documentElement when a pointer-related down event occurs
  const POINTER_DEACTIVATION_EVENT_TYPES = ['touchend', 'pointerup', 'mouseup'];

  // Tracks activations that have occurred on the current frame, to avoid simultaneous nested activations
  /** @type {!Array<!EventTarget>} */
  let activatedTargets = [];

  /**
   * @extends {MDCFoundation<!MDCRippleAdapter>}
   */
  class MDCRippleFoundation extends MDCFoundation {
    static get cssClasses() {
      return cssClasses;
    }

    static get strings() {
      return strings;
    }

    static get numbers() {
      return numbers;
    }

    static get defaultAdapter() {
      return {
        browserSupportsCssVars: () => /* boolean - cached */ {},
        isUnbounded: () => /* boolean */ {},
        isSurfaceActive: () => /* boolean */ {},
        isSurfaceDisabled: () => /* boolean */ {},
        addClass: (/* className: string */) => {},
        removeClass: (/* className: string */) => {},
        containsEventTarget: (/* target: !EventTarget */) => {},
        registerInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerResizeHandler: (/* handler: EventListener */) => {},
        deregisterResizeHandler: (/* handler: EventListener */) => {},
        updateCssVariable: (/* varName: string, value: string */) => {},
        computeBoundingRect: () => /* ClientRect */ {},
        getWindowPageOffset: () => /* {x: number, y: number} */ {},
      };
    }

    constructor(adapter) {
      super(Object.assign(MDCRippleFoundation.defaultAdapter, adapter));

      /** @private {number} */
      this.layoutFrame_ = 0;

      /** @private {!ClientRect} */
      this.frame_ = /** @type {!ClientRect} */ ({width: 0, height: 0});

      /** @private {!ActivationStateType} */
      this.activationState_ = this.defaultActivationState_();

      /** @private {number} */
      this.initialSize_ = 0;

      /** @private {number} */
      this.maxRadius_ = 0;

      /** @private {function(!Event)} */
      this.activateHandler_ = (e) => this.activate_(e);

      /** @private {function(!Event)} */
      this.deactivateHandler_ = (e) => this.deactivate_(e);

      /** @private {function(?Event=)} */
      this.focusHandler_ = () => this.handleFocus();

      /** @private {function(?Event=)} */
      this.blurHandler_ = () => this.handleBlur();

      /** @private {!Function} */
      this.resizeHandler_ = () => this.layout();

      /** @private {{left: number, top:number}} */
      this.unboundedCoords_ = {
        left: 0,
        top: 0,
      };

      /** @private {number} */
      this.fgScale_ = 0;

      /** @private {number} */
      this.activationTimer_ = 0;

      /** @private {number} */
      this.fgDeactivationRemovalTimer_ = 0;

      /** @private {boolean} */
      this.activationAnimationHasEnded_ = false;

      /** @private {!Function} */
      this.activationTimerCallback_ = () => {
        this.activationAnimationHasEnded_ = true;
        this.runDeactivationUXLogicIfReady_();
      };

      /** @private {?Event} */
      this.previousActivationEvent_ = null;
    }

    /**
     * We compute this property so that we are not querying information about the client
     * until the point in time where the foundation requests it. This prevents scenarios where
     * client-side feature-detection may happen too early, such as when components are rendered on the server
     * and then initialized at mount time on the client.
     * @return {boolean}
     * @private
     */
    isSupported_() {
      return this.adapter_.browserSupportsCssVars();
    }

    /**
     * @return {!ActivationStateType}
     */
    defaultActivationState_() {
      return {
        isActivated: false,
        hasDeactivationUXRun: false,
        wasActivatedByPointer: false,
        wasElementMadeActive: false,
        activationEvent: null,
        isProgrammatic: false,
      };
    }

    /** @override */
    init() {
      if (!this.isSupported_()) {
        return;
      }
      this.registerRootHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.addClass(ROOT);
        if (this.adapter_.isUnbounded()) {
          this.adapter_.addClass(UNBOUNDED);
          // Unbounded ripples need layout logic applied immediately to set coordinates for both shade and ripple
          this.layoutInternal_();
        }
      });
    }

    /** @override */
    destroy() {
      if (!this.isSupported_()) {
        return;
      }

      if (this.activationTimer_) {
        clearTimeout(this.activationTimer_);
        this.activationTimer_ = 0;
        const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
        this.adapter_.removeClass(FG_ACTIVATION);
      }

      this.deregisterRootHandlers_();
      this.deregisterDeactivationHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.removeClass(ROOT);
        this.adapter_.removeClass(UNBOUNDED);
        this.removeCssVars_();
      });
    }

    /** @private */
    registerRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.registerInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.registerInteractionHandler('focus', this.focusHandler_);
      this.adapter_.registerInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.registerResizeHandler(this.resizeHandler_);
      }
    }

    /**
     * @param {!Event} e
     * @private
     */
    registerDeactivationHandlers_(e) {
      if (e.type === 'keydown') {
        this.adapter_.registerInteractionHandler('keyup', this.deactivateHandler_);
      } else {
        POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
          this.adapter_.registerDocumentInteractionHandler(type, this.deactivateHandler_);
        });
      }
    }

    /** @private */
    deregisterRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.deregisterInteractionHandler('focus', this.focusHandler_);
      this.adapter_.deregisterInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.deregisterResizeHandler(this.resizeHandler_);
      }
    }

    /** @private */
    deregisterDeactivationHandlers_() {
      this.adapter_.deregisterInteractionHandler('keyup', this.deactivateHandler_);
      POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterDocumentInteractionHandler(type, this.deactivateHandler_);
      });
    }

    /** @private */
    removeCssVars_() {
      const {strings: strings$$1} = MDCRippleFoundation;
      Object.keys(strings$$1).forEach((k) => {
        if (k.indexOf('VAR_') === 0) {
          this.adapter_.updateCssVariable(strings$$1[k], null);
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    activate_(e) {
      if (this.adapter_.isSurfaceDisabled()) {
        return;
      }

      const activationState = this.activationState_;
      if (activationState.isActivated) {
        return;
      }

      // Avoid reacting to follow-on events fired by touch device after an already-processed user interaction
      const previousActivationEvent = this.previousActivationEvent_;
      const isSameInteraction = previousActivationEvent && e && previousActivationEvent.type !== e.type;
      if (isSameInteraction) {
        return;
      }

      activationState.isActivated = true;
      activationState.isProgrammatic = e === null;
      activationState.activationEvent = e;
      activationState.wasActivatedByPointer = activationState.isProgrammatic ? false : (
        e.type === 'mousedown' || e.type === 'touchstart' || e.type === 'pointerdown'
      );

      const hasActivatedChild =
        e && activatedTargets.length > 0 && activatedTargets.some((target) => this.adapter_.containsEventTarget(target));
      if (hasActivatedChild) {
        // Immediately reset activation state, while preserving logic that prevents touch follow-on events
        this.resetActivationState_();
        return;
      }

      if (e) {
        activatedTargets.push(/** @type {!EventTarget} */ (e.target));
        this.registerDeactivationHandlers_(e);
      }

      activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
      if (activationState.wasElementMadeActive) {
        this.animateActivation_();
      }

      requestAnimationFrame(() => {
        // Reset array on next frame after the current event has had a chance to bubble to prevent ancestor ripples
        activatedTargets = [];

        if (!activationState.wasElementMadeActive && (e.key === ' ' || e.keyCode === 32)) {
          // If space was pressed, try again within an rAF call to detect :active, because different UAs report
          // active states inconsistently when they're called within event handling code:
          // - https://bugs.chromium.org/p/chromium/issues/detail?id=635971
          // - https://bugzilla.mozilla.org/show_bug.cgi?id=1293741
          // We try first outside rAF to support Edge, which does not exhibit this problem, but will crash if a CSS
          // variable is set within a rAF callback for a submit button interaction (#2241).
          activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
          if (activationState.wasElementMadeActive) {
            this.animateActivation_();
          }
        }

        if (!activationState.wasElementMadeActive) {
          // Reset activation state immediately if element was not made active.
          this.activationState_ = this.defaultActivationState_();
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    checkElementMadeActive_(e) {
      return (e && e.type === 'keydown') ? this.adapter_.isSurfaceActive() : true;
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    activate(event = null) {
      this.activate_(event);
    }

    /** @private */
    animateActivation_() {
      const {VAR_FG_TRANSLATE_START, VAR_FG_TRANSLATE_END} = MDCRippleFoundation.strings;
      const {FG_DEACTIVATION, FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      const {DEACTIVATION_TIMEOUT_MS} = MDCRippleFoundation.numbers;

      this.layoutInternal_();

      let translateStart = '';
      let translateEnd = '';

      if (!this.adapter_.isUnbounded()) {
        const {startPoint, endPoint} = this.getFgTranslationCoordinates_();
        translateStart = `${startPoint.x}px, ${startPoint.y}px`;
        translateEnd = `${endPoint.x}px, ${endPoint.y}px`;
      }

      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_START, translateStart);
      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_END, translateEnd);
      // Cancel any ongoing activation/deactivation animations
      clearTimeout(this.activationTimer_);
      clearTimeout(this.fgDeactivationRemovalTimer_);
      this.rmBoundedActivationClasses_();
      this.adapter_.removeClass(FG_DEACTIVATION);

      // Force layout in order to re-trigger the animation.
      this.adapter_.computeBoundingRect();
      this.adapter_.addClass(FG_ACTIVATION);
      this.activationTimer_ = setTimeout(() => this.activationTimerCallback_(), DEACTIVATION_TIMEOUT_MS);
    }

    /**
     * @private
     * @return {{startPoint: PointType, endPoint: PointType}}
     */
    getFgTranslationCoordinates_() {
      const {activationEvent, wasActivatedByPointer} = this.activationState_;

      let startPoint;
      if (wasActivatedByPointer) {
        startPoint = getNormalizedEventCoords(
          /** @type {!Event} */ (activationEvent),
          this.adapter_.getWindowPageOffset(), this.adapter_.computeBoundingRect()
        );
      } else {
        startPoint = {
          x: this.frame_.width / 2,
          y: this.frame_.height / 2,
        };
      }
      // Center the element around the start point.
      startPoint = {
        x: startPoint.x - (this.initialSize_ / 2),
        y: startPoint.y - (this.initialSize_ / 2),
      };

      const endPoint = {
        x: (this.frame_.width / 2) - (this.initialSize_ / 2),
        y: (this.frame_.height / 2) - (this.initialSize_ / 2),
      };

      return {startPoint, endPoint};
    }

    /** @private */
    runDeactivationUXLogicIfReady_() {
      // This method is called both when a pointing device is released, and when the activation animation ends.
      // The deactivation animation should only run after both of those occur.
      const {FG_DEACTIVATION} = MDCRippleFoundation.cssClasses;
      const {hasDeactivationUXRun, isActivated} = this.activationState_;
      const activationHasEnded = hasDeactivationUXRun || !isActivated;

      if (activationHasEnded && this.activationAnimationHasEnded_) {
        this.rmBoundedActivationClasses_();
        this.adapter_.addClass(FG_DEACTIVATION);
        this.fgDeactivationRemovalTimer_ = setTimeout(() => {
          this.adapter_.removeClass(FG_DEACTIVATION);
        }, numbers.FG_DEACTIVATION_MS);
      }
    }

    /** @private */
    rmBoundedActivationClasses_() {
      const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      this.adapter_.removeClass(FG_ACTIVATION);
      this.activationAnimationHasEnded_ = false;
      this.adapter_.computeBoundingRect();
    }

    resetActivationState_() {
      this.previousActivationEvent_ = this.activationState_.activationEvent;
      this.activationState_ = this.defaultActivationState_();
      // Touch devices may fire additional events for the same interaction within a short time.
      // Store the previous event until it's safe to assume that subsequent events are for new interactions.
      setTimeout(() => this.previousActivationEvent_ = null, MDCRippleFoundation.numbers.TAP_DELAY_MS);
    }

    /**
     * @param {?Event} e
     * @private
     */
    deactivate_(e) {
      const activationState = this.activationState_;
      // This can happen in scenarios such as when you have a keyup event that blurs the element.
      if (!activationState.isActivated) {
        return;
      }

      const state = /** @type {!ActivationStateType} */ (Object.assign({}, activationState));

      if (activationState.isProgrammatic) {
        const evtObject = null;
        requestAnimationFrame(() => this.animateDeactivation_(evtObject, state));
        this.resetActivationState_();
      } else {
        this.deregisterDeactivationHandlers_();
        requestAnimationFrame(() => {
          this.activationState_.hasDeactivationUXRun = true;
          this.animateDeactivation_(e, state);
          this.resetActivationState_();
        });
      }
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    deactivate(event = null) {
      this.deactivate_(event);
    }

    /**
     * @param {Event} e
     * @param {!ActivationStateType} options
     * @private
     */
    animateDeactivation_(e, {wasActivatedByPointer, wasElementMadeActive}) {
      if (wasActivatedByPointer || wasElementMadeActive) {
        this.runDeactivationUXLogicIfReady_();
      }
    }

    layout() {
      if (this.layoutFrame_) {
        cancelAnimationFrame(this.layoutFrame_);
      }
      this.layoutFrame_ = requestAnimationFrame(() => {
        this.layoutInternal_();
        this.layoutFrame_ = 0;
      });
    }

    /** @private */
    layoutInternal_() {
      this.frame_ = this.adapter_.computeBoundingRect();
      const maxDim = Math.max(this.frame_.height, this.frame_.width);

      // Surface diameter is treated differently for unbounded vs. bounded ripples.
      // Unbounded ripple diameter is calculated smaller since the surface is expected to already be padded appropriately
      // to extend the hitbox, and the ripple is expected to meet the edges of the padded hitbox (which is typically
      // square). Bounded ripples, on the other hand, are fully expected to expand beyond the surface's longest diameter
      // (calculated based on the diagonal plus a constant padding), and are clipped at the surface's border via
      // `overflow: hidden`.
      const getBoundedRadius = () => {
        const hypotenuse = Math.sqrt(Math.pow(this.frame_.width, 2) + Math.pow(this.frame_.height, 2));
        return hypotenuse + MDCRippleFoundation.numbers.PADDING;
      };

      this.maxRadius_ = this.adapter_.isUnbounded() ? maxDim : getBoundedRadius();

      // Ripple is sized as a fraction of the largest dimension of the surface, then scales up using a CSS scale transform
      this.initialSize_ = maxDim * MDCRippleFoundation.numbers.INITIAL_ORIGIN_SCALE;
      this.fgScale_ = this.maxRadius_ / this.initialSize_;

      this.updateLayoutCssVars_();
    }

    /** @private */
    updateLayoutCssVars_() {
      const {
        VAR_FG_SIZE, VAR_LEFT, VAR_TOP, VAR_FG_SCALE,
      } = MDCRippleFoundation.strings;

      this.adapter_.updateCssVariable(VAR_FG_SIZE, `${this.initialSize_}px`);
      this.adapter_.updateCssVariable(VAR_FG_SCALE, this.fgScale_);

      if (this.adapter_.isUnbounded()) {
        this.unboundedCoords_ = {
          left: Math.round((this.frame_.width / 2) - (this.initialSize_ / 2)),
          top: Math.round((this.frame_.height / 2) - (this.initialSize_ / 2)),
        };

        this.adapter_.updateCssVariable(VAR_LEFT, `${this.unboundedCoords_.left}px`);
        this.adapter_.updateCssVariable(VAR_TOP, `${this.unboundedCoords_.top}px`);
      }
    }

    /** @param {boolean} unbounded */
    setUnbounded(unbounded) {
      const {UNBOUNDED} = MDCRippleFoundation.cssClasses;
      if (unbounded) {
        this.adapter_.addClass(UNBOUNDED);
      } else {
        this.adapter_.removeClass(UNBOUNDED);
      }
    }

    handleFocus() {
      requestAnimationFrame(() =>
        this.adapter_.addClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }

    handleBlur() {
      requestAnimationFrame(() =>
        this.adapter_.removeClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends MDCComponent<!MDCRippleFoundation>
   */
  class MDCRipple extends MDCComponent {
    /** @param {...?} args */
    constructor(...args) {
      super(...args);

      /** @type {boolean} */
      this.disabled = false;

      /** @private {boolean} */
      this.unbounded_;
    }

    /**
     * @param {!Element} root
     * @param {{isUnbounded: (boolean|undefined)}=} options
     * @return {!MDCRipple}
     */
    static attachTo(root, {isUnbounded = undefined} = {}) {
      const ripple = new MDCRipple(root);
      // Only override unbounded behavior if option is explicitly specified
      if (isUnbounded !== undefined) {
        ripple.unbounded = /** @type {boolean} */ (isUnbounded);
      }
      return ripple;
    }

    /**
     * @param {!RippleCapableSurface} instance
     * @return {!MDCRippleAdapter}
     */
    static createAdapter(instance) {
      const MATCHES = getMatchesProperty(HTMLElement.prototype);

      return {
        browserSupportsCssVars: () => supportsCssVariables(window),
        isUnbounded: () => instance.unbounded,
        isSurfaceActive: () => instance.root_[MATCHES](':active'),
        isSurfaceDisabled: () => instance.disabled,
        addClass: (className) => instance.root_.classList.add(className),
        removeClass: (className) => instance.root_.classList.remove(className),
        containsEventTarget: (target) => instance.root_.contains(target),
        registerInteractionHandler: (evtType, handler) =>
          instance.root_.addEventListener(evtType, handler, applyPassive()),
        deregisterInteractionHandler: (evtType, handler) =>
          instance.root_.removeEventListener(evtType, handler, applyPassive()),
        registerDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.addEventListener(evtType, handler, applyPassive()),
        deregisterDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.removeEventListener(evtType, handler, applyPassive()),
        registerResizeHandler: (handler) => window.addEventListener('resize', handler),
        deregisterResizeHandler: (handler) => window.removeEventListener('resize', handler),
        updateCssVariable: (varName, value) => instance.root_.style.setProperty(varName, value),
        computeBoundingRect: () => instance.root_.getBoundingClientRect(),
        getWindowPageOffset: () => ({x: window.pageXOffset, y: window.pageYOffset}),
      };
    }

    /** @return {boolean} */
    get unbounded() {
      return this.unbounded_;
    }

    /** @param {boolean} unbounded */
    set unbounded(unbounded) {
      this.unbounded_ = Boolean(unbounded);
      this.setUnbounded_();
    }

    /**
     * Closure Compiler throws an access control error when directly accessing a
     * protected or private property inside a getter/setter, like unbounded above.
     * By accessing the protected property inside a method, we solve that problem.
     * That's why this function exists.
     * @private
     */
    setUnbounded_() {
      this.foundation_.setUnbounded(this.unbounded_);
    }

    activate() {
      this.foundation_.activate();
    }

    deactivate() {
      this.foundation_.deactivate();
    }

    layout() {
      this.foundation_.layout();
    }

    /**
     * @return {!MDCRippleFoundation}
     * @override
     */
    getDefaultFoundation() {
      return new MDCRippleFoundation(MDCRipple.createAdapter(this));
    }

    /** @override */
    initialSyncWithDOM() {
      this.unbounded = 'mdcRippleIsUnbounded' in this.root_.dataset;
    }
  }

  /**
   * See Material Design spec for more details on when to use ripples.
   * https://material.io/guidelines/motion/choreography.html#choreography-creation
   * @record
   */
  class RippleCapableSurface {}

  /** @protected {!Element} */
  RippleCapableSurface.prototype.root_;

  /**
   * Whether or not the ripple bleeds out of the bounds of the element.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.unbounded;

  /**
   * Whether or not the ripple is attached to a disabled component.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.disabled;

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /** @enum {string} */
  const cssClasses$1 = {
    ACTIVE: 'mdc-tab-indicator--active',
    FADE: 'mdc-tab-indicator--fade',
    FADING_ACTIVATE: 'mdc-tab-indicator--fading-activate',
    FADING_DEACTIVATE: 'mdc-tab-indicator--fading-deactivate',
    SLIDING_ACTIVATE: 'mdc-tab-indicator--sliding-activate',
  };

  /** @enum {string} */
  const strings$1 = {
    CONTENT_SELECTOR: '.mdc-tab-indicator__content',
  };

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCFoundation<!MDCTabIndicatorAdapter>}
   * @abstract
   */
  class MDCTabIndicatorFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$1;
    }

    /** @return enum {string} */
    static get strings() {
      return strings$1;
    }

    /**
     * @see MDCTabIndicatorAdapter for typing information
     * @return {!MDCTabIndicatorAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTabIndicatorAdapter} */ ({
        registerEventHandler: () => {},
        deregisterEventHandler: () => {},
        addClass: () => {},
        removeClass: () => {},
        computeContentClientRect: () => {},
        setContentStyleProperty: () => {},
      });
    }

    /** @param {!MDCTabIndicatorAdapter} adapter */
    constructor(adapter) {
      super(Object.assign(MDCTabIndicatorFoundation.defaultAdapter, adapter));
    }

    /** @return {!ClientRect} */
    computeContentClientRect() {
      return this.adapter_.computeContentClientRect();
    }

    /**
     * Activates the indicator
     * @param {!ClientRect=} previousIndicatorClientRect
     * @abstract
     */
    activate(previousIndicatorClientRect) {} // eslint-disable-line no-unused-vars

    /** @abstract */
    deactivate() {}
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCTabIndicatorFoundation}
   * @final
   */
  class MDCSlidingTabIndicatorFoundation extends MDCTabIndicatorFoundation {
    /** @param {...?} args */
    constructor(...args) {
      super(...args);

      /** @private {function(?Event): undefined} */
      this.handleTransitionEnd_ = () => this.handleTransitionEnd();
    }

    /** Handles the transitionend event */
    handleTransitionEnd() {
      this.adapter_.deregisterEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.SLIDING_ACTIVATE);
    }

    /** @param {!ClientRect=} previousIndicatorClientRect */
    activate(previousIndicatorClientRect) {
      this.adapter_.addClass(MDCTabIndicatorFoundation.cssClasses.ACTIVE);

      // Early exit if no indicator is present to handle cases where an indicator
      // may be activated without a prior indicator state
      if (!previousIndicatorClientRect) {
        return;
      }

      // This animation uses the FLIP approach. You can read more about it at the link below:
      // https://aerotwist.com/blog/flip-your-animations/

      // Calculate the dimensions based on the dimensions of the previous indicator
      const currentClientRect = this.computeContentClientRect();
      const widthDelta = previousIndicatorClientRect.width / currentClientRect.width;
      const xPosition = previousIndicatorClientRect.left - currentClientRect.left;
      this.adapter_.setContentStyleProperty('transform', `translateX(${xPosition}px) scaleX(${widthDelta})`);

      // Force repaint
      this.computeContentClientRect();

      // Add animating class and remove transformation in a new frame
      requestAnimationFrame(() => {
        this.adapter_.addClass(MDCTabIndicatorFoundation.cssClasses.SLIDING_ACTIVATE);
        this.adapter_.setContentStyleProperty('transform', '');
      });

      this.adapter_.registerEventHandler('transitionend', this.handleTransitionEnd_);
    }

    deactivate() {
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.ACTIVE);
      // We remove the animating class in deactivate in case the Tab is deactivated before the animation completes and
      // the "transitionend" handler isn't called.
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.SLIDING_ACTIVATE);
      this.adapter_.deregisterEventHandler('transitionend', this.handleTransitionEnd_);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCTabIndicatorFoundation}
   * @final
   */
  class MDCFadingTabIndicatorFoundation extends MDCTabIndicatorFoundation {
    /** @param {...?} args */
    constructor(...args) {
      super(...args);

      /** @private {function(?Event): undefined} */
      this.handleTransitionEnd_ = () => this.handleTransitionEnd();
    }

    /** Handles the transitionend event */
    handleTransitionEnd() {
      this.adapter_.deregisterEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.FADING_ACTIVATE);
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.FADING_DEACTIVATE);
    }

    activate() {
      this.adapter_.registerEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.addClass(MDCTabIndicatorFoundation.cssClasses.FADING_ACTIVATE);
      this.adapter_.addClass(MDCTabIndicatorFoundation.cssClasses.ACTIVE);
    }

    deactivate() {
      this.adapter_.registerEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.addClass(MDCTabIndicatorFoundation.cssClasses.FADING_DEACTIVATE);
      this.adapter_.removeClass(MDCTabIndicatorFoundation.cssClasses.ACTIVE);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCComponent<!MDCTabIndicatorFoundation>}
   * @final
   */
  class MDCTabIndicator extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCTabIndicator}
     */
    static attachTo(root) {
      return new MDCTabIndicator(root);
    }

    /**
     * @param {...?} args
     */
    constructor(...args) {
      super(...args);
      /** @type {?Element} */
      this.content_;
    }

    initialize() {
      this.content_ = this.root_.querySelector(MDCTabIndicatorFoundation.strings.CONTENT_SELECTOR);
    }

    /**
     * @return {!ClientRect}
     */
    computeContentClientRect() {
      return this.foundation_.computeContentClientRect();
    }

    /**
     * @return {!MDCTabIndicatorFoundation}
     */
    getDefaultFoundation() {
      const adapter = /** @type {!MDCTabIndicatorAdapter} */ (Object.assign({
        registerEventHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterEventHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        computeContentClientRect: () => this.content_.getBoundingClientRect(),
        setContentStyleProperty: (prop, value) => this.content_.style.setProperty(prop, value),
      }));

      if (this.root_.classList.contains(MDCTabIndicatorFoundation.cssClasses.FADE)) {
        return new MDCFadingTabIndicatorFoundation(adapter);
      }

      // Default to the sliding indicator
      return new MDCSlidingTabIndicatorFoundation(adapter);
    }

    /**
     * @param {!ClientRect=} previousIndicatorClientRect
     */
    activate(previousIndicatorClientRect) {
      this.foundation_.activate(previousIndicatorClientRect);
    }

    deactivate() {
      this.foundation_.deactivate();
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const cssClasses$2 = {
    ACTIVE: 'mdc-tab--active',
    ANIMATING_ACTIVATE: 'mdc-tab--animating-activate',
    ANIMATING_DEACTIVATE: 'mdc-tab--animating-deactivate',
  };

  /** @enum {string} */
  const strings$2 = {
    ARIA_SELECTED: 'aria-selected',
    RIPPLE_SELECTOR: '.mdc-tab__ripple',
    CONTENT_SELECTOR: '.mdc-tab__content',
    TAB_INDICATOR_SELECTOR: '.mdc-tab-indicator',
    TABINDEX: 'tabIndex',
    INTERACTED_EVENT: 'MDCTab:interacted',
  };

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCFoundation<!MDCTabAdapter>}
   * @final
   */
  class MDCTabFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$2;
    }

    /** @return enum {string} */
    static get strings() {
      return strings$2;
    }

    /**
     * @see MDCTabAdapter for typing information
     * @return {!MDCTabAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTabAdapter} */ ({
        registerEventHandler: () => {},
        deregisterEventHandler: () => {},
        addClass: () => {},
        removeClass: () => {},
        hasClass: () => {},
        setAttr: () => {},
        activateIndicator: () => {},
        deactivateIndicator: () => {},
        computeIndicatorClientRect: () => {},
        notifyInteracted: () => {},
        getOffsetLeft: () => {},
        getOffsetWidth: () => {},
        getContentOffsetLeft: () => {},
        getContentOffsetWidth: () => {},
        focus: () => {},
      });
    }

    /** @param {!MDCTabAdapter} adapter */
    constructor(adapter) {
      super(Object.assign(MDCTabFoundation.defaultAdapter, adapter));

      /** @private {function(!Event): undefined} */
      this.handleTransitionEnd_ = (evt) => this.handleTransitionEnd(evt);

      /** @private {function(?Event): undefined} */
      this.handleClick_ = () => this.handleClick();
    }

    init() {
      this.adapter_.registerEventHandler('click', this.handleClick_);
    }

    /**
     * Handles the "transitionend" event
     * @param {!Event} evt A browser event
     */
    handleTransitionEnd(evt) {
      // Early exit for ripple
      if (evt.pseudoElement) {
        return;
      }
      this.adapter_.deregisterEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.removeClass(cssClasses$2.ANIMATING_ACTIVATE);
      this.adapter_.removeClass(cssClasses$2.ANIMATING_DEACTIVATE);
    }

    /**
     * Handles the "click" event
     */
    handleClick() {
      // It's up to the parent component to keep track of the active Tab and
      // ensure we don't activate a Tab that's already active.
      this.adapter_.notifyInteracted();
    }

    /**
     * Returns the Tab's active state
     * @return {boolean}
     */
    isActive() {
      return this.adapter_.hasClass(cssClasses$2.ACTIVE);
    }

    /**
     * Activates the Tab
     * @param {!ClientRect=} previousIndicatorClientRect
     */
    activate(previousIndicatorClientRect) {
      // Early exit
      if (this.isActive()) {
        return;
      }

      this.adapter_.registerEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.addClass(cssClasses$2.ANIMATING_ACTIVATE);
      this.adapter_.addClass(cssClasses$2.ACTIVE);
      this.adapter_.setAttr(strings$2.ARIA_SELECTED, 'true');
      this.adapter_.setAttr(strings$2.TABINDEX, '0');
      this.adapter_.activateIndicator(previousIndicatorClientRect);
      this.adapter_.focus();
    }

    /**
     * Deactivates the Tab
     */
    deactivate() {
      // Early exit
      if (!this.isActive()) {
        return;
      }

      this.adapter_.registerEventHandler('transitionend', this.handleTransitionEnd_);
      this.adapter_.addClass(cssClasses$2.ANIMATING_DEACTIVATE);
      this.adapter_.removeClass(cssClasses$2.ACTIVE);
      this.adapter_.setAttr(strings$2.ARIA_SELECTED, 'false');
      this.adapter_.setAttr(strings$2.TABINDEX, '-1');
      this.adapter_.deactivateIndicator();
    }

    /**
     * Returns the indicator's client rect
     * @return {!ClientRect}
     */
    computeIndicatorClientRect() {
      return this.adapter_.computeIndicatorClientRect();
    }

    /**
     * Returns the dimensions of the Tab
     * @return {!MDCTabDimensions}
     */
    computeDimensions() {
      const rootWidth = this.adapter_.getOffsetWidth();
      const rootLeft = this.adapter_.getOffsetLeft();
      const contentWidth = this.adapter_.getContentOffsetWidth();
      const contentLeft = this.adapter_.getContentOffsetLeft();

      return {
        rootLeft,
        rootRight: rootLeft + rootWidth,
        contentLeft: rootLeft + contentLeft,
        contentRight: rootLeft + contentLeft + contentWidth,
      };
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCTabFoundation>}
   * @final
   */
  class MDCTab extends MDCComponent {
    /**
     * @param {...?} args
     */
    constructor(...args) {
      super(...args);
      /** @private {?MDCRipple} */
      this.ripple_;
      /** @private {?MDCTabIndicator} */
      this.tabIndicator_;
      /** @private {?Element} */
      this.content_;
    }

    /**
     * @param {!Element} root
     * @return {!MDCTab}
     */
    static attachTo(root) {
      return new MDCTab(root);
    }

    initialize(
      rippleFactory = (el, foundation) => new MDCRipple(el, foundation),
      tabIndicatorFactory = (el) => new MDCTabIndicator(el)) {
      const rippleSurface = this.root_.querySelector(MDCTabFoundation.strings.RIPPLE_SELECTOR);
      const rippleAdapter = Object.assign(MDCRipple.createAdapter(/** @type {!RippleCapableSurface} */ (this)), {
        addClass: (className) => rippleSurface.classList.add(className),
        removeClass: (className) => rippleSurface.classList.remove(className),
        updateCssVariable: (varName, value) => rippleSurface.style.setProperty(varName, value),
      });
      const rippleFoundation = new MDCRippleFoundation(rippleAdapter);
      this.ripple_ = rippleFactory(this.root_, rippleFoundation);

      const tabIndicatorElement = this.root_.querySelector(MDCTabFoundation.strings.TAB_INDICATOR_SELECTOR);
      this.tabIndicator_ = tabIndicatorFactory(tabIndicatorElement);

      this.content_ = this.root_.querySelector(MDCTabFoundation.strings.CONTENT_SELECTOR);
    }

    destroy() {
      this.ripple_.destroy();
      super.destroy();
    }

    /**
     * @return {!MDCTabFoundation}
     */
    getDefaultFoundation() {
      return new MDCTabFoundation(
        /** @type {!MDCTabAdapter} */ ({
          setAttr: (attr, value) => this.root_.setAttribute(attr, value),
          registerEventHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
          deregisterEventHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
          addClass: (className) => this.root_.classList.add(className),
          removeClass: (className) => this.root_.classList.remove(className),
          hasClass: (className) => this.root_.classList.contains(className),
          activateIndicator: (previousIndicatorClientRect) => this.tabIndicator_.activate(previousIndicatorClientRect),
          deactivateIndicator: () => this.tabIndicator_.deactivate(),
          computeIndicatorClientRect: () => this.tabIndicator_.computeContentClientRect(),
          notifyInteracted: () => this.emit(MDCTabFoundation.strings.INTERACTED_EVENT, {tab: this}, true /* bubble */),
          getOffsetLeft: () => this.root_.offsetLeft,
          getOffsetWidth: () => this.root_.offsetWidth,
          getContentOffsetLeft: () => this.content_.offsetLeft,
          getContentOffsetWidth: () => this.content_.offsetWidth,
          focus: () => this.root_.focus(),
        }));
    }

    /**
     * Getter for the active state of the tab
     * @return {boolean}
     */
    get active() {
      return this.foundation_.isActive();
    }

    /**
     * Activates the tab
     * @param {!ClientRect=} computeIndicatorClientRect
     */
    activate(computeIndicatorClientRect) {
      this.foundation_.activate(computeIndicatorClientRect);
    }

    /**
     * Deactivates the tab
     */
    deactivate() {
      this.foundation_.deactivate();
    }

    /**
     * Returns the indicator's client rect
     * @return {!ClientRect}
     */
    computeIndicatorClientRect() {
      return this.foundation_.computeIndicatorClientRect();
    }

    /**
     * @return {!MDCTabDimensions}
     */
    computeDimensions() {
      return this.foundation_.computeDimensions();
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /** @enum {string} */
  const cssClasses$3 = {
    ANIMATING: 'mdc-tab-scroller--animating',
    SCROLL_TEST: 'mdc-tab-scroller__test',
    SCROLL_AREA_SCROLL: 'mdc-tab-scroller__scroll-area--scroll',
  };

  /** @enum {string} */
  const strings$3 = {
    AREA_SELECTOR: '.mdc-tab-scroller__scroll-area',
    CONTENT_SELECTOR: '.mdc-tab-scroller__scroll-content',
  };

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */
  /* eslint-enable no-unused-vars */

  /**
   * @abstract
   */
  class MDCTabScrollerRTL {
    /** @param {!MDCTabScrollerAdapter} adapter */
    constructor(adapter) {
      /** @private */
      this.adapter_ = adapter;
    }

    /**
     * @param {number} translateX The current translateX position
     * @return {number}
     * @abstract
     */
    getScrollPositionRTL(translateX) {}

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     * @abstract
     */
    scrollToRTL(scrollX) {}

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     * @abstract
     */
    incrementScrollRTL(scrollX) {}

    /**
     * @param {number} scrollX The current scrollX position
     * @param {number} translateX The current translateX position
     * @return {number}
     * @abstract
     */
    getAnimatingScrollPosition(scrollX, translateX) {}
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */
  /* eslint-enable no-unused-vars */

  /**
   * @extends {MDCTabScrollerRTL}
   * @final
   */
  class MDCTabScrollerRTLDefault extends MDCTabScrollerRTL {
    /**
     * @return {number}
     */
    getScrollPositionRTL() {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const {right} = this.calculateScrollEdges_();
      // Scroll values on most browsers are ints instead of floats so we round
      return Math.round(right - currentScrollLeft);
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    scrollToRTL(scrollX) {
      const edges = this.calculateScrollEdges_();
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(edges.right - scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: clampedScrollLeft - currentScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    incrementScrollRTL(scrollX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(currentScrollLeft - scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: clampedScrollLeft - currentScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @return {number}
     */
    getAnimatingScrollPosition(scrollX) {
      return scrollX;
    }

    /**
     * @return {!MDCTabScrollerHorizontalEdges}
     * @private
     */
    calculateScrollEdges_() {
      const contentWidth = this.adapter_.getScrollContentOffsetWidth();
      const rootWidth = this.adapter_.getScrollAreaOffsetWidth();
      return /** @type {!MDCTabScrollerHorizontalEdges} */ ({
        left: 0,
        right: contentWidth - rootWidth,
      });
    }

    /**
     * @param {number} scrollX
     * @return {number}
     * @private
     */
    clampScrollValue_(scrollX) {
      const edges = this.calculateScrollEdges_();
      return Math.min(Math.max(edges.left, scrollX), edges.right);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */
  /* eslint-enable no-unused-vars */

  /**
   * @extends {MDCTabScrollerRTL}
   * @final
   */
  class MDCTabScrollerRTLNegative extends MDCTabScrollerRTL {
    /**
     * @param {number} translateX The current translateX position
     * @return {number}
     */
    getScrollPositionRTL(translateX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      return Math.round(translateX - currentScrollLeft);
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    scrollToRTL(scrollX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(-scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: clampedScrollLeft - currentScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    incrementScrollRTL(scrollX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(currentScrollLeft - scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: clampedScrollLeft - currentScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @param {number} translateX
     * @return {number}
     */
    getAnimatingScrollPosition(scrollX, translateX) {
      return scrollX - translateX;
    }

    /**
     * @return {!MDCTabScrollerHorizontalEdges}
     * @private
     */
    calculateScrollEdges_() {
      const contentWidth = this.adapter_.getScrollContentOffsetWidth();
      const rootWidth = this.adapter_.getScrollAreaOffsetWidth();
      return /** @type {!MDCTabScrollerHorizontalEdges} */ ({
        left: rootWidth - contentWidth,
        right: 0,
      });
    }

    /**
     * @param {number} scrollX
     * @return {number}
     * @private
     */
    clampScrollValue_(scrollX) {
      const edges = this.calculateScrollEdges_();
      return Math.max(Math.min(edges.right, scrollX), edges.left);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */
  /* eslint-enable no-unused-vars */

  /**
   * @extends {MDCTabScrollerRTL}
   * @final
   */
  class MDCTabScrollerRTLReverse extends MDCTabScrollerRTL {
    /**
     * @param {number} translateX
     * @return {number}
     */
    getScrollPositionRTL(translateX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      // Scroll values on most browsers are ints instead of floats so we round
      return Math.round(currentScrollLeft - translateX);
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    scrollToRTL(scrollX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: currentScrollLeft - clampedScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @return {!MDCTabScrollerAnimation}
     */
    incrementScrollRTL(scrollX) {
      const currentScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      const clampedScrollLeft = this.clampScrollValue_(currentScrollLeft + scrollX);
      return /** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: clampedScrollLeft,
        scrollDelta: currentScrollLeft - clampedScrollLeft,
      });
    }

    /**
     * @param {number} scrollX
     * @return {number}
     */
    getAnimatingScrollPosition(scrollX, translateX) {
      return scrollX + translateX;
    }

    /**
     * @return {!MDCTabScrollerHorizontalEdges}
     * @private
     */
    calculateScrollEdges_() {
      const contentWidth = this.adapter_.getScrollContentOffsetWidth();
      const rootWidth = this.adapter_.getScrollAreaOffsetWidth();
      return /** @type {!MDCTabScrollerHorizontalEdges} */ ({
        left: contentWidth - rootWidth,
        right: 0,
      });
    }

    /**
     * @param {number} scrollX
     * @return {number}
     * @private
     */
    clampScrollValue_(scrollX) {
      const edges = this.calculateScrollEdges_();
      return Math.min(Math.max(edges.right, scrollX), edges.left);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCFoundation<!MDCTabScrollerAdapter>}
   * @final
   */
  class MDCTabScrollerFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$3;
    }

    /** @return enum {string} */
    static get strings() {
      return strings$3;
    }

    /**
     * @see MDCTabScrollerAdapter for typing information
     * @return {!MDCTabScrollerAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTabScrollerAdapter} */ ({
        eventTargetMatchesSelector: () => {},
        addClass: () => {},
        removeClass: () => {},
        addScrollAreaClass: () => {},
        setScrollAreaStyleProperty: () => {},
        setScrollContentStyleProperty: () => {},
        getScrollContentStyleValue: () => {},
        setScrollAreaScrollLeft: () => {},
        getScrollAreaScrollLeft: () => {},
        getScrollContentOffsetWidth: () => {},
        getScrollAreaOffsetWidth: () => {},
        computeScrollAreaClientRect: () => {},
        computeScrollContentClientRect: () => {},
        computeHorizontalScrollbarHeight: () => {},
      });
    }

    /** @param {!MDCTabScrollerAdapter} adapter */
    constructor(adapter) {
      super(Object.assign(MDCTabScrollerFoundation.defaultAdapter, adapter));

      /**
       * This boolean controls whether we should handle the transitionend and interaction events during the animation.
       * @private {boolean}
       */
      this.isAnimating_ = false;

      /**
       * The MDCTabScrollerRTL instance varies per browser and allows us to encapsulate the peculiar browser behavior
       * of RTL scrolling in it's own class.
       * @private {?MDCTabScrollerRTL}
       */
      this.rtlScrollerInstance_;
    }

    init() {
      // Compute horizontal scrollbar height on scroller with overflow initially hidden, then update overflow to scroll
      // and immediately adjust bottom margin to avoid the scrollbar initially appearing before JS runs.
      const horizontalScrollbarHeight = this.adapter_.computeHorizontalScrollbarHeight();
      this.adapter_.setScrollAreaStyleProperty('margin-bottom', -horizontalScrollbarHeight + 'px');
      this.adapter_.addScrollAreaClass(MDCTabScrollerFoundation.cssClasses.SCROLL_AREA_SCROLL);
    }

    /**
     * Computes the current visual scroll position
     * @return {number}
     */
    getScrollPosition() {
      if (this.isRTL_()) {
        return this.computeCurrentScrollPositionRTL_();
      }

      const currentTranslateX = this.calculateCurrentTranslateX_();
      const scrollLeft = this.adapter_.getScrollAreaScrollLeft();
      return scrollLeft - currentTranslateX;
    }

    /**
     * Handles interaction events that occur during transition
     */
    handleInteraction() {
      // Early exit if we aren't animating
      if (!this.isAnimating_) {
        return;
      }

      // Prevent other event listeners from handling this event
      this.stopScrollAnimation_();
    }

    /**
     * Handles the transitionend event
     * @param {!Event} evt
     */
    handleTransitionEnd(evt) {
      // Early exit if we aren't animating or the event was triggered by a different element.
      if (!this.isAnimating_
        || !this.adapter_.eventTargetMatchesSelector(evt.target, MDCTabScrollerFoundation.strings.CONTENT_SELECTOR)) {
        return;
      }

      this.isAnimating_ = false;
      this.adapter_.removeClass(MDCTabScrollerFoundation.cssClasses.ANIMATING);
    }

    /**
     * Increment the scroll value by the scrollXIncrement
     * @param {number} scrollXIncrement The value by which to increment the scroll position
     */
    incrementScroll(scrollXIncrement) {
      // Early exit for non-operational increment values
      if (scrollXIncrement === 0) {
        return;
      }

      if (this.isRTL_()) {
        return this.incrementScrollRTL_(scrollXIncrement);
      }

      this.incrementScroll_(scrollXIncrement);
    }

    /**
     * Scrolls to the given scrollX value
     * @param {number} scrollX
     */
    scrollTo(scrollX) {
      if (this.isRTL_()) {
        return this.scrollToRTL_(scrollX);
      }

      this.scrollTo_(scrollX);
    }

    /**
     * Returns the appropriate version of the MDCTabScrollerRTL
     * @return {!MDCTabScrollerRTL}
     */
    getRTLScroller() {
      if (!this.rtlScrollerInstance_) {
        this.rtlScrollerInstance_ = this.rtlScrollerFactory_();
      }

      return this.rtlScrollerInstance_;
    }

    /**
     * Returns the translateX value from a CSS matrix transform function string
     * @return {number}
     * @private
     */
    calculateCurrentTranslateX_() {
      const transformValue = this.adapter_.getScrollContentStyleValue('transform');
      // Early exit if no transform is present
      if (transformValue === 'none') {
        return 0;
      }

      // The transform value comes back as a matrix transformation in the form
      // of `matrix(a, b, c, d, tx, ty)`. We only care about tx (translateX) so
      // we're going to grab all the parenthesized values, strip out tx, and
      // parse it.
      const results = /\((.+)\)/.exec(transformValue)[1];
      const parts = results.split(',');
      return parseFloat(parts[4]);
    }

    /**
     * Calculates a safe scroll value that is > 0 and < the max scroll value
     * @param {number} scrollX The distance to scroll
     * @return {number}
     * @private
     */
    clampScrollValue_(scrollX) {
      const edges = this.calculateScrollEdges_();
      return Math.min(Math.max(edges.left, scrollX), edges.right);
    }

    /**
     * @return {number}
     * @private
     */
    computeCurrentScrollPositionRTL_() {
      const translateX = this.calculateCurrentTranslateX_();
      return this.getRTLScroller().getScrollPositionRTL(translateX);
    }

    /**
     * @return {!MDCTabScrollerHorizontalEdges}
     * @private
     */
    calculateScrollEdges_() {
      const contentWidth = this.adapter_.getScrollContentOffsetWidth();
      const rootWidth = this.adapter_.getScrollAreaOffsetWidth();
      return /** @type {!MDCTabScrollerHorizontalEdges} */ ({
        left: 0,
        right: contentWidth - rootWidth,
      });
    }

    /**
     * Internal scroll method
     * @param {number} scrollX The new scroll position
     * @private
     */
    scrollTo_(scrollX) {
      const currentScrollX = this.getScrollPosition();
      const safeScrollX = this.clampScrollValue_(scrollX);
      const scrollDelta = safeScrollX - currentScrollX;
      this.animate_(/** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: safeScrollX,
        scrollDelta: scrollDelta,
      }));
    }

    /**
     * Internal RTL scroll method
     * @param {number} scrollX The new scroll position
     * @private
     */
    scrollToRTL_(scrollX) {
      const animation = this.getRTLScroller().scrollToRTL(scrollX);
      this.animate_(animation);
    }

    /**
     * Internal increment scroll method
     * @param {number} scrollX The new scroll position increment
     * @private
     */
    incrementScroll_(scrollX) {
      const currentScrollX = this.getScrollPosition();
      const targetScrollX = scrollX + currentScrollX;
      const safeScrollX = this.clampScrollValue_(targetScrollX);
      const scrollDelta = safeScrollX - currentScrollX;
      this.animate_(/** @type {!MDCTabScrollerAnimation} */ ({
        finalScrollPosition: safeScrollX,
        scrollDelta: scrollDelta,
      }));
    }

    /**
     * Internal incremenet scroll RTL method
     * @param {number} scrollX The new scroll position RTL increment
     * @private
     */
    incrementScrollRTL_(scrollX) {
      const animation = this.getRTLScroller().incrementScrollRTL(scrollX);
      this.animate_(animation);
    }

    /**
     * Animates the tab scrolling
     * @param {!MDCTabScrollerAnimation} animation The animation to apply
     * @private
     */
    animate_(animation) {
      // Early exit if translateX is 0, which means there's no animation to perform
      if (animation.scrollDelta === 0) {
        return;
      }

      this.stopScrollAnimation_();
      // This animation uses the FLIP approach.
      // Read more here: https://aerotwist.com/blog/flip-your-animations/
      this.adapter_.setScrollAreaScrollLeft(animation.finalScrollPosition);
      this.adapter_.setScrollContentStyleProperty('transform', `translateX(${animation.scrollDelta}px)`);
      // Force repaint
      this.adapter_.computeScrollAreaClientRect();

      requestAnimationFrame(() => {
        this.adapter_.addClass(MDCTabScrollerFoundation.cssClasses.ANIMATING);
        this.adapter_.setScrollContentStyleProperty('transform', 'none');
      });

      this.isAnimating_ = true;
    }

    /**
     * Stops scroll animation
     * @private
     */
    stopScrollAnimation_() {
      this.isAnimating_ = false;
      const currentScrollPosition = this.getAnimatingScrollPosition_();
      this.adapter_.removeClass(MDCTabScrollerFoundation.cssClasses.ANIMATING);
      this.adapter_.setScrollContentStyleProperty('transform', 'translateX(0px)');
      this.adapter_.setScrollAreaScrollLeft(currentScrollPosition);
    }

    /**
     * Gets the current scroll position during animation
     * @return {number}
     * @private
     */
    getAnimatingScrollPosition_() {
      const currentTranslateX = this.calculateCurrentTranslateX_();
      const scrollLeft = this.adapter_.getScrollAreaScrollLeft();
      if (this.isRTL_()) {
        return this.getRTLScroller().getAnimatingScrollPosition(scrollLeft, currentTranslateX);
      }

      return scrollLeft - currentTranslateX;
    }

    /**
     * Determines the RTL Scroller to use
     * @return {!MDCTabScrollerRTL}
     * @private
     */
    rtlScrollerFactory_() {
      // Browsers have three different implementations of scrollLeft in RTL mode,
      // dependent on the browser. The behavior is based off the max LTR
      // scrollleft value and 0.
      //
      // * Default scrolling in RTL *
      //    - Left-most value: 0
      //    - Right-most value: Max LTR scrollLeft value
      //
      // * Negative scrolling in RTL *
      //    - Left-most value: Negated max LTR scrollLeft value
      //    - Right-most value: 0
      //
      // * Reverse scrolling in RTL *
      //    - Left-most value: Max LTR scrollLeft value
      //    - Right-most value: 0
      //
      // We use those principles below to determine which RTL scrollLeft
      // behavior is implemented in the current browser.
      const initialScrollLeft = this.adapter_.getScrollAreaScrollLeft();
      this.adapter_.setScrollAreaScrollLeft(initialScrollLeft - 1);
      const newScrollLeft = this.adapter_.getScrollAreaScrollLeft();

      // If the newScrollLeft value is negative,then we know that the browser has
      // implemented negative RTL scrolling, since all other implementations have
      // only positive values.
      if (newScrollLeft < 0) {
        // Undo the scrollLeft test check
        this.adapter_.setScrollAreaScrollLeft(initialScrollLeft);
        return new MDCTabScrollerRTLNegative(this.adapter_);
      }

      const rootClientRect = this.adapter_.computeScrollAreaClientRect();
      const contentClientRect = this.adapter_.computeScrollContentClientRect();
      const rightEdgeDelta = Math.round(contentClientRect.right - rootClientRect.right);
      // Undo the scrollLeft test check
      this.adapter_.setScrollAreaScrollLeft(initialScrollLeft);

      // By calculating the clientRect of the root element and the clientRect of
      // the content element, we can determine how much the scroll value changed
      // when we performed the scrollLeft subtraction above.
      if (rightEdgeDelta === newScrollLeft) {
        return new MDCTabScrollerRTLReverse(this.adapter_);
      }

      return new MDCTabScrollerRTLDefault(this.adapter_);
    }

    /**
     * @return {boolean}
     * @private
     */
    isRTL_() {
      return this.adapter_.getScrollContentStyleValue('direction') === 'rtl';
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Stores result from computeHorizontalScrollbarHeight to avoid redundant processing.
   * @private {number|undefined}
   */
  let horizontalScrollbarHeight_;

  /**
   * Computes the height of browser-rendered horizontal scrollbars using a self-created test element.
   * May return 0 (e.g. on OS X browsers under default configuration).
   * @param {!Document} documentObj
   * @param {boolean=} shouldCacheResult
   * @return {number}
   */
  function computeHorizontalScrollbarHeight(documentObj, shouldCacheResult = true) {
    if (shouldCacheResult && typeof horizontalScrollbarHeight_ !== 'undefined') {
      return horizontalScrollbarHeight_;
    }

    const el = documentObj.createElement('div');
    el.classList.add(cssClasses$3.SCROLL_TEST);
    documentObj.body.appendChild(el);

    const horizontalScrollbarHeight = el.offsetHeight - el.clientHeight;
    documentObj.body.removeChild(el);

    if (shouldCacheResult) {
      horizontalScrollbarHeight_ = horizontalScrollbarHeight;
    }
    return horizontalScrollbarHeight;
  }

  /**
   * @param {!Object} HTMLElementPrototype
   * @return {!Array<string>}
   */
  function getMatchesProperty$1(HTMLElementPrototype) {
    return [
      'msMatchesSelector', 'matches',
    ].filter((p) => p in HTMLElementPrototype).pop();
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
   * @extends {MDCComponent<!MDCTabScrollerFoundation>}
   * @final
   */
  class MDCTabScroller extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCTabScroller}
     */
    static attachTo(root) {
      return new MDCTabScroller(root);
    }

    constructor(...args) {
      super(...args);

      /** @private {?Element} */
      this.content_;

      /** @private {?Element} */
      this.area_;

      /** @private {?function(?Event): undefined} */
      this.handleInteraction_;

      /** @private {?function(!Event): undefined} */
      this.handleTransitionEnd_;
    }

    initialize() {
      this.area_ = this.root_.querySelector(MDCTabScrollerFoundation.strings.AREA_SELECTOR);
      this.content_ = this.root_.querySelector(MDCTabScrollerFoundation.strings.CONTENT_SELECTOR);
    }

    initialSyncWithDOM() {
      this.handleInteraction_ = () => this.foundation_.handleInteraction();
      this.handleTransitionEnd_ = (evt) => this.foundation_.handleTransitionEnd(evt);

      this.area_.addEventListener('wheel', this.handleInteraction_);
      this.area_.addEventListener('touchstart', this.handleInteraction_);
      this.area_.addEventListener('pointerdown', this.handleInteraction_);
      this.area_.addEventListener('mousedown', this.handleInteraction_);
      this.area_.addEventListener('keydown', this.handleInteraction_);
      this.content_.addEventListener('transitionend', this.handleTransitionEnd_);
    }

    destroy() {
      super.destroy();

      this.area_.removeEventListener('wheel', this.handleInteraction_);
      this.area_.removeEventListener('touchstart', this.handleInteraction_);
      this.area_.removeEventListener('pointerdown', this.handleInteraction_);
      this.area_.removeEventListener('mousedown', this.handleInteraction_);
      this.area_.removeEventListener('keydown', this.handleInteraction_);
      this.content_.removeEventListener('transitionend', this.handleTransitionEnd_);
    }

    /**
     * @return {!MDCTabScrollerFoundation}
     */
    getDefaultFoundation() {
      const adapter = /** @type {!MDCTabScrollerAdapter} */ ({
        eventTargetMatchesSelector: (evtTarget, selector) => {
          const MATCHES = getMatchesProperty$1(HTMLElement.prototype);
          return evtTarget[MATCHES](selector);
        },
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        addScrollAreaClass: (className) => this.area_.classList.add(className),
        setScrollAreaStyleProperty: (prop, value) => this.area_.style.setProperty(prop, value),
        setScrollContentStyleProperty: (prop, value) => this.content_.style.setProperty(prop, value),
        getScrollContentStyleValue: (propName) => window.getComputedStyle(this.content_).getPropertyValue(propName),
        setScrollAreaScrollLeft: (scrollX) => this.area_.scrollLeft = scrollX,
        getScrollAreaScrollLeft: () => this.area_.scrollLeft,
        getScrollContentOffsetWidth: () => this.content_.offsetWidth,
        getScrollAreaOffsetWidth: () => this.area_.offsetWidth,
        computeScrollAreaClientRect: () => this.area_.getBoundingClientRect(),
        computeScrollContentClientRect: () => this.content_.getBoundingClientRect(),
        computeHorizontalScrollbarHeight: () => computeHorizontalScrollbarHeight(document),
      });

      return new MDCTabScrollerFoundation(adapter);
    }

    /**
     * Returns the current visual scroll position
     * @return {number}
     */
    getScrollPosition() {
      return this.foundation_.getScrollPosition();
    }

    /**
     * Returns the width of the scroll content
     * @return {number}
     */
    getScrollContentWidth() {
      return this.content_.offsetWidth;
    }

    /**
     * Increments the scroll value by the given amount
     * @param {number} scrollXIncrement The pixel value by which to increment the scroll value
     */
    incrementScroll(scrollXIncrement) {
      this.foundation_.incrementScroll(scrollXIncrement);
    }

    /**
     * Scrolls to the given pixel position
     * @param {number} scrollX The pixel value to scroll to
     */
    scrollTo(scrollX) {
      this.foundation_.scrollTo(scrollX);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
    */

  /** @enum {string} */
  const strings$4 = {
    TAB_ACTIVATED_EVENT: 'MDCTabBar:activated',
    TAB_SCROLLER_SELECTOR: '.mdc-tab-scroller',
    TAB_SELECTOR: '.mdc-tab',
    END_KEY: 'End',
    HOME_KEY: 'Home',
    ARROW_LEFT_KEY: 'ArrowLeft',
    ARROW_RIGHT_KEY: 'ArrowRight',
  };

  /** @enum {number} */
  const numbers$1 = {
    EXTRA_SCROLL_AMOUNT: 20,
    END_KEYCODE: 35,
    HOME_KEYCODE: 36,
    ARROW_LEFT_KEYCODE: 37,
    ARROW_RIGHT_KEYCODE: 39,
  };

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
  */
  /* eslint-enable no-unused-vars */

  /**
   * @type {Set<string>}
   */
  const ACCEPTABLE_KEYS = new Set();
  // IE11 has no support for new Set with iterable so we need to initialize this by hand
  ACCEPTABLE_KEYS.add(strings$4.ARROW_LEFT_KEY);
  ACCEPTABLE_KEYS.add(strings$4.ARROW_RIGHT_KEY);
  ACCEPTABLE_KEYS.add(strings$4.END_KEY);
  ACCEPTABLE_KEYS.add(strings$4.HOME_KEY);

  /**
   * @type {Map<number, string>}
   */
  const KEYCODE_MAP = new Map();
  // IE11 has no support for new Map with iterable so we need to initialize this by hand
  KEYCODE_MAP.set(numbers$1.HOME_KEYCODE, strings$4.HOME_KEY);
  KEYCODE_MAP.set(numbers$1.END_KEYCODE, strings$4.END_KEY);
  KEYCODE_MAP.set(numbers$1.ARROW_LEFT_KEYCODE, strings$4.ARROW_LEFT_KEY);
  KEYCODE_MAP.set(numbers$1.ARROW_RIGHT_KEYCODE, strings$4.ARROW_RIGHT_KEY);

  /**
   * @extends {MDCFoundation<!MDCTabBarAdapter>}
   * @final
   */
  class MDCTabBarFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get strings() {
      return strings$4;
    }

    /** @return enum {number} */
    static get numbers() {
      return numbers$1;
    }

    /**
     * @see MDCTabBarAdapter for typing information
     * @return {!MDCTabBarAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTabBarAdapter} */ ({
        scrollTo: () => {},
        incrementScroll: () => {},
        getScrollPosition: () => {},
        getScrollContentWidth: () => {},
        getOffsetWidth: () => {},
        isRTL: () => {},
        activateTabAtIndex: () => {},
        deactivateTabAtIndex: () => {},
        getTabIndicatorClientRectAtIndex: () => {},
        getTabDimensionsAtIndex: () => {},
        getActiveTabIndex: () => {},
        getIndexOfTab: () => {},
        getTabListLength: () => {},
        notifyTabActivated: () => {},
      });
    }

    /**
     * @param {!MDCTabBarAdapter} adapter
     * */
    constructor(adapter) {
      super(Object.assign(MDCTabBarFoundation.defaultAdapter, adapter));
    }

    init() {
      const activeIndex = this.adapter_.getActiveTabIndex();
      this.scrollIntoView(activeIndex);
    }

    /**
     * Activates the tab at the given index
     * @param {number} index
     */
    activateTab(index) {
      const previousActiveIndex = this.adapter_.getActiveTabIndex();
      if (!this.indexIsInRange_(index)) {
        return;
      }

      this.adapter_.deactivateTabAtIndex(previousActiveIndex);
      this.adapter_.activateTabAtIndex(index, this.adapter_.getTabIndicatorClientRectAtIndex(previousActiveIndex));
      this.scrollIntoView(index);

      // Only notify the tab activation if the index is different than the previously active index
      if (index !== previousActiveIndex) {
        this.adapter_.notifyTabActivated(index);
      }
    }

    /**
     * Handles the keydown event
     * @param {!Event} evt
     */
    handleKeyDown(evt) {
      // Get the key from the event
      const key = this.getKeyFromEvent_(evt);

      // Early exit if the event key isn't one of the keyboard navigation keys
      if (key === undefined) {
        return;
      }

      evt.preventDefault();
      this.activateTabFromKey_(key);
    }

    /**
     * Handles the MDCTab:interacted event
     * @param {!Event} evt
     */
    handleTabInteraction(evt) {
      this.activateTab(this.adapter_.getIndexOfTab(evt.detail.tab));
    }

    /**
     * Scrolls the tab at the given index into view
     * @param {number} index The tab index to make visible
     */
    scrollIntoView(index) {
      // Early exit if the index is out of range
      if (!this.indexIsInRange_(index)) {
        return;
      }

      // Always scroll to 0 if scrolling to the 0th index
      if (index === 0) {
        return this.adapter_.scrollTo(0);
      }

      // Always scroll to the max value if scrolling to the Nth index
      // MDCTabScroller.scrollTo() will never scroll past the max possible value
      if (index === this.adapter_.getTabListLength() - 1) {
        return this.adapter_.scrollTo(this.adapter_.getScrollContentWidth());
      }

      if (this.isRTL_()) {
        return this.scrollIntoViewRTL_(index);
      }

      this.scrollIntoView_(index);
    }

    /**
     * Private method for activating a tab from a key
     * @param {string} key The name of the key
     * @private
     */
    activateTabFromKey_(key) {
      const isRTL = this.isRTL_();
      const maxTabIndex = this.adapter_.getTabListLength() - 1;
      const shouldGoToEnd = key === strings$4.END_KEY;
      const shouldDecrement = key === strings$4.ARROW_LEFT_KEY && !isRTL || key === strings$4.ARROW_RIGHT_KEY && isRTL;
      const shouldIncrement = key === strings$4.ARROW_RIGHT_KEY && !isRTL || key === strings$4.ARROW_LEFT_KEY && isRTL;
      let tabIndex = this.adapter_.getActiveTabIndex();

      if (shouldGoToEnd) {
        tabIndex = maxTabIndex;
      } else if (shouldDecrement) {
        tabIndex -= 1;
      } else if (shouldIncrement) {
        tabIndex += 1;
      } else {
        tabIndex = 0;
      }

      if (tabIndex < 0) {
        tabIndex = maxTabIndex;
      } else if (tabIndex > maxTabIndex) {
        tabIndex = 0;
      }

      this.activateTab(tabIndex);
    }

    /**
     * Calculates the scroll increment that will make the tab at the given index visible
     * @param {number} index The index of the tab
     * @param {number} nextIndex The index of the next tab
     * @param {number} scrollPosition The current scroll position
     * @param {number} barWidth The width of the Tab Bar
     * @return {number}
     * @private
     */
    calculateScrollIncrement_(index, nextIndex, scrollPosition, barWidth) {
      const nextTabDimensions = this.adapter_.getTabDimensionsAtIndex(nextIndex);
      const relativeContentLeft = nextTabDimensions.contentLeft - scrollPosition - barWidth;
      const relativeContentRight = nextTabDimensions.contentRight - scrollPosition;
      const leftIncrement = relativeContentRight - numbers$1.EXTRA_SCROLL_AMOUNT;
      const rightIncrement = relativeContentLeft + numbers$1.EXTRA_SCROLL_AMOUNT;

      if (nextIndex < index) {
        return Math.min(leftIncrement, 0);
      }

      return Math.max(rightIncrement, 0);
    }

    /**
     * Calculates the scroll increment that will make the tab at the given index visible in RTL
     * @param {number} index The index of the tab
     * @param {number} nextIndex The index of the next tab
     * @param {number} scrollPosition The current scroll position
     * @param {number} barWidth The width of the Tab Bar
     * @param {number} scrollContentWidth The width of the scroll content
     * @return {number}
     * @private
     */
    calculateScrollIncrementRTL_(index, nextIndex, scrollPosition, barWidth, scrollContentWidth, ) {
      const nextTabDimensions = this.adapter_.getTabDimensionsAtIndex(nextIndex);
      const relativeContentLeft = scrollContentWidth - nextTabDimensions.contentLeft - scrollPosition;
      const relativeContentRight = scrollContentWidth - nextTabDimensions.contentRight - scrollPosition - barWidth;
      const leftIncrement = relativeContentRight + numbers$1.EXTRA_SCROLL_AMOUNT;
      const rightIncrement = relativeContentLeft - numbers$1.EXTRA_SCROLL_AMOUNT;

      if (nextIndex > index) {
        return Math.max(leftIncrement, 0);
      }

      return Math.min(rightIncrement, 0);
    }

    /**
     * Determines the index of the adjacent tab closest to either edge of the Tab Bar
     * @param {number} index The index of the tab
     * @param {!MDCTabDimensions} tabDimensions The dimensions of the tab
     * @param {number} scrollPosition The current scroll position
     * @param {number} barWidth The width of the tab bar
     * @return {number}
     * @private
     */
    findAdjacentTabIndexClosestToEdge_(index, tabDimensions, scrollPosition, barWidth) {
      /**
       * Tabs are laid out in the Tab Scroller like this:
       *
       *    Scroll Position
       *    +---+
       *    |   |   Bar Width
       *    |   +-----------------------------------+
       *    |   |                                   |
       *    |   V                                   V
       *    |   +-----------------------------------+
       *    V   |             Tab Scroller          |
       *    +------------+--------------+-------------------+
       *    |    Tab     |      Tab     |        Tab        |
       *    +------------+--------------+-------------------+
       *        |                                   |
       *        +-----------------------------------+
       *
       * To determine the next adjacent index, we look at the Tab root left and
       * Tab root right, both relative to the scroll position. If the Tab root
       * left is less than 0, then we know it's out of view to the left. If the
       * Tab root right minus the bar width is greater than 0, we know the Tab is
       * out of view to the right. From there, we either increment or decrement
       * the index.
       */
      const relativeRootLeft = tabDimensions.rootLeft - scrollPosition;
      const relativeRootRight = tabDimensions.rootRight - scrollPosition - barWidth;
      const relativeRootDelta = relativeRootLeft + relativeRootRight;
      const leftEdgeIsCloser = relativeRootLeft < 0 || relativeRootDelta < 0;
      const rightEdgeIsCloser = relativeRootRight > 0 || relativeRootDelta > 0;

      if (leftEdgeIsCloser) {
        return index - 1;
      }

      if (rightEdgeIsCloser) {
        return index + 1;
      }

      return -1;
    }

    /**
     * Determines the index of the adjacent tab closest to either edge of the Tab Bar in RTL
     * @param {number} index The index of the tab
     * @param {!MDCTabDimensions} tabDimensions The dimensions of the tab
     * @param {number} scrollPosition The current scroll position
     * @param {number} barWidth The width of the tab bar
     * @param {number} scrollContentWidth The width of the scroller content
     * @return {number}
     * @private
     */
    findAdjacentTabIndexClosestToEdgeRTL_(index, tabDimensions, scrollPosition, barWidth, scrollContentWidth) {
      const rootLeft = scrollContentWidth - tabDimensions.rootLeft - barWidth - scrollPosition;
      const rootRight = scrollContentWidth - tabDimensions.rootRight - scrollPosition;
      const rootDelta = rootLeft + rootRight;
      const leftEdgeIsCloser = rootLeft > 0 || rootDelta > 0;
      const rightEdgeIsCloser = rootRight < 0 || rootDelta < 0;

      if (leftEdgeIsCloser) {
        return index + 1;
      }

      if (rightEdgeIsCloser) {
        return index - 1;
      }

      return -1;
    }

    /**
     * Returns the key associated with a keydown event
     * @param {!Event} evt The keydown event
     * @return {string}
     * @private
     */
    getKeyFromEvent_(evt) {
      if (ACCEPTABLE_KEYS.has(evt.key)) {
        return evt.key;
      }

      return KEYCODE_MAP.get(evt.keyCode);
    }

    /**
     * Returns whether a given index is inclusively between the ends
     * @param {number} index The index to test
     * @private
     */
    indexIsInRange_(index) {
      return index >= 0 && index < this.adapter_.getTabListLength();
    }

    /**
     * Returns the view's RTL property
     * @return {boolean}
     * @private
     */
    isRTL_() {
      return this.adapter_.isRTL();
    }

    /**
     * Scrolls the tab at the given index into view for left-to-right useragents
     * @param {number} index The index of the tab to scroll into view
     * @private
     */
    scrollIntoView_(index) {
      const scrollPosition = this.adapter_.getScrollPosition();
      const barWidth = this.adapter_.getOffsetWidth();
      const tabDimensions = this.adapter_.getTabDimensionsAtIndex(index);
      const nextIndex = this.findAdjacentTabIndexClosestToEdge_(index, tabDimensions, scrollPosition, barWidth);

      if (!this.indexIsInRange_(nextIndex)) {
        return;
      }

      const scrollIncrement = this.calculateScrollIncrement_(index, nextIndex, scrollPosition, barWidth);
      this.adapter_.incrementScroll(scrollIncrement);
    }

    /**
     * Scrolls the tab at the given index into view in RTL
     * @param {number} index The tab index to make visible
     * @private
     */
    scrollIntoViewRTL_(index) {
      const scrollPosition = this.adapter_.getScrollPosition();
      const barWidth = this.adapter_.getOffsetWidth();
      const tabDimensions = this.adapter_.getTabDimensionsAtIndex(index);
      const scrollWidth = this.adapter_.getScrollContentWidth();
      const nextIndex = this.findAdjacentTabIndexClosestToEdgeRTL_(
        index, tabDimensions, scrollPosition, barWidth, scrollWidth);

      if (!this.indexIsInRange_(nextIndex)) {
        return;
      }

      const scrollIncrement = this.calculateScrollIncrementRTL_(index, nextIndex, scrollPosition, barWidth, scrollWidth);
      this.adapter_.incrementScroll(scrollIncrement);
    }
  }

  /**
    * @license
    * Copyright 2018 Google Inc. All Rights Reserved.
    *
    * Licensed under the Apache License, Version 2.0 (the "License")
    * you may not use this file except in compliance with the License.
    * You may obtain a copy of the License at
    *
    *      http://www.apache.org/licenses/LICENSE-2.0
    *
    * Unless required by applicable law or agreed to in writing, software
    * distributed under the License is distributed on an "AS IS" BASIS,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    * See the License for the specific language governing permissions and
    * limitations under the License.
    */

  /**
   * @extends {MDCComponent<!MDCTabBarFoundation>}
   * @final
   */
  class MDCTabBar extends MDCComponent {
    /**
     * @param {...?} args
     */
    constructor(...args) {
      super(...args);

      /** @private {!Array<!MDCTab>} */
      this.tabList_;

      /** @type {(function(!Element): !MDCTab)} */
      this.tabFactory_;

      /** @private {?MDCTabScroller} */
      this.tabScroller_;

      /** @type {(function(!Element): !MDCTabScroller)} */
      this.tabScrollerFactory_;

      /** @private {?function(?Event): undefined} */
      this.handleTabInteraction_;

      /** @private {?function(?Event): undefined} */
      this.handleKeyDown_;
    }

    /**
     * @param {!Element} root
     * @return {!MDCTabBar}
     */
    static attachTo(root) {
      return new MDCTabBar(root);
    }

    /**
     * @param {(function(!Element): !MDCTab)=} tabFactory A function which creates a new MDCTab
     * @param {(function(!Element): !MDCTabScroller)=} tabScrollerFactory A function which creates a new MDCTabScroller
     */
    initialize(
      tabFactory = (el) => new MDCTab(el),
      tabScrollerFactory = (el) => new MDCTabScroller(el),
    ) {
      this.tabFactory_ = tabFactory;
      this.tabScrollerFactory_ = tabScrollerFactory;

      const tabElements = [].slice.call(this.root_.querySelectorAll(MDCTabBarFoundation.strings.TAB_SELECTOR));
      this.tabList_ = tabElements.map((el) => this.tabFactory_(el));

      const tabScrollerElement = this.root_.querySelector(MDCTabBarFoundation.strings.TAB_SCROLLER_SELECTOR);
      if (tabScrollerElement) {
        this.tabScroller_ = this.tabScrollerFactory_(tabScrollerElement);
      }
    }

    initialSyncWithDOM() {
      this.handleTabInteraction_ = (evt) => this.foundation_.handleTabInteraction(evt);
      this.handleKeyDown_ = (evt) => this.foundation_.handleKeyDown(evt);

      this.root_.addEventListener(MDCTabFoundation.strings.INTERACTED_EVENT, this.handleTabInteraction_);
      this.root_.addEventListener('keydown', this.handleKeyDown_);
    }

    destroy() {
      super.destroy();
      this.root_.removeEventListener(MDCTabFoundation.strings.INTERACTED_EVENT, this.handleTabInteraction_);
      this.root_.removeEventListener('keydown', this.handleKeyDown_);
      this.tabList_.forEach((tab) => tab.destroy());
      this.tabScroller_.destroy();
    }

    /**
     * @return {!MDCTabBarFoundation}
     */
    getDefaultFoundation() {
      return new MDCTabBarFoundation(
        /** @type {!MDCTabBarAdapter} */ ({
          scrollTo: (scrollX) => this.tabScroller_.scrollTo(scrollX),
          incrementScroll: (scrollXIncrement) => this.tabScroller_.incrementScroll(scrollXIncrement),
          getScrollPosition: () => this.tabScroller_.getScrollPosition(),
          getScrollContentWidth: () => this.tabScroller_.getScrollContentWidth(),
          getOffsetWidth: () => this.root_.offsetWidth,
          isRTL: () => window.getComputedStyle(this.root_).getPropertyValue('direction') === 'rtl',
          activateTabAtIndex: (index, clientRect) => this.tabList_[index].activate(clientRect),
          deactivateTabAtIndex: (index) => this.tabList_[index].deactivate(),
          getTabIndicatorClientRectAtIndex: (index) => this.tabList_[index].computeIndicatorClientRect(),
          getTabDimensionsAtIndex: (index) => this.tabList_[index].computeDimensions(),
          getActiveTabIndex: () => {
            for (let i = 0; i < this.tabList_.length; i++) {
              if (this.tabList_[i].active) {
                return i;
              }
            }
            return -1;
          },
          getIndexOfTab: (tabToFind) => this.tabList_.indexOf(tabToFind),
          getTabListLength: () => this.tabList_.length,
          notifyTabActivated: (index) => this.emit(MDCTabBarFoundation.strings.TAB_ACTIVATED_EVENT, {index}, true),
        })
      );
    }

    /**
     * Activates the tab at the given index
     * @param {number} index The index of the tab
     */
    activateTab(index) {
      this.foundation_.activateTab(index);
    }

    /**
     * Scrolls the tab at the given index into view
     * @param {number} index THe index of the tab
     */
    scrollIntoView(index) {
      this.foundation_.scrollIntoView(index);
    }
  }

  const regex = /%s/gi;

  function processClasses(classes, userClasses) {
      if (!userClasses) {
          return classes.join(" ");
      }
      if (Array.isArray(userClasses)) {
          userClasses = userClasses.join(" ");
      }
      if (!regex.test(userClasses)) {
          return userClasses+" "+classes.join(" ");
      }
      return userClasses.replace(regex, classes.join(" "));
  }

  /* src/TabBar.html generated by Svelte v2.9.11 */

  function attrs(attributes) {
      let result = Object.assign({role:"tablist"}, attributes);
      let classes = ["mdc-tab-bar"];
      delete result.attrs;
      result["class"] = processClasses(classes, result["class"]);
      return result;
  }
  function oncreate() {
      if (this.root && this.root.getTabsComponent) {
          this.mdcComponent = this.root.getTabsComponent();
      } else {
          this.mdcComponent = new MDCTabBar(this.refs.bar);
      }
      this.mdcComponent.listen("MDCTabBar:activated", (data) => {
          this.fire("change", {"activeTabIndex": data.detail.index});
          if (data.detail.index !== this.get().activeTabIndex) {
              this.set({"activeTabIndex": data.detail.index});
          }
      });
  }
  function ondestroy() {
      if (!!this.mdcComponent) {
          this.mdcComponent.destroy();
      }
  }
  function onstate({current, changed}) {
      if (changed["activeTabIndex"] && current["activeTabIndex"]) {
          this.mdcComponent.activateTab(current["activeTabIndex"]);
      }
  }
  function create_main_fragment(component, ctx) {
  	var div, slot_content_default = component._slotted.default;

  	var div_levels = [
  		ctx.attrs
  	];

  	var div_data = {};
  	for (var i = 0; i < div_levels.length; i += 1) {
  		div_data = shared_js.assign(div_data, div_levels[i]);
  	}

  	return {
  		c() {
  			div = shared_js.createElement("div");
  			shared_js.setAttributes(div, div_data);
  		},

  		m(target, anchor) {
  			shared_js.insert(target, div, anchor);

  			if (slot_content_default) {
  				shared_js.append(div, slot_content_default);
  			}

  			component.refs.bar = div;
  		},

  		p(changed, ctx) {
  			shared_js.setAttributes(div, shared_js.getSpreadUpdate(div_levels, [
  				(changed.attrs) && ctx.attrs
  			]));
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(div);
  			}

  			if (slot_content_default) {
  				shared_js.reinsertChildren(div, slot_content_default);
  			}

  			if (component.refs.bar === div) component.refs.bar = null;
  		}
  	};
  }

  function TabBar(options) {
  	shared_js.init(this, options);
  	this.refs = {};
  	this._state = shared_js.assign({}, options.data);
  	this._recompute({  }, this._state);
  	this._intro = true;

  	this._handlers.state = [onstate];

  	this._handlers.destroy = [ondestroy];

  	this._slotted = options.slots || {};

  	if (!options.root) {
  		this._oncreate = [];
  	}

  	this.slots = {};

  	this._fragment = create_main_fragment(this, this._state);

  	this.root._oncreate.push(() => {
  		onstate.call(this, { changed: shared_js.assignTrue({}, this._state), current: this._state });
  		oncreate.call(this);
  		this.fire("update", { changed: shared_js.assignTrue({}, this._state), current: this._state });
  	});

  	if (options.target) {
  		this._fragment.c();
  		this._mount(options.target, options.anchor);

  		shared_js.callAll(this._oncreate);
  	}
  }

  shared_js.assign(TabBar.prototype, shared_js.proto);

  TabBar.prototype._recompute = function _recompute(changed, state) {
  	if (this._differs(state.attrs, (state.attrs = attrs(state)))) changed.attrs = true;
  };

  return TabBar;

})));
