(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('./shared.js')) :
  typeof define === 'function' && define.amd ? define(['./shared.js'], factory) :
  (global.TextArea = factory(global.svelteShared));
}(this, (function (shared_js) { 'use strict';

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template A
   */
  class MDCFoundation {
    /** @return enum{cssClasses} */
    static get cssClasses() {
      // Classes extending MDCFoundation should implement this method to return an object which exports every
      // CSS class the foundation class needs as a property. e.g. {ACTIVE: 'mdc-component--active'}
      return {};
    }

    /** @return enum{strings} */
    static get strings() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // semantic strings as constants. e.g. {ARIA_ROLE: 'tablist'}
      return {};
    }

    /** @return enum{numbers} */
    static get numbers() {
      // Classes extending MDCFoundation should implement this method to return an object which exports all
      // of its semantic numbers as constants. e.g. {ANIMATION_DELAY_MS: 350}
      return {};
    }

    /** @return {!Object} */
    static get defaultAdapter() {
      // Classes extending MDCFoundation may choose to implement this getter in order to provide a convenient
      // way of viewing the necessary methods of an adapter. In the future, this could also be used for adapter
      // validation.
      return {};
    }

    /**
     * @param {A=} adapter
     */
    constructor(adapter = {}) {
      /** @protected {!A} */
      this.adapter_ = adapter;
    }

    init() {
      // Subclasses should override this method to perform initialization routines (registering events, etc.)
    }

    destroy() {
      // Subclasses should override this method to perform de-initialization routines (de-registering events, etc.)
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *   http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @template F
   */
  class MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCComponent}
     */
    static attachTo(root) {
      // Subclasses which extend MDCBase should provide an attachTo() method that takes a root element and
      // returns an instantiated component with its root set to that element. Also note that in the cases of
      // subclasses, an explicit foundation class will not have to be passed in; it will simply be initialized
      // from getDefaultFoundation().
      return new MDCComponent(root, new MDCFoundation());
    }

    /**
     * @param {!Element} root
     * @param {F=} foundation
     * @param {...?} args
     */
    constructor(root, foundation = undefined, ...args) {
      /** @protected {!Element} */
      this.root_ = root;
      this.initialize(...args);
      // Note that we initialize foundation here and not within the constructor's default param so that
      // this.root_ is defined and can be used within the foundation class.
      /** @protected {!F} */
      this.foundation_ = foundation === undefined ? this.getDefaultFoundation() : foundation;
      this.foundation_.init();
      this.initialSyncWithDOM();
    }

    initialize(/* ...args */) {
      // Subclasses can override this to do any additional setup work that would be considered part of a
      // "constructor". Essentially, it is a hook into the parent constructor before the foundation is
      // initialized. Any additional arguments besides root and foundation will be passed in here.
    }

    /**
     * @return {!F} foundation
     */
    getDefaultFoundation() {
      // Subclasses must override this method to return a properly configured foundation class for the
      // component.
      throw new Error('Subclasses must override getDefaultFoundation to return a properly configured ' +
        'foundation class');
    }

    initialSyncWithDOM() {
      // Subclasses should override this method if they need to perform work to synchronize with a host DOM
      // object. An example of this would be a form control wrapper that needs to synchronize its internal state
      // to some property or attribute of the host DOM. Please note: this is *not* the place to perform DOM
      // reads/writes that would cause layout / paint, as this is called synchronously from within the constructor.
    }

    destroy() {
      // Subclasses may implement this method to release any resources / deregister any listeners they have
      // attached. An example of this might be deregistering a resize event from the window object.
      this.foundation_.destroy();
    }

    /**
     * Wrapper method to add an event listener to the component's root element. This is most useful when
     * listening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    listen(evtType, handler) {
      this.root_.addEventListener(evtType, handler);
    }

    /**
     * Wrapper method to remove an event listener to the component's root element. This is most useful when
     * unlistening for custom events.
     * @param {string} evtType
     * @param {!Function} handler
     */
    unlisten(evtType, handler) {
      this.root_.removeEventListener(evtType, handler);
    }

    /**
     * Fires a cross-browser-compatible custom event from the component root of the given type,
     * with the given data.
     * @param {string} evtType
     * @param {!Object} evtData
     * @param {boolean=} shouldBubble
     */
    emit(evtType, evtData, shouldBubble = false) {
      let evt;
      if (typeof CustomEvent === 'function') {
        evt = new CustomEvent(evtType, {
          detail: evtData,
          bubbles: shouldBubble,
        });
      } else {
        evt = document.createEvent('CustomEvent');
        evt.initCustomEvent(evtType, shouldBubble, false, evtData);
      }

      this.root_.dispatchEvent(evt);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  const cssClasses = {
    // Ripple is a special case where the "root" component is really a "mixin" of sorts,
    // given that it's an 'upgrade' to an existing component. That being said it is the root
    // CSS class that all other CSS classes derive from.
    ROOT: 'mdc-ripple-upgraded',
    UNBOUNDED: 'mdc-ripple-upgraded--unbounded',
    BG_FOCUSED: 'mdc-ripple-upgraded--background-focused',
    FG_ACTIVATION: 'mdc-ripple-upgraded--foreground-activation',
    FG_DEACTIVATION: 'mdc-ripple-upgraded--foreground-deactivation',
  };

  const strings = {
    VAR_LEFT: '--mdc-ripple-left',
    VAR_TOP: '--mdc-ripple-top',
    VAR_FG_SIZE: '--mdc-ripple-fg-size',
    VAR_FG_SCALE: '--mdc-ripple-fg-scale',
    VAR_FG_TRANSLATE_START: '--mdc-ripple-fg-translate-start',
    VAR_FG_TRANSLATE_END: '--mdc-ripple-fg-translate-end',
  };

  const numbers = {
    PADDING: 10,
    INITIAL_ORIGIN_SCALE: 0.6,
    DEACTIVATION_TIMEOUT_MS: 225, // Corresponds to $mdc-ripple-translate-duration (i.e. activation animation duration)
    FG_DEACTIVATION_MS: 150, // Corresponds to $mdc-ripple-fade-out-duration (i.e. deactivation animation duration)
    TAP_DELAY_MS: 300, // Delay between touch and simulated mouse events on touch devices
  };

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * Stores result from supportsCssVariables to avoid redundant processing to detect CSS custom variable support.
   * @private {boolean|undefined}
   */
  let supportsCssVariables_;

  /**
   * Stores result from applyPassive to avoid redundant processing to detect passive event listener support.
   * @private {boolean|undefined}
   */
  let supportsPassive_;

  /**
   * @param {!Window} windowObj
   * @return {boolean}
   */
  function detectEdgePseudoVarBug(windowObj) {
    // Detect versions of Edge with buggy var() support
    // See: https://developer.microsoft.com/en-us/microsoft-edge/platform/issues/11495448/
    const document = windowObj.document;
    const node = document.createElement('div');
    node.className = 'mdc-ripple-surface--test-edge-var-bug';
    document.body.appendChild(node);

    // The bug exists if ::before style ends up propagating to the parent element.
    // Additionally, getComputedStyle returns null in iframes with display: "none" in Firefox,
    // but Firefox is known to support CSS custom properties correctly.
    // See: https://bugzilla.mozilla.org/show_bug.cgi?id=548397
    const computedStyle = windowObj.getComputedStyle(node);
    const hasPseudoVarBug = computedStyle !== null && computedStyle.borderTopStyle === 'solid';
    node.remove();
    return hasPseudoVarBug;
  }

  /**
   * @param {!Window} windowObj
   * @param {boolean=} forceRefresh
   * @return {boolean|undefined}
   */

  function supportsCssVariables(windowObj, forceRefresh = false) {
    let supportsCssVariables = supportsCssVariables_;
    if (typeof supportsCssVariables_ === 'boolean' && !forceRefresh) {
      return supportsCssVariables;
    }

    const supportsFunctionPresent = windowObj.CSS && typeof windowObj.CSS.supports === 'function';
    if (!supportsFunctionPresent) {
      return;
    }

    const explicitlySupportsCssVars = windowObj.CSS.supports('--css-vars', 'yes');
    // See: https://bugs.webkit.org/show_bug.cgi?id=154669
    // See: README section on Safari
    const weAreFeatureDetectingSafari10plus = (
      windowObj.CSS.supports('(--css-vars: yes)') &&
      windowObj.CSS.supports('color', '#00000000')
    );

    if (explicitlySupportsCssVars || weAreFeatureDetectingSafari10plus) {
      supportsCssVariables = !detectEdgePseudoVarBug(windowObj);
    } else {
      supportsCssVariables = false;
    }

    if (!forceRefresh) {
      supportsCssVariables_ = supportsCssVariables;
    }
    return supportsCssVariables;
  }

  //
  /**
   * Determine whether the current browser supports passive event listeners, and if so, use them.
   * @param {!Window=} globalObj
   * @param {boolean=} forceRefresh
   * @return {boolean|{passive: boolean}}
   */
  function applyPassive(globalObj = window, forceRefresh = false) {
    if (supportsPassive_ === undefined || forceRefresh) {
      let isSupported = false;
      try {
        globalObj.document.addEventListener('test', null, {get passive() {
          isSupported = true;
        }});
      } catch (e) { }

      supportsPassive_ = isSupported;
    }

    return supportsPassive_ ? {passive: true} : false;
  }

  /**
   * @param {!Object} HTMLElementPrototype
   * @return {!Array<string>}
   */
  function getMatchesProperty(HTMLElementPrototype) {
    return [
      'webkitMatchesSelector', 'msMatchesSelector', 'matches',
    ].filter((p) => p in HTMLElementPrototype).pop();
  }

  /**
   * @param {!Event} ev
   * @param {{x: number, y: number}} pageOffset
   * @param {!ClientRect} clientRect
   * @return {{x: number, y: number}}
   */
  function getNormalizedEventCoords(ev, pageOffset, clientRect) {
    const {x, y} = pageOffset;
    const documentX = x + clientRect.left;
    const documentY = y + clientRect.top;

    let normalizedX;
    let normalizedY;
    // Determine touch point relative to the ripple container.
    if (ev.type === 'touchstart') {
      normalizedX = ev.changedTouches[0].pageX - documentX;
      normalizedY = ev.changedTouches[0].pageY - documentY;
    } else {
      normalizedX = ev.pageX - documentX;
      normalizedY = ev.pageY - documentY;
    }

    return {x: normalizedX, y: normalizedY};
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  // Activation events registered on the root element of each instance for activation
  const ACTIVATION_EVENT_TYPES = ['touchstart', 'pointerdown', 'mousedown', 'keydown'];

  // Deactivation events registered on documentElement when a pointer-related down event occurs
  const POINTER_DEACTIVATION_EVENT_TYPES = ['touchend', 'pointerup', 'mouseup'];

  // Tracks activations that have occurred on the current frame, to avoid simultaneous nested activations
  /** @type {!Array<!EventTarget>} */
  let activatedTargets = [];

  /**
   * @extends {MDCFoundation<!MDCRippleAdapter>}
   */
  class MDCRippleFoundation extends MDCFoundation {
    static get cssClasses() {
      return cssClasses;
    }

    static get strings() {
      return strings;
    }

    static get numbers() {
      return numbers;
    }

    static get defaultAdapter() {
      return {
        browserSupportsCssVars: () => /* boolean - cached */ {},
        isUnbounded: () => /* boolean */ {},
        isSurfaceActive: () => /* boolean */ {},
        isSurfaceDisabled: () => /* boolean */ {},
        addClass: (/* className: string */) => {},
        removeClass: (/* className: string */) => {},
        containsEventTarget: (/* target: !EventTarget */) => {},
        registerInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        deregisterDocumentInteractionHandler: (/* evtType: string, handler: EventListener */) => {},
        registerResizeHandler: (/* handler: EventListener */) => {},
        deregisterResizeHandler: (/* handler: EventListener */) => {},
        updateCssVariable: (/* varName: string, value: string */) => {},
        computeBoundingRect: () => /* ClientRect */ {},
        getWindowPageOffset: () => /* {x: number, y: number} */ {},
      };
    }

    constructor(adapter) {
      super(Object.assign(MDCRippleFoundation.defaultAdapter, adapter));

      /** @private {number} */
      this.layoutFrame_ = 0;

      /** @private {!ClientRect} */
      this.frame_ = /** @type {!ClientRect} */ ({width: 0, height: 0});

      /** @private {!ActivationStateType} */
      this.activationState_ = this.defaultActivationState_();

      /** @private {number} */
      this.initialSize_ = 0;

      /** @private {number} */
      this.maxRadius_ = 0;

      /** @private {function(!Event)} */
      this.activateHandler_ = (e) => this.activate_(e);

      /** @private {function(!Event)} */
      this.deactivateHandler_ = (e) => this.deactivate_(e);

      /** @private {function(?Event=)} */
      this.focusHandler_ = () => this.handleFocus();

      /** @private {function(?Event=)} */
      this.blurHandler_ = () => this.handleBlur();

      /** @private {!Function} */
      this.resizeHandler_ = () => this.layout();

      /** @private {{left: number, top:number}} */
      this.unboundedCoords_ = {
        left: 0,
        top: 0,
      };

      /** @private {number} */
      this.fgScale_ = 0;

      /** @private {number} */
      this.activationTimer_ = 0;

      /** @private {number} */
      this.fgDeactivationRemovalTimer_ = 0;

      /** @private {boolean} */
      this.activationAnimationHasEnded_ = false;

      /** @private {!Function} */
      this.activationTimerCallback_ = () => {
        this.activationAnimationHasEnded_ = true;
        this.runDeactivationUXLogicIfReady_();
      };

      /** @private {?Event} */
      this.previousActivationEvent_ = null;
    }

    /**
     * We compute this property so that we are not querying information about the client
     * until the point in time where the foundation requests it. This prevents scenarios where
     * client-side feature-detection may happen too early, such as when components are rendered on the server
     * and then initialized at mount time on the client.
     * @return {boolean}
     * @private
     */
    isSupported_() {
      return this.adapter_.browserSupportsCssVars();
    }

    /**
     * @return {!ActivationStateType}
     */
    defaultActivationState_() {
      return {
        isActivated: false,
        hasDeactivationUXRun: false,
        wasActivatedByPointer: false,
        wasElementMadeActive: false,
        activationEvent: null,
        isProgrammatic: false,
      };
    }

    /** @override */
    init() {
      if (!this.isSupported_()) {
        return;
      }
      this.registerRootHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.addClass(ROOT);
        if (this.adapter_.isUnbounded()) {
          this.adapter_.addClass(UNBOUNDED);
          // Unbounded ripples need layout logic applied immediately to set coordinates for both shade and ripple
          this.layoutInternal_();
        }
      });
    }

    /** @override */
    destroy() {
      if (!this.isSupported_()) {
        return;
      }

      if (this.activationTimer_) {
        clearTimeout(this.activationTimer_);
        this.activationTimer_ = 0;
        const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
        this.adapter_.removeClass(FG_ACTIVATION);
      }

      this.deregisterRootHandlers_();
      this.deregisterDeactivationHandlers_();

      const {ROOT, UNBOUNDED} = MDCRippleFoundation.cssClasses;
      requestAnimationFrame(() => {
        this.adapter_.removeClass(ROOT);
        this.adapter_.removeClass(UNBOUNDED);
        this.removeCssVars_();
      });
    }

    /** @private */
    registerRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.registerInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.registerInteractionHandler('focus', this.focusHandler_);
      this.adapter_.registerInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.registerResizeHandler(this.resizeHandler_);
      }
    }

    /**
     * @param {!Event} e
     * @private
     */
    registerDeactivationHandlers_(e) {
      if (e.type === 'keydown') {
        this.adapter_.registerInteractionHandler('keyup', this.deactivateHandler_);
      } else {
        POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
          this.adapter_.registerDocumentInteractionHandler(type, this.deactivateHandler_);
        });
      }
    }

    /** @private */
    deregisterRootHandlers_() {
      ACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterInteractionHandler(type, this.activateHandler_);
      });
      this.adapter_.deregisterInteractionHandler('focus', this.focusHandler_);
      this.adapter_.deregisterInteractionHandler('blur', this.blurHandler_);

      if (this.adapter_.isUnbounded()) {
        this.adapter_.deregisterResizeHandler(this.resizeHandler_);
      }
    }

    /** @private */
    deregisterDeactivationHandlers_() {
      this.adapter_.deregisterInteractionHandler('keyup', this.deactivateHandler_);
      POINTER_DEACTIVATION_EVENT_TYPES.forEach((type) => {
        this.adapter_.deregisterDocumentInteractionHandler(type, this.deactivateHandler_);
      });
    }

    /** @private */
    removeCssVars_() {
      const {strings: strings$$1} = MDCRippleFoundation;
      Object.keys(strings$$1).forEach((k) => {
        if (k.indexOf('VAR_') === 0) {
          this.adapter_.updateCssVariable(strings$$1[k], null);
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    activate_(e) {
      if (this.adapter_.isSurfaceDisabled()) {
        return;
      }

      const activationState = this.activationState_;
      if (activationState.isActivated) {
        return;
      }

      // Avoid reacting to follow-on events fired by touch device after an already-processed user interaction
      const previousActivationEvent = this.previousActivationEvent_;
      const isSameInteraction = previousActivationEvent && e && previousActivationEvent.type !== e.type;
      if (isSameInteraction) {
        return;
      }

      activationState.isActivated = true;
      activationState.isProgrammatic = e === null;
      activationState.activationEvent = e;
      activationState.wasActivatedByPointer = activationState.isProgrammatic ? false : (
        e.type === 'mousedown' || e.type === 'touchstart' || e.type === 'pointerdown'
      );

      const hasActivatedChild =
        e && activatedTargets.length > 0 && activatedTargets.some((target) => this.adapter_.containsEventTarget(target));
      if (hasActivatedChild) {
        // Immediately reset activation state, while preserving logic that prevents touch follow-on events
        this.resetActivationState_();
        return;
      }

      if (e) {
        activatedTargets.push(/** @type {!EventTarget} */ (e.target));
        this.registerDeactivationHandlers_(e);
      }

      activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
      if (activationState.wasElementMadeActive) {
        this.animateActivation_();
      }

      requestAnimationFrame(() => {
        // Reset array on next frame after the current event has had a chance to bubble to prevent ancestor ripples
        activatedTargets = [];

        if (!activationState.wasElementMadeActive && (e.key === ' ' || e.keyCode === 32)) {
          // If space was pressed, try again within an rAF call to detect :active, because different UAs report
          // active states inconsistently when they're called within event handling code:
          // - https://bugs.chromium.org/p/chromium/issues/detail?id=635971
          // - https://bugzilla.mozilla.org/show_bug.cgi?id=1293741
          // We try first outside rAF to support Edge, which does not exhibit this problem, but will crash if a CSS
          // variable is set within a rAF callback for a submit button interaction (#2241).
          activationState.wasElementMadeActive = this.checkElementMadeActive_(e);
          if (activationState.wasElementMadeActive) {
            this.animateActivation_();
          }
        }

        if (!activationState.wasElementMadeActive) {
          // Reset activation state immediately if element was not made active.
          this.activationState_ = this.defaultActivationState_();
        }
      });
    }

    /**
     * @param {?Event} e
     * @private
     */
    checkElementMadeActive_(e) {
      return (e && e.type === 'keydown') ? this.adapter_.isSurfaceActive() : true;
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    activate(event = null) {
      this.activate_(event);
    }

    /** @private */
    animateActivation_() {
      const {VAR_FG_TRANSLATE_START, VAR_FG_TRANSLATE_END} = MDCRippleFoundation.strings;
      const {FG_DEACTIVATION, FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      const {DEACTIVATION_TIMEOUT_MS} = MDCRippleFoundation.numbers;

      this.layoutInternal_();

      let translateStart = '';
      let translateEnd = '';

      if (!this.adapter_.isUnbounded()) {
        const {startPoint, endPoint} = this.getFgTranslationCoordinates_();
        translateStart = `${startPoint.x}px, ${startPoint.y}px`;
        translateEnd = `${endPoint.x}px, ${endPoint.y}px`;
      }

      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_START, translateStart);
      this.adapter_.updateCssVariable(VAR_FG_TRANSLATE_END, translateEnd);
      // Cancel any ongoing activation/deactivation animations
      clearTimeout(this.activationTimer_);
      clearTimeout(this.fgDeactivationRemovalTimer_);
      this.rmBoundedActivationClasses_();
      this.adapter_.removeClass(FG_DEACTIVATION);

      // Force layout in order to re-trigger the animation.
      this.adapter_.computeBoundingRect();
      this.adapter_.addClass(FG_ACTIVATION);
      this.activationTimer_ = setTimeout(() => this.activationTimerCallback_(), DEACTIVATION_TIMEOUT_MS);
    }

    /**
     * @private
     * @return {{startPoint: PointType, endPoint: PointType}}
     */
    getFgTranslationCoordinates_() {
      const {activationEvent, wasActivatedByPointer} = this.activationState_;

      let startPoint;
      if (wasActivatedByPointer) {
        startPoint = getNormalizedEventCoords(
          /** @type {!Event} */ (activationEvent),
          this.adapter_.getWindowPageOffset(), this.adapter_.computeBoundingRect()
        );
      } else {
        startPoint = {
          x: this.frame_.width / 2,
          y: this.frame_.height / 2,
        };
      }
      // Center the element around the start point.
      startPoint = {
        x: startPoint.x - (this.initialSize_ / 2),
        y: startPoint.y - (this.initialSize_ / 2),
      };

      const endPoint = {
        x: (this.frame_.width / 2) - (this.initialSize_ / 2),
        y: (this.frame_.height / 2) - (this.initialSize_ / 2),
      };

      return {startPoint, endPoint};
    }

    /** @private */
    runDeactivationUXLogicIfReady_() {
      // This method is called both when a pointing device is released, and when the activation animation ends.
      // The deactivation animation should only run after both of those occur.
      const {FG_DEACTIVATION} = MDCRippleFoundation.cssClasses;
      const {hasDeactivationUXRun, isActivated} = this.activationState_;
      const activationHasEnded = hasDeactivationUXRun || !isActivated;

      if (activationHasEnded && this.activationAnimationHasEnded_) {
        this.rmBoundedActivationClasses_();
        this.adapter_.addClass(FG_DEACTIVATION);
        this.fgDeactivationRemovalTimer_ = setTimeout(() => {
          this.adapter_.removeClass(FG_DEACTIVATION);
        }, numbers.FG_DEACTIVATION_MS);
      }
    }

    /** @private */
    rmBoundedActivationClasses_() {
      const {FG_ACTIVATION} = MDCRippleFoundation.cssClasses;
      this.adapter_.removeClass(FG_ACTIVATION);
      this.activationAnimationHasEnded_ = false;
      this.adapter_.computeBoundingRect();
    }

    resetActivationState_() {
      this.previousActivationEvent_ = this.activationState_.activationEvent;
      this.activationState_ = this.defaultActivationState_();
      // Touch devices may fire additional events for the same interaction within a short time.
      // Store the previous event until it's safe to assume that subsequent events are for new interactions.
      setTimeout(() => this.previousActivationEvent_ = null, MDCRippleFoundation.numbers.TAP_DELAY_MS);
    }

    /**
     * @param {?Event} e
     * @private
     */
    deactivate_(e) {
      const activationState = this.activationState_;
      // This can happen in scenarios such as when you have a keyup event that blurs the element.
      if (!activationState.isActivated) {
        return;
      }

      const state = /** @type {!ActivationStateType} */ (Object.assign({}, activationState));

      if (activationState.isProgrammatic) {
        const evtObject = null;
        requestAnimationFrame(() => this.animateDeactivation_(evtObject, state));
        this.resetActivationState_();
      } else {
        this.deregisterDeactivationHandlers_();
        requestAnimationFrame(() => {
          this.activationState_.hasDeactivationUXRun = true;
          this.animateDeactivation_(e, state);
          this.resetActivationState_();
        });
      }
    }

    /**
     * @param {?Event=} event Optional event containing position information.
     */
    deactivate(event = null) {
      this.deactivate_(event);
    }

    /**
     * @param {Event} e
     * @param {!ActivationStateType} options
     * @private
     */
    animateDeactivation_(e, {wasActivatedByPointer, wasElementMadeActive}) {
      if (wasActivatedByPointer || wasElementMadeActive) {
        this.runDeactivationUXLogicIfReady_();
      }
    }

    layout() {
      if (this.layoutFrame_) {
        cancelAnimationFrame(this.layoutFrame_);
      }
      this.layoutFrame_ = requestAnimationFrame(() => {
        this.layoutInternal_();
        this.layoutFrame_ = 0;
      });
    }

    /** @private */
    layoutInternal_() {
      this.frame_ = this.adapter_.computeBoundingRect();
      const maxDim = Math.max(this.frame_.height, this.frame_.width);

      // Surface diameter is treated differently for unbounded vs. bounded ripples.
      // Unbounded ripple diameter is calculated smaller since the surface is expected to already be padded appropriately
      // to extend the hitbox, and the ripple is expected to meet the edges of the padded hitbox (which is typically
      // square). Bounded ripples, on the other hand, are fully expected to expand beyond the surface's longest diameter
      // (calculated based on the diagonal plus a constant padding), and are clipped at the surface's border via
      // `overflow: hidden`.
      const getBoundedRadius = () => {
        const hypotenuse = Math.sqrt(Math.pow(this.frame_.width, 2) + Math.pow(this.frame_.height, 2));
        return hypotenuse + MDCRippleFoundation.numbers.PADDING;
      };

      this.maxRadius_ = this.adapter_.isUnbounded() ? maxDim : getBoundedRadius();

      // Ripple is sized as a fraction of the largest dimension of the surface, then scales up using a CSS scale transform
      this.initialSize_ = maxDim * MDCRippleFoundation.numbers.INITIAL_ORIGIN_SCALE;
      this.fgScale_ = this.maxRadius_ / this.initialSize_;

      this.updateLayoutCssVars_();
    }

    /** @private */
    updateLayoutCssVars_() {
      const {
        VAR_FG_SIZE, VAR_LEFT, VAR_TOP, VAR_FG_SCALE,
      } = MDCRippleFoundation.strings;

      this.adapter_.updateCssVariable(VAR_FG_SIZE, `${this.initialSize_}px`);
      this.adapter_.updateCssVariable(VAR_FG_SCALE, this.fgScale_);

      if (this.adapter_.isUnbounded()) {
        this.unboundedCoords_ = {
          left: Math.round((this.frame_.width / 2) - (this.initialSize_ / 2)),
          top: Math.round((this.frame_.height / 2) - (this.initialSize_ / 2)),
        };

        this.adapter_.updateCssVariable(VAR_LEFT, `${this.unboundedCoords_.left}px`);
        this.adapter_.updateCssVariable(VAR_TOP, `${this.unboundedCoords_.top}px`);
      }
    }

    /** @param {boolean} unbounded */
    setUnbounded(unbounded) {
      const {UNBOUNDED} = MDCRippleFoundation.cssClasses;
      if (unbounded) {
        this.adapter_.addClass(UNBOUNDED);
      } else {
        this.adapter_.removeClass(UNBOUNDED);
      }
    }

    handleFocus() {
      requestAnimationFrame(() =>
        this.adapter_.addClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }

    handleBlur() {
      requestAnimationFrame(() =>
        this.adapter_.removeClass(MDCRippleFoundation.cssClasses.BG_FOCUSED));
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends MDCComponent<!MDCRippleFoundation>
   */
  class MDCRipple extends MDCComponent {
    /** @param {...?} args */
    constructor(...args) {
      super(...args);

      /** @type {boolean} */
      this.disabled = false;

      /** @private {boolean} */
      this.unbounded_;
    }

    /**
     * @param {!Element} root
     * @param {{isUnbounded: (boolean|undefined)}=} options
     * @return {!MDCRipple}
     */
    static attachTo(root, {isUnbounded = undefined} = {}) {
      const ripple = new MDCRipple(root);
      // Only override unbounded behavior if option is explicitly specified
      if (isUnbounded !== undefined) {
        ripple.unbounded = /** @type {boolean} */ (isUnbounded);
      }
      return ripple;
    }

    /**
     * @param {!RippleCapableSurface} instance
     * @return {!MDCRippleAdapter}
     */
    static createAdapter(instance) {
      const MATCHES = getMatchesProperty(HTMLElement.prototype);

      return {
        browserSupportsCssVars: () => supportsCssVariables(window),
        isUnbounded: () => instance.unbounded,
        isSurfaceActive: () => instance.root_[MATCHES](':active'),
        isSurfaceDisabled: () => instance.disabled,
        addClass: (className) => instance.root_.classList.add(className),
        removeClass: (className) => instance.root_.classList.remove(className),
        containsEventTarget: (target) => instance.root_.contains(target),
        registerInteractionHandler: (evtType, handler) =>
          instance.root_.addEventListener(evtType, handler, applyPassive()),
        deregisterInteractionHandler: (evtType, handler) =>
          instance.root_.removeEventListener(evtType, handler, applyPassive()),
        registerDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.addEventListener(evtType, handler, applyPassive()),
        deregisterDocumentInteractionHandler: (evtType, handler) =>
          document.documentElement.removeEventListener(evtType, handler, applyPassive()),
        registerResizeHandler: (handler) => window.addEventListener('resize', handler),
        deregisterResizeHandler: (handler) => window.removeEventListener('resize', handler),
        updateCssVariable: (varName, value) => instance.root_.style.setProperty(varName, value),
        computeBoundingRect: () => instance.root_.getBoundingClientRect(),
        getWindowPageOffset: () => ({x: window.pageXOffset, y: window.pageYOffset}),
      };
    }

    /** @return {boolean} */
    get unbounded() {
      return this.unbounded_;
    }

    /** @param {boolean} unbounded */
    set unbounded(unbounded) {
      this.unbounded_ = Boolean(unbounded);
      this.setUnbounded_();
    }

    /**
     * Closure Compiler throws an access control error when directly accessing a
     * protected or private property inside a getter/setter, like unbounded above.
     * By accessing the protected property inside a method, we solve that problem.
     * That's why this function exists.
     * @private
     */
    setUnbounded_() {
      this.foundation_.setUnbounded(this.unbounded_);
    }

    activate() {
      this.foundation_.activate();
    }

    deactivate() {
      this.foundation_.deactivate();
    }

    layout() {
      this.foundation_.layout();
    }

    /**
     * @return {!MDCRippleFoundation}
     * @override
     */
    getDefaultFoundation() {
      return new MDCRippleFoundation(MDCRipple.createAdapter(this));
    }

    /** @override */
    initialSyncWithDOM() {
      this.unbounded = 'mdcRippleIsUnbounded' in this.root_.dataset;
    }
  }

  /**
   * See Material Design spec for more details on when to use ripples.
   * https://material.io/guidelines/motion/choreography.html#choreography-creation
   * @record
   */
  class RippleCapableSurface {}

  /** @protected {!Element} */
  RippleCapableSurface.prototype.root_;

  /**
   * Whether or not the ripple bleeds out of the bounds of the element.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.unbounded;

  /**
   * Whether or not the ripple is attached to a disabled component.
   * @type {boolean|undefined}
   */
  RippleCapableSurface.prototype.disabled;

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const strings$1 = {
    ARIA_CONTROLS: 'aria-controls',
    INPUT_SELECTOR: '.mdc-text-field__input',
    LABEL_SELECTOR: '.mdc-floating-label',
    ICON_SELECTOR: '.mdc-text-field__icon',
    OUTLINE_SELECTOR: '.mdc-notched-outline',
    LINE_RIPPLE_SELECTOR: '.mdc-line-ripple',
  };

  /** @enum {string} */
  const cssClasses$1 = {
    ROOT: 'mdc-text-field',
    UPGRADED: 'mdc-text-field--upgraded',
    DISABLED: 'mdc-text-field--disabled',
    DENSE: 'mdc-text-field--dense',
    FOCUSED: 'mdc-text-field--focused',
    INVALID: 'mdc-text-field--invalid',
    BOX: 'mdc-text-field--box',
    OUTLINED: 'mdc-text-field--outlined',
  };

  /** @enum {number} */
  const numbers$1 = {
    LABEL_SCALE: 0.75,
    DENSE_LABEL_SCALE: 0.923,
  };

  // whitelist based off of https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/HTML5/Constraint_validation
  // under section: `Validation-related attributes`
  const VALIDATION_ATTR_WHITELIST = [
    'pattern', 'min', 'max', 'required', 'step', 'minlength', 'maxlength',
  ];

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const strings$2 = {
    ARIA_HIDDEN: 'aria-hidden',
    ROLE: 'role',
  };

  /** @enum {string} */
  const cssClasses$2 = {
    HELPER_TEXT_PERSISTENT: 'mdc-text-field-helper-text--persistent',
    HELPER_TEXT_VALIDATION_MSG: 'mdc-text-field-helper-text--validation-msg',
  };

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  /**
   * @extends {MDCFoundation<!MDCTextFieldHelperTextAdapter>}
   * @final
   */
  class MDCTextFieldHelperTextFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$2;
    }

    /** @return enum {string} */
    static get strings() {
      return strings$2;
    }

    /**
     * {@see MDCTextFieldHelperTextAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCTextFieldHelperTextAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTextFieldHelperTextAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        hasClass: () => {},
        setAttr: () => {},
        removeAttr: () => {},
        setContent: () => {},
      });
    }

    /**
     * @param {!MDCTextFieldHelperTextAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCTextFieldHelperTextFoundation.defaultAdapter, adapter));
    }

    /**
     * Sets the content of the helper text field.
     * @param {string} content
     */
    setContent(content) {
      this.adapter_.setContent(content);
    }

    /** @param {boolean} isPersistent Sets the persistency of the helper text. */
    setPersistent(isPersistent) {
      if (isPersistent) {
        this.adapter_.addClass(cssClasses$2.HELPER_TEXT_PERSISTENT);
      } else {
        this.adapter_.removeClass(cssClasses$2.HELPER_TEXT_PERSISTENT);
      }
    }

    /**
     * @param {boolean} isValidation True to make the helper text act as an
     *   error validation message.
     */
    setValidation(isValidation) {
      if (isValidation) {
        this.adapter_.addClass(cssClasses$2.HELPER_TEXT_VALIDATION_MSG);
      } else {
        this.adapter_.removeClass(cssClasses$2.HELPER_TEXT_VALIDATION_MSG);
      }
    }

    /** Makes the helper text visible to the screen reader. */
    showToScreenReader() {
      this.adapter_.removeAttr(strings$2.ARIA_HIDDEN);
    }

    /**
     * Sets the validity of the helper text based on the input validity.
     * @param {boolean} inputIsValid
     */
    setValidity(inputIsValid) {
      const helperTextIsPersistent = this.adapter_.hasClass(cssClasses$2.HELPER_TEXT_PERSISTENT);
      const helperTextIsValidationMsg = this.adapter_.hasClass(cssClasses$2.HELPER_TEXT_VALIDATION_MSG);
      const validationMsgNeedsDisplay = helperTextIsValidationMsg && !inputIsValid;

      if (validationMsgNeedsDisplay) {
        this.adapter_.setAttr(strings$2.ROLE, 'alert');
      } else {
        this.adapter_.removeAttr(strings$2.ROLE);
      }

      if (!helperTextIsPersistent && !validationMsgNeedsDisplay) {
        this.hide_();
      }
    }

    /**
     * Hides the help text from screen readers.
     * @private
     */
    hide_() {
      this.adapter_.setAttr(strings$2.ARIA_HIDDEN, 'true');
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const strings$3 = {
    ICON_EVENT: 'MDCTextField:icon',
    ICON_ROLE: 'button',
  };

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  /**
   * @extends {MDCFoundation<!MDCTextFieldIconAdapter>}
   * @final
   */
  class MDCTextFieldIconFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get strings() {
      return strings$3;
    }

    /**
     * {@see MDCTextFieldIconAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCTextFieldIconAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTextFieldIconAdapter} */ ({
        getAttr: () => {},
        setAttr: () => {},
        removeAttr: () => {},
        setContent: () => {},
        registerInteractionHandler: () => {},
        deregisterInteractionHandler: () => {},
        notifyIconAction: () => {},
      });
    }

    /**
     * @param {!MDCTextFieldIconAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCTextFieldIconFoundation.defaultAdapter, adapter));

      /** @private {string?} */
      this.savedTabIndex_ = null;

      /** @private {function(!Event): undefined} */
      this.interactionHandler_ = (evt) => this.handleInteraction(evt);
    }

    init() {
      this.savedTabIndex_ = this.adapter_.getAttr('tabindex');

      ['click', 'keydown'].forEach((evtType) => {
        this.adapter_.registerInteractionHandler(evtType, this.interactionHandler_);
      });
    }

    destroy() {
      ['click', 'keydown'].forEach((evtType) => {
        this.adapter_.deregisterInteractionHandler(evtType, this.interactionHandler_);
      });
    }

    /** @param {boolean} disabled */
    setDisabled(disabled) {
      if (!this.savedTabIndex_) {
        return;
      }

      if (disabled) {
        this.adapter_.setAttr('tabindex', '-1');
        this.adapter_.removeAttr('role');
      } else {
        this.adapter_.setAttr('tabindex', this.savedTabIndex_);
        this.adapter_.setAttr('role', strings$3.ICON_ROLE);
      }
    }

    /** @param {string} label */
    setAriaLabel(label) {
      this.adapter_.setAttr('aria-label', label);
    }

    /** @param {string} content */
    setContent(content) {
      this.adapter_.setContent(content);
    }

    /**
     * Handles an interaction event
     * @param {!Event} evt
     */
    handleInteraction(evt) {
      if (evt.type === 'click' || evt.key === 'Enter' || evt.keyCode === 13) {
        this.adapter_.notifyIconAction();
      }
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  /**
   * @extends {MDCFoundation<!MDCTextFieldAdapter>}
   * @final
   */
  class MDCTextFieldFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$1;
    }

    /** @return enum {string} */
    static get strings() {
      return strings$1;
    }

    /** @return enum {string} */
    static get numbers() {
      return numbers$1;
    }

    /** @return {boolean} */
    get shouldShake() {
      return !this.isValid() && !this.isFocused_;
    }

    /** @return {boolean} */
    get shouldFloat() {
      return this.isFocused_ || !!this.getValue() || this.isBadInput_();
    }

    /**
     * {@see MDCTextFieldAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCTextFieldAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCTextFieldAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        hasClass: () => {},
        registerTextFieldInteractionHandler: () => {},
        deregisterTextFieldInteractionHandler: () => {},
        registerInputInteractionHandler: () => {},
        deregisterInputInteractionHandler: () => {},
        registerValidationAttributeChangeHandler: () => {},
        deregisterValidationAttributeChangeHandler: () => {},
        getNativeInput: () => {},
        isFocused: () => {},
        isRtl: () => {},
        activateLineRipple: () => {},
        deactivateLineRipple: () => {},
        setLineRippleTransformOrigin: () => {},
        shakeLabel: () => {},
        floatLabel: () => {},
        hasLabel: () => {},
        getLabelWidth: () => {},
        hasOutline: () => {},
        notchOutline: () => {},
        closeOutline: () => {},
      });
    }

    /**
     * @param {!MDCTextFieldAdapter} adapter
     * @param {!FoundationMapType=} foundationMap Map from subcomponent names to their subfoundations.
     */
    constructor(adapter, foundationMap = /** @type {!FoundationMapType} */ ({})) {
      super(Object.assign(MDCTextFieldFoundation.defaultAdapter, adapter));

      /** @type {!MDCTextFieldHelperTextFoundation|undefined} */
      this.helperText_ = foundationMap.helperText;
      /** @type {!MDCTextFieldIconFoundation|undefined} */
      this.icon_ = foundationMap.icon;

      /** @private {boolean} */
      this.isFocused_ = false;
      /** @private {boolean} */
      this.receivedUserInput_ = false;
      /** @private {boolean} */
      this.useCustomValidityChecking_ = false;
      /** @private {boolean} */
      this.isValid_ = true;
      /** @private {function(): undefined} */
      this.inputFocusHandler_ = () => this.activateFocus();
      /** @private {function(): undefined} */
      this.inputBlurHandler_ = () => this.deactivateFocus();
      /** @private {function(): undefined} */
      this.inputInputHandler_ = () => this.autoCompleteFocus();
      /** @private {function(!Event): undefined} */
      this.setPointerXOffset_ = (evt) => this.setTransformOrigin(evt);
      /** @private {function(!Event): undefined} */
      this.textFieldInteractionHandler_ = () => this.handleTextFieldInteraction();
      /** @private {function(!Array): undefined} */
      this.validationAttributeChangeHandler_ = (attributesList) => this.handleValidationAttributeChange(attributesList);

      /** @private {!MutationObserver} */
      this.validationObserver_;
    }

    init() {
      this.adapter_.addClass(MDCTextFieldFoundation.cssClasses.UPGRADED);
      // Ensure label does not collide with any pre-filled value.
      if (this.adapter_.hasLabel() && (this.getValue() || this.isBadInput_())) {
        this.adapter_.floatLabel(this.shouldFloat);
        this.notchOutline(this.shouldFloat);
      }

      if (this.adapter_.isFocused()) {
        this.inputFocusHandler_();
      }

      this.adapter_.registerInputInteractionHandler('focus', this.inputFocusHandler_);
      this.adapter_.registerInputInteractionHandler('blur', this.inputBlurHandler_);
      this.adapter_.registerInputInteractionHandler('input', this.inputInputHandler_);
      ['mousedown', 'touchstart'].forEach((evtType) => {
        this.adapter_.registerInputInteractionHandler(evtType, this.setPointerXOffset_);
      });
      ['click', 'keydown'].forEach((evtType) => {
        this.adapter_.registerTextFieldInteractionHandler(evtType, this.textFieldInteractionHandler_);
      });
      this.validationObserver_ =
          this.adapter_.registerValidationAttributeChangeHandler(this.validationAttributeChangeHandler_);
    }

    destroy() {
      this.adapter_.removeClass(MDCTextFieldFoundation.cssClasses.UPGRADED);
      this.adapter_.deregisterInputInteractionHandler('focus', this.inputFocusHandler_);
      this.adapter_.deregisterInputInteractionHandler('blur', this.inputBlurHandler_);
      this.adapter_.deregisterInputInteractionHandler('input', this.inputInputHandler_);
      ['mousedown', 'touchstart'].forEach((evtType) => {
        this.adapter_.deregisterInputInteractionHandler(evtType, this.setPointerXOffset_);
      });
      ['click', 'keydown'].forEach((evtType) => {
        this.adapter_.deregisterTextFieldInteractionHandler(evtType, this.textFieldInteractionHandler_);
      });
      this.adapter_.deregisterValidationAttributeChangeHandler(this.validationObserver_);
    }

    /**
     * Handles user interactions with the Text Field.
     */
    handleTextFieldInteraction() {
      if (this.adapter_.getNativeInput().disabled) {
        return;
      }
      this.receivedUserInput_ = true;
    }

    /**
     * Handles validation attribute changes
     * @param {!Array<string>} attributesList
     */
    handleValidationAttributeChange(attributesList) {
      attributesList.some((attributeName) => {
        if (VALIDATION_ATTR_WHITELIST.indexOf(attributeName) > -1) {
          this.styleValidity_(true);
          return true;
        }
      });
    }

    /**
     * Opens/closes the notched outline.
     * @param {boolean} openNotch
     */
    notchOutline(openNotch) {
      if (!this.adapter_.hasOutline() || !this.adapter_.hasLabel()) {
        return;
      }

      if (openNotch) {
        const isDense = this.adapter_.hasClass(cssClasses$1.DENSE);
        const labelScale = isDense ? numbers$1.DENSE_LABEL_SCALE : numbers$1.LABEL_SCALE;
        const labelWidth = this.adapter_.getLabelWidth() * labelScale;
        const isRtl = this.adapter_.isRtl();
        this.adapter_.notchOutline(labelWidth, isRtl);
      } else {
        this.adapter_.closeOutline();
      }
    }

    /**
     * Activates the text field focus state.
     */
    activateFocus() {
      this.isFocused_ = true;
      this.styleFocused_(this.isFocused_);
      this.adapter_.activateLineRipple();
      this.notchOutline(this.shouldFloat);
      if (this.adapter_.hasLabel()) {
        this.adapter_.shakeLabel(this.shouldShake);
        this.adapter_.floatLabel(this.shouldFloat);
      }
      if (this.helperText_) {
        this.helperText_.showToScreenReader();
      }
    }

    /**
     * Sets the line ripple's transform origin, so that the line ripple activate
     * animation will animate out from the user's click location.
     * @param {!Event} evt
     */
    setTransformOrigin(evt) {
      const targetClientRect = evt.target.getBoundingClientRect();
      const evtCoords = {x: evt.clientX, y: evt.clientY};
      const normalizedX = evtCoords.x - targetClientRect.left;
      this.adapter_.setLineRippleTransformOrigin(normalizedX);
    }

    /**
     * Activates the Text Field's focus state in cases when the input value
     * changes without user input (e.g. programatically).
     */
    autoCompleteFocus() {
      if (!this.receivedUserInput_) {
        this.activateFocus();
      }
    }

    /**
     * Deactivates the Text Field's focus state.
     */
    deactivateFocus() {
      this.isFocused_ = false;
      this.adapter_.deactivateLineRipple();
      const input = this.getNativeInput_();
      const shouldRemoveLabelFloat = !input.value && !this.isBadInput_();
      const isValid = this.isValid();
      this.styleValidity_(isValid);
      this.styleFocused_(this.isFocused_);
      if (this.adapter_.hasLabel()) {
        this.adapter_.shakeLabel(this.shouldShake);
        this.adapter_.floatLabel(this.shouldFloat);
        this.notchOutline(this.shouldFloat);
      }
      if (shouldRemoveLabelFloat) {
        this.receivedUserInput_ = false;
      }
    }

    /**
     * @return {string} The value of the input Element.
     */
    getValue() {
      return this.getNativeInput_().value;
    }

    /**
     * @param {string} value The value to set on the input Element.
     */
    setValue(value) {
      this.getNativeInput_().value = value;
      const isValid = this.isValid();
      this.styleValidity_(isValid);
      if (this.adapter_.hasLabel()) {
        this.adapter_.shakeLabel(this.shouldShake);
        this.adapter_.floatLabel(this.shouldFloat);
        this.notchOutline(this.shouldFloat);
      }
    }

    /**
     * @return {boolean} If a custom validity is set, returns that value.
     *     Otherwise, returns the result of native validity checks.
     */
    isValid() {
      return this.useCustomValidityChecking_
        ? this.isValid_ : this.isNativeInputValid_();
    }

    /**
     * @param {boolean} isValid Sets the validity state of the Text Field.
     */
    setValid(isValid) {
      this.useCustomValidityChecking_ = true;
      this.isValid_ = isValid;
      // Retrieve from the getter to ensure correct logic is applied.
      isValid = this.isValid();
      this.styleValidity_(isValid);
      if (this.adapter_.hasLabel()) {
        this.adapter_.shakeLabel(this.shouldShake);
      }
    }

    /**
     * @return {boolean} True if the Text Field is disabled.
     */
    isDisabled() {
      return this.getNativeInput_().disabled;
    }

    /**
     * @param {boolean} disabled Sets the text-field disabled or enabled.
     */
    setDisabled(disabled) {
      this.getNativeInput_().disabled = disabled;
      this.styleDisabled_(disabled);
    }

    /**
     * @param {string} content Sets the content of the helper text.
     */
    setHelperTextContent(content) {
      if (this.helperText_) {
        this.helperText_.setContent(content);
      }
    }

    /**
     * Sets the aria label of the icon.
     * @param {string} label
     */
    setIconAriaLabel(label) {
      if (this.icon_) {
        this.icon_.setAriaLabel(label);
      }
    }

    /**
     * Sets the text content of the icon.
     * @param {string} content
     */
    setIconContent(content) {
      if (this.icon_) {
        this.icon_.setContent(content);
      }
    }

    /**
     * @return {boolean} True if the Text Field input fails in converting the
     *     user-supplied value.
     * @private
     */
    isBadInput_() {
      return this.getNativeInput_().validity.badInput;
    }

    /**
     * @return {boolean} The result of native validity checking
     *     (ValidityState.valid).
     */
    isNativeInputValid_() {
      return this.getNativeInput_().validity.valid;
    }

    /**
     * Styles the component based on the validity state.
     * @param {boolean} isValid
     * @private
     */
    styleValidity_(isValid) {
      const {INVALID} = MDCTextFieldFoundation.cssClasses;
      if (isValid) {
        this.adapter_.removeClass(INVALID);
      } else {
        this.adapter_.addClass(INVALID);
      }
      if (this.helperText_) {
        this.helperText_.setValidity(isValid);
      }
    }

    /**
     * Styles the component based on the focused state.
     * @param {boolean} isFocused
     * @private
     */
    styleFocused_(isFocused) {
      const {FOCUSED} = MDCTextFieldFoundation.cssClasses;
      if (isFocused) {
        this.adapter_.addClass(FOCUSED);
      } else {
        this.adapter_.removeClass(FOCUSED);
      }
    }

    /**
     * Styles the component based on the disabled state.
     * @param {boolean} isDisabled
     * @private
     */
    styleDisabled_(isDisabled) {
      const {DISABLED, INVALID} = MDCTextFieldFoundation.cssClasses;
      if (isDisabled) {
        this.adapter_.addClass(DISABLED);
        this.adapter_.removeClass(INVALID);
      } else {
        this.adapter_.removeClass(DISABLED);
      }
      if (this.icon_) {
        this.icon_.setDisabled(isDisabled);
      }
    }

    /**
     * @return {!Element|!NativeInputType} The native text input from the
     * host environment, or a dummy if none exists.
     * @private
     */
    getNativeInput_() {
      return this.adapter_.getNativeInput() ||
      /** @type {!NativeInputType} */ ({
        value: '',
        disabled: false,
        validity: {
          badInput: false,
          valid: true,
        },
      });
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const cssClasses$3 = {
    LINE_RIPPLE_ACTIVE: 'mdc-line-ripple--active',
    LINE_RIPPLE_DEACTIVATING: 'mdc-line-ripple--deactivating',
  };

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */


  /**
   * @extends {MDCFoundation<!MDCLineRippleAdapter>}
   * @final
   */
  class MDCLineRippleFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$3;
    }

    /**
     * {@see MDCLineRippleAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCLineRippleAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCLineRippleAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        hasClass: () => {},
        setStyle: () => {},
        registerEventHandler: () => {},
        deregisterEventHandler: () => {},
      });
    }

    /**
     * @param {!MDCLineRippleAdapter=} adapter
     */
    constructor(adapter = /** @type {!MDCLineRippleAdapter} */ ({})) {
      super(Object.assign(MDCLineRippleFoundation.defaultAdapter, adapter));

      /** @private {function(!Event): undefined} */
      this.transitionEndHandler_ = (evt) => this.handleTransitionEnd(evt);
    }

    init() {
      this.adapter_.registerEventHandler('transitionend', this.transitionEndHandler_);
    }

    destroy() {
      this.adapter_.deregisterEventHandler('transitionend', this.transitionEndHandler_);
    }

    /**
     * Activates the line ripple
     */
    activate() {
      this.adapter_.removeClass(cssClasses$3.LINE_RIPPLE_DEACTIVATING);
      this.adapter_.addClass(cssClasses$3.LINE_RIPPLE_ACTIVE);
    }

    /**
     * Sets the center of the ripple animation to the given X coordinate.
     * @param {number} xCoordinate
     */
    setRippleCenter(xCoordinate) {
      this.adapter_.setStyle('transform-origin', `${xCoordinate}px center`);
    }

    /**
     * Deactivates the line ripple
     */
    deactivate() {
      this.adapter_.addClass(cssClasses$3.LINE_RIPPLE_DEACTIVATING);
    }

    /**
     * Handles a transition end event
     * @param {!Event} evt
     */
    handleTransitionEnd(evt) {
      // Wait for the line ripple to be either transparent or opaque
      // before emitting the animation end event
      const isDeactivating = this.adapter_.hasClass(cssClasses$3.LINE_RIPPLE_DEACTIVATING);

      if (evt.propertyName === 'opacity') {
        if (isDeactivating) {
          this.adapter_.removeClass(cssClasses$3.LINE_RIPPLE_ACTIVE);
          this.adapter_.removeClass(cssClasses$3.LINE_RIPPLE_DEACTIVATING);
        }
      }
    }
  }

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCLineRippleFoundation>}
   * @final
   */
  class MDCLineRipple extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCLineRipple}
     */
    static attachTo(root) {
      return new MDCLineRipple(root);
    }

    /**
     * Activates the line ripple
     */
    activate() {
      this.foundation_.activate();
    }

    /**
     * Deactivates the line ripple
     */
    deactivate() {
      this.foundation_.deactivate();
    }

    /**
     * Sets the transform origin given a user's click location. The `rippleCenter` is the
     * x-coordinate of the middle of the ripple.
     * @param {number} xCoordinate
     */
    setRippleCenter(xCoordinate) {
      this.foundation_.setRippleCenter(xCoordinate);
    }

    /**
     * @return {!MDCLineRippleFoundation}
     */
    getDefaultFoundation() {
      return new MDCLineRippleFoundation(/** @type {!MDCLineRippleAdapter} */ (Object.assign({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        hasClass: (className) => this.root_.classList.contains(className),
        setStyle: (propertyName, value) => this.root_.style[propertyName] = value,
        registerEventHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterEventHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
      })));
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCTextFieldHelperTextFoundation>}
   * @final
   */
  class MDCTextFieldHelperText extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCTextFieldHelperText}
     */
    static attachTo(root) {
      return new MDCTextFieldHelperText(root);
    }

    /**
     * @return {!MDCTextFieldHelperTextFoundation}
     */
    get foundation() {
      return this.foundation_;
    }

    /**
     * @return {!MDCTextFieldHelperTextFoundation}
     */
    getDefaultFoundation() {
      return new MDCTextFieldHelperTextFoundation(/** @type {!MDCTextFieldHelperTextAdapter} */ (Object.assign({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        hasClass: (className) => this.root_.classList.contains(className),
        setAttr: (attr, value) => this.root_.setAttribute(attr, value),
        removeAttr: (attr) => this.root_.removeAttribute(attr),
        setContent: (content) => {
          this.root_.textContent = content;
        },
      })));
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCTextFieldIconFoundation>}
   * @final
   */
  class MDCTextFieldIcon extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCTextFieldIcon}
     */
    static attachTo(root) {
      return new MDCTextFieldIcon(root);
    }

    /**
     * @return {!MDCTextFieldIconFoundation}
     */
    get foundation() {
      return this.foundation_;
    }

    /**
     * @return {!MDCTextFieldIconFoundation}
     */
    getDefaultFoundation() {
      return new MDCTextFieldIconFoundation(/** @type {!MDCTextFieldIconAdapter} */ (Object.assign({
        getAttr: (attr) => this.root_.getAttribute(attr),
        setAttr: (attr, value) => this.root_.setAttribute(attr, value),
        removeAttr: (attr) => this.root_.removeAttribute(attr),
        setContent: (content) => {
          this.root_.textContent = content;
        },
        registerInteractionHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterInteractionHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
        notifyIconAction: () => this.emit(
          MDCTextFieldIconFoundation.strings.ICON_EVENT, {} /* evtData */, true /* shouldBubble */),
      })));
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const cssClasses$4 = {
    LABEL_FLOAT_ABOVE: 'mdc-floating-label--float-above',
    LABEL_SHAKE: 'mdc-floating-label--shake',
  };

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCFoundation<!MDCFloatingLabelAdapter>}
   * @final
   */
  class MDCFloatingLabelFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$4;
    }

    /**
     * {@see MDCFloatingLabelAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCFloatingLabelAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCFloatingLabelAdapter} */ ({
        addClass: () => {},
        removeClass: () => {},
        getWidth: () => {},
        registerInteractionHandler: () => {},
        deregisterInteractionHandler: () => {},
      });
    }

    /**
     * @param {!MDCFloatingLabelAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCFloatingLabelFoundation.defaultAdapter, adapter));

      /** @private {function(!Event): undefined} */
      this.shakeAnimationEndHandler_ = () => this.handleShakeAnimationEnd_();
    }

    init() {
      this.adapter_.registerInteractionHandler('animationend', this.shakeAnimationEndHandler_);
    }

    destroy() {
      this.adapter_.deregisterInteractionHandler('animationend', this.shakeAnimationEndHandler_);
    }

    /**
     * Returns the width of the label element.
     * @return {number}
     */
    getWidth() {
      return this.adapter_.getWidth();
    }

    /**
     * Styles the label to produce the label shake for errors.
     * @param {boolean} shouldShake adds shake class if true,
     * otherwise removes shake class.
     */
    shake(shouldShake) {
      const {LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      if (shouldShake) {
        this.adapter_.addClass(LABEL_SHAKE);
      } else {
        this.adapter_.removeClass(LABEL_SHAKE);
      }
    }

    /**
     * Styles the label to float or dock.
     * @param {boolean} shouldFloat adds float class if true, otherwise remove
     * float and shake class to dock label.
     */
    float(shouldFloat) {
      const {LABEL_FLOAT_ABOVE, LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      if (shouldFloat) {
        this.adapter_.addClass(LABEL_FLOAT_ABOVE);
      } else {
        this.adapter_.removeClass(LABEL_FLOAT_ABOVE);
        this.adapter_.removeClass(LABEL_SHAKE);
      }
    }

    /**
     * Handles an interaction event on the root element.
     */
    handleShakeAnimationEnd_() {
      const {LABEL_SHAKE} = MDCFloatingLabelFoundation.cssClasses;
      this.adapter_.removeClass(LABEL_SHAKE);
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCFloatingLabelFoundation>}
   * @final
   */
  class MDCFloatingLabel extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCFloatingLabel}
     */
    static attachTo(root) {
      return new MDCFloatingLabel(root);
    }

    /**
     * Styles the label to produce the label shake for errors.
     * @param {boolean} shouldShake styles the label to shake by adding shake class
     * if true, otherwise will stop shaking by removing shake class.
     */
    shake(shouldShake) {
      this.foundation_.shake(shouldShake);
    }

    /**
     * Styles label to float/dock.
     * @param {boolean} shouldFloat styles the label to float by adding float class
     * if true, otherwise docks the label by removing the float class.
     */
    float(shouldFloat) {
      this.foundation_.float(shouldFloat);
    }

    /**
     * @return {number}
     */
    getWidth() {
      return this.foundation_.getWidth();
    }

    /**
     * @return {!MDCFloatingLabelFoundation}
     */
    getDefaultFoundation() {
      return new MDCFloatingLabelFoundation({
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        getWidth: () => this.root_.offsetWidth,
        registerInteractionHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
        deregisterInteractionHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
      });
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @license
   * Copyright 2018 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /** @enum {string} */
  const strings$4 = {
    PATH_SELECTOR: '.mdc-notched-outline__path',
    IDLE_OUTLINE_SELECTOR: '.mdc-notched-outline__idle',
  };

  /** @enum {string} */
  const cssClasses$5 = {
    OUTLINE_NOTCHED: 'mdc-notched-outline--notched',
  };

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCFoundation<!MDCNotchedOutlineAdapter>}
   * @final
   */
  class MDCNotchedOutlineFoundation extends MDCFoundation {
    /** @return enum {string} */
    static get strings() {
      return strings$4;
    }

    /** @return enum {string} */
    static get cssClasses() {
      return cssClasses$5;
    }

    /**
     * {@see MDCNotchedOutlineAdapter} for typing information on parameters and return
     * types.
     * @return {!MDCNotchedOutlineAdapter}
     */
    static get defaultAdapter() {
      return /** @type {!MDCNotchedOutlineAdapter} */ ({
        getWidth: () => {},
        getHeight: () => {},
        addClass: () => {},
        removeClass: () => {},
        setOutlinePathAttr: () => {},
        getIdleOutlineStyleValue: () => {},
      });
    }

    /**
     * @param {!MDCNotchedOutlineAdapter} adapter
     */
    constructor(adapter) {
      super(Object.assign(MDCNotchedOutlineFoundation.defaultAdapter, adapter));
    }

    /**
     * Adds the outline notched selector and updates the notch width
     * calculated based off of notchWidth and isRtl.
     * @param {number} notchWidth
     * @param {boolean=} isRtl
     */
    notch(notchWidth, isRtl = false) {
      const {OUTLINE_NOTCHED} = MDCNotchedOutlineFoundation.cssClasses;
      this.adapter_.addClass(OUTLINE_NOTCHED);
      this.updateSvgPath_(notchWidth, isRtl);
    }

    /**
     * Removes notched outline selector to close the notch in the outline.
     */
    closeNotch() {
      const {OUTLINE_NOTCHED} = MDCNotchedOutlineFoundation.cssClasses;
      this.adapter_.removeClass(OUTLINE_NOTCHED);
    }

    /**
     * Updates the SVG path of the focus outline element based on the notchWidth
     * and the RTL context.
     * @param {number} notchWidth
     * @param {boolean=} isRtl
     * @private
     */
    updateSvgPath_(notchWidth, isRtl) {
      // Fall back to reading a specific corner's style because Firefox doesn't report the style on border-radius.
      const radiusStyleValue = this.adapter_.getIdleOutlineStyleValue('border-radius') ||
          this.adapter_.getIdleOutlineStyleValue('border-top-left-radius');
      const radius = parseFloat(radiusStyleValue);
      const width = this.adapter_.getWidth();
      const height = this.adapter_.getHeight();
      const cornerWidth = radius + 1.2;
      const leadingStrokeLength = Math.abs(11 - cornerWidth);
      const paddedNotchWidth = notchWidth + 8;

      // The right, bottom, and left sides of the outline follow the same SVG path.
      const pathMiddle = 'a' + radius + ',' + radius + ' 0 0 1 ' + radius + ',' + radius
        + 'v' + (height - (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + -radius + ',' + radius
        + 'h' + (-width + (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + -radius + ',' + -radius
        + 'v' + (-height + (2 * cornerWidth))
        + 'a' + radius + ',' + radius + ' 0 0 1 ' + radius + ',' + -radius;

      let path;
      if (!isRtl) {
        path = 'M' + (cornerWidth + leadingStrokeLength + paddedNotchWidth) + ',' + 1
          + 'h' + (width - (2 * cornerWidth) - paddedNotchWidth - leadingStrokeLength)
          + pathMiddle
          + 'h' + leadingStrokeLength;
      } else {
        path = 'M' + (width - cornerWidth - leadingStrokeLength) + ',' + 1
          + 'h' + leadingStrokeLength
          + pathMiddle
          + 'h' + (width - (2 * cornerWidth) - paddedNotchWidth - leadingStrokeLength);
      }

      this.adapter_.setOutlinePathAttr(path);
    }
  }

  /**
   * @license
   * Copyright 2017 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */

  /**
   * @extends {MDCComponent<!MDCNotchedOutlineFoundation>}
   * @final
   */
  class MDCNotchedOutline extends MDCComponent {
    /**
     * @param {!Element} root
     * @return {!MDCNotchedOutline}
     */
    static attachTo(root) {
      return new MDCNotchedOutline(root);
    }

    /**
      * Updates outline selectors and SVG path to open notch.
      * @param {number} notchWidth The notch width in the outline.
      * @param {boolean=} isRtl Determines if outline is rtl. If rtl is true, notch
      * will be right justified in outline path, otherwise left justified.
      */
    notch(notchWidth, isRtl) {
      this.foundation_.notch(notchWidth, isRtl);
    }

    /**
     * Updates the outline selectors to close notch and return it to idle state.
     */
    closeNotch() {
      this.foundation_.closeNotch();
    }

    /**
     * @return {!MDCNotchedOutlineFoundation}
     */
    getDefaultFoundation() {
      return new MDCNotchedOutlineFoundation({
        getWidth: () => this.root_.offsetWidth,
        getHeight: () => this.root_.offsetHeight,
        addClass: (className) => this.root_.classList.add(className),
        removeClass: (className) => this.root_.classList.remove(className),
        setOutlinePathAttr: (value) => {
          const path = this.root_.querySelector(strings$4.PATH_SELECTOR);
          path.setAttribute('d', value);
        },
        getIdleOutlineStyleValue: (propertyName) => {
          const idleOutlineElement = this.root_.parentNode.querySelector(strings$4.IDLE_OUTLINE_SELECTOR);
          return window.getComputedStyle(idleOutlineElement).getPropertyValue(propertyName);
        },
      });
    }
  }

  /**
   * @license
   * Copyright 2016 Google Inc. All Rights Reserved.
   *
   * Licensed under the Apache License, Version 2.0 (the "License");
   * you may not use this file except in compliance with the License.
   * You may obtain a copy of the License at
   *
   *      http://www.apache.org/licenses/LICENSE-2.0
   *
   * Unless required by applicable law or agreed to in writing, software
   * distributed under the License is distributed on an "AS IS" BASIS,
   * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   * See the License for the specific language governing permissions and
   * limitations under the License.
   */
  /* eslint-enable no-unused-vars */

  /**
   * @extends {MDCComponent<!MDCTextFieldFoundation>}
   * @final
   */
  class MDCTextField extends MDCComponent {
    /**
     * @param {...?} args
     */
    constructor(...args) {
      super(...args);
      /** @private {?Element} */
      this.input_;
      /** @type {?MDCRipple} */
      this.ripple;
      /** @private {?MDCLineRipple} */
      this.lineRipple_;
      /** @private {?MDCTextFieldHelperText} */
      this.helperText_;
      /** @private {?MDCTextFieldIcon} */
      this.icon_;
      /** @private {?MDCFloatingLabel} */
      this.label_;
      /** @private {?MDCNotchedOutline} */
      this.outline_;
    }

    /**
     * @param {!Element} root
     * @return {!MDCTextField}
     */
    static attachTo(root) {
      return new MDCTextField(root);
    }

    /**
     * @param {(function(!Element): !MDCRipple)=} rippleFactory A function which
     * creates a new MDCRipple.
     * @param {(function(!Element): !MDCLineRipple)=} lineRippleFactory A function which
     * creates a new MDCLineRipple.
     * @param {(function(!Element): !MDCTextFieldHelperText)=} helperTextFactory A function which
     * creates a new MDCTextFieldHelperText.
     * @param {(function(!Element): !MDCTextFieldIcon)=} iconFactory A function which
     * creates a new MDCTextFieldIcon.
     * @param {(function(!Element): !MDCFloatingLabel)=} labelFactory A function which
     * creates a new MDCFloatingLabel.
     * @param {(function(!Element): !MDCNotchedOutline)=} outlineFactory A function which
     * creates a new MDCNotchedOutline.
     */
    initialize(
      rippleFactory = (el, foundation) => new MDCRipple(el, foundation),
      lineRippleFactory = (el) => new MDCLineRipple(el),
      helperTextFactory = (el) => new MDCTextFieldHelperText(el),
      iconFactory = (el) => new MDCTextFieldIcon(el),
      labelFactory = (el) => new MDCFloatingLabel(el),
      outlineFactory = (el) => new MDCNotchedOutline(el)) {
      this.input_ = this.root_.querySelector(strings$1.INPUT_SELECTOR);
      const labelElement = this.root_.querySelector(strings$1.LABEL_SELECTOR);
      if (labelElement) {
        this.label_ = labelFactory(labelElement);
      }
      const lineRippleElement = this.root_.querySelector(strings$1.LINE_RIPPLE_SELECTOR);
      if (lineRippleElement) {
        this.lineRipple_ = lineRippleFactory(lineRippleElement);
      }
      const outlineElement = this.root_.querySelector(strings$1.OUTLINE_SELECTOR);
      if (outlineElement) {
        this.outline_ = outlineFactory(outlineElement);
      }
      if (this.input_.hasAttribute(strings$1.ARIA_CONTROLS)) {
        const helperTextElement = document.getElementById(this.input_.getAttribute(strings$1.ARIA_CONTROLS));
        if (helperTextElement) {
          this.helperText_ = helperTextFactory(helperTextElement);
        }
      }
      const iconElement = this.root_.querySelector(strings$1.ICON_SELECTOR);
      if (iconElement) {
        this.icon_ = iconFactory(iconElement);
      }

      this.ripple = null;
      if (this.root_.classList.contains(cssClasses$1.BOX)) {
        const MATCHES = getMatchesProperty(HTMLElement.prototype);
        const adapter =
          Object.assign(MDCRipple.createAdapter(/** @type {!RippleCapableSurface} */ (this)), {
            isSurfaceActive: () => this.input_[MATCHES](':active'),
            registerInteractionHandler: (type, handler) => this.input_.addEventListener(type, handler),
            deregisterInteractionHandler: (type, handler) => this.input_.removeEventListener(type, handler),
          });
        const foundation = new MDCRippleFoundation(adapter);
        this.ripple = rippleFactory(this.root_, foundation);
      }
    }

    destroy() {
      if (this.ripple) {
        this.ripple.destroy();
      }
      if (this.lineRipple_) {
        this.lineRipple_.destroy();
      }
      if (this.helperText_) {
        this.helperText_.destroy();
      }
      if (this.icon_) {
        this.icon_.destroy();
      }
      if (this.label_) {
        this.label_.destroy();
      }
      if (this.outline_) {
        this.outline_.destroy();
      }
      super.destroy();
    }

    /**
     * Initiliazes the Text Field's internal state based on the environment's
     * state.
     */
    initialSyncWithDom() {
      this.disabled = this.input_.disabled;
    }

    /**
     * @return {string} The value of the input.
     */
    get value() {
      return this.foundation_.getValue();
    }

    /**
     * @param {string} value The value to set on the input.
     */
    set value(value) {
      this.foundation_.setValue(value);
    }

    /**
     * @return {boolean} True if the Text Field is disabled.
     */
    get disabled() {
      return this.foundation_.isDisabled();
    }

    /**
     * @param {boolean} disabled Sets the Text Field disabled or enabled.
     */
    set disabled(disabled) {
      this.foundation_.setDisabled(disabled);
    }

    /**
     * @return {boolean} valid True if the Text Field is valid.
     */
    get valid() {
      return this.foundation_.isValid();
    }

    /**
     * @param {boolean} valid Sets the Text Field valid or invalid.
     */
    set valid(valid) {
      this.foundation_.setValid(valid);
    }

    /**
     * @return {boolean} True if the Text Field is required.
     */
    get required() {
      return this.input_.required;
    }

    /**
     * @param {boolean} required Sets the Text Field to required.
     */
    set required(required) {
      this.input_.required = required;
    }

    /**
     * @return {string} The input element's validation pattern.
     */
    get pattern() {
      return this.input_.pattern;
    }

    /**
     * @param {string} pattern Sets the input element's validation pattern.
     */
    set pattern(pattern) {
      this.input_.pattern = pattern;
    }

    /**
     * @return {number} The input element's minLength.
     */
    get minLength() {
      return this.input_.minLength;
    }

    /**
     * @param {number} minLength Sets the input element's minLength.
     */
    set minLength(minLength) {
      this.input_.minLength = minLength;
    }

    /**
     * @return {number} The input element's maxLength.
     */
    get maxLength() {
      return this.input_.maxLength;
    }

    /**
     * @param {number} maxLength Sets the input element's maxLength.
     */
    set maxLength(maxLength) {
      // Chrome throws exception if maxLength is set < 0
      if (maxLength < 0) {
        this.input_.removeAttribute('maxLength');
      } else {
        this.input_.maxLength = maxLength;
      }
    }

    /**
     * @return {string} The input element's min.
     */
    get min() {
      return this.input_.min;
    }

    /**
     * @param {string} min Sets the input element's min.
     */
    set min(min) {
      this.input_.min = min;
    }

    /**
     * @return {string} The input element's max.
     */
    get max() {
      return this.input_.max;
    }

    /**
     * @param {string} max Sets the input element's max.
     */
    set max(max) {
      this.input_.max = max;
    }

    /**
     * @return {string} The input element's step.
     */
    get step() {
      return this.input_.step;
    }

    /**
     * @param {string} step Sets the input element's step.
     */
    set step(step) {
      this.input_.step = step;
    }

    /**
     * Sets the helper text element content.
     * @param {string} content
     */
    set helperTextContent(content) {
      this.foundation_.setHelperTextContent(content);
    }

    /**
     * Sets the aria label of the icon.
     * @param {string} label
     */
    set iconAriaLabel(label) {
      this.foundation_.setIconAriaLabel(label);
    }

    /**
     * Sets the text content of the icon.
     * @param {string} content
     */
    set iconContent(content) {
      this.foundation_.setIconContent(content);
    }

    /**
     * Recomputes the outline SVG path for the outline element.
     */
    layout() {
      const openNotch = this.foundation_.shouldFloat;
      this.foundation_.notchOutline(openNotch);
    }

    /**
     * @return {!MDCTextFieldFoundation}
     */
    getDefaultFoundation() {
      return new MDCTextFieldFoundation(
        /** @type {!MDCTextFieldAdapter} */ (Object.assign({
          addClass: (className) => this.root_.classList.add(className),
          removeClass: (className) => this.root_.classList.remove(className),
          hasClass: (className) => this.root_.classList.contains(className),
          registerTextFieldInteractionHandler: (evtType, handler) => this.root_.addEventListener(evtType, handler),
          deregisterTextFieldInteractionHandler: (evtType, handler) => this.root_.removeEventListener(evtType, handler),
          registerValidationAttributeChangeHandler: (handler) => {
            const getAttributesList = (mutationsList) => mutationsList.map((mutation) => mutation.attributeName);
            const observer = new MutationObserver((mutationsList) => handler(getAttributesList(mutationsList)));
            const targetNode = this.root_.querySelector(strings$1.INPUT_SELECTOR);
            const config = {attributes: true};
            observer.observe(targetNode, config);
            return observer;
          },
          deregisterValidationAttributeChangeHandler: (observer) => observer.disconnect(),
          isFocused: () => {
            return document.activeElement === this.root_.querySelector(strings$1.INPUT_SELECTOR);
          },
          isRtl: () => window.getComputedStyle(this.root_).getPropertyValue('direction') === 'rtl',
        },
        this.getInputAdapterMethods_(),
        this.getLabelAdapterMethods_(),
        this.getLineRippleAdapterMethods_(),
        this.getOutlineAdapterMethods_())),
        this.getFoundationMap_());
    }

    /**
     * @return {!{
     *   shakeLabel: function(boolean): undefined,
     *   floatLabel: function(boolean): undefined,
     *   hasLabel: function(): boolean,
     *   getLabelWidth: function(): number,
     * }}
     */
    getLabelAdapterMethods_() {
      return {
        shakeLabel: (shouldShake) => this.label_.shake(shouldShake),
        floatLabel: (shouldFloat) => this.label_.float(shouldFloat),
        hasLabel: () => !!this.label_,
        getLabelWidth: () => this.label_.getWidth(),
      };
    }

    /**
     * @return {!{
     *   activateLineRipple: function(): undefined,
     *   deactivateLineRipple: function(): undefined,
     *   setLineRippleTransformOrigin: function(number): undefined,
     * }}
     */
    getLineRippleAdapterMethods_() {
      return {
        activateLineRipple: () => {
          if (this.lineRipple_) {
            this.lineRipple_.activate();
          }
        },
        deactivateLineRipple: () => {
          if (this.lineRipple_) {
            this.lineRipple_.deactivate();
          }
        },
        setLineRippleTransformOrigin: (normalizedX) => {
          if (this.lineRipple_) {
            this.lineRipple_.setRippleCenter(normalizedX);
          }
        },
      };
    }

    /**
     * @return {!{
     *   notchOutline: function(number, boolean): undefined,
     *   hasOutline: function(): boolean,
     * }}
     */
    getOutlineAdapterMethods_() {
      return {
        notchOutline: (labelWidth, isRtl) => this.outline_.notch(labelWidth, isRtl),
        closeOutline: () => this.outline_.closeNotch(),
        hasOutline: () => !!this.outline_,
      };
    }

    /**
     * @return {!{
     *   registerInputInteractionHandler: function(string, function()): undefined,
     *   deregisterInputInteractionHandler: function(string, function()): undefined,
     *   getNativeInput: function(): ?Element,
     * }}
     */
    getInputAdapterMethods_() {
      return {
        registerInputInteractionHandler: (evtType, handler) => this.input_.addEventListener(evtType, handler),
        deregisterInputInteractionHandler: (evtType, handler) => this.input_.removeEventListener(evtType, handler),
        getNativeInput: () => this.input_,
      };
    }

    /**
     * Returns a map of all subcomponents to subfoundations.
     * @return {!FoundationMapType}
     */
    getFoundationMap_() {
      return {
        helperText: this.helperText_ ? this.helperText_.foundation : undefined,
        icon: this.icon_ ? this.icon_.foundation : undefined,
      };
    }
  }

  const regex = /%s/gi;

  function processClasses(classes, userClasses) {
      if (!userClasses) {
          return classes.join(" ");
      }
      if (Array.isArray(userClasses)) {
          userClasses = userClasses.join(" ");
      }
      if (!regex.test(userClasses)) {
          return userClasses+" "+classes.join(" ");
      }
      return userClasses.replace(regex, classes.join(" "));
  }

  /* src/Icon.html generated by Svelte v2.9.11 */

  function attrs(attributes) {
      let result = Object.assign({"aria_hidden": "true"}, attributes);
      let classes = ["material-icons"];
      delete result.attrs;
      result["class"] = processClasses(classes, result["class"]);
      return result;
  }
  var methods = {
  	getElement() {
  		return this.refs.icon;
  	}
  };

  function create_main_fragment(component, ctx) {
  	var i, slot_content_default = component._slotted.default;

  	function click_handler(event) {
  		component.fire("click", event);
  	}

  	var i_levels = [
  		ctx.attrs
  	];

  	var i_data = {};
  	for (var i_1 = 0; i_1 < i_levels.length; i_1 += 1) {
  		i_data = shared_js.assign(i_data, i_levels[i_1]);
  	}

  	return {
  		c() {
  			i = shared_js.createElement("i");
  			shared_js.addListener(i, "click", click_handler);
  			shared_js.setAttributes(i, i_data);
  		},

  		m(target, anchor) {
  			shared_js.insert(target, i, anchor);

  			if (slot_content_default) {
  				shared_js.append(i, slot_content_default);
  			}

  			component.refs.icon = i;
  		},

  		p(changed, ctx) {
  			shared_js.setAttributes(i, shared_js.getSpreadUpdate(i_levels, [
  				(changed.attrs) && ctx.attrs
  			]));
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(i);
  			}

  			if (slot_content_default) {
  				shared_js.reinsertChildren(i, slot_content_default);
  			}

  			shared_js.removeListener(i, "click", click_handler);
  			if (component.refs.icon === i) component.refs.icon = null;
  		}
  	};
  }

  function Icon(options) {
  	shared_js.init(this, options);
  	this.refs = {};
  	this._state = shared_js.assign({}, options.data);
  	this._recompute({  }, this._state);
  	this._intro = true;

  	this._slotted = options.slots || {};

  	this.slots = {};

  	this._fragment = create_main_fragment(this, this._state);

  	if (options.target) {
  		this._fragment.c();
  		this._mount(options.target, options.anchor);
  	}
  }

  shared_js.assign(Icon.prototype, shared_js.proto);
  shared_js.assign(Icon.prototype, methods);

  Icon.prototype._recompute = function _recompute(changed, state) {
  	if (this._differs(state.attrs, (state.attrs = attrs(state)))) changed.attrs = true;
  };

  /* src/TextArea.html generated by Svelte v2.9.11 */

  function inputAttrs({placeholder, helperText, id}) {
      var result = {};
      if (helperText) {
          let elem = id+"-helper-text";
          result["aria-controls"] = elem;
          result["aria-describedby"] = elem;
      }
      return result;
  }
  function attrs$1(attributes) {
      let result = Object.assign({}, attributes);
      let cls = "mdc-text-field";
      result.textarea = true;
      for (let key of ["box", "textarea", "dense", "fullwidth", "outlined", "disabled", "focused"]) {
          if (result[key]) {
              classes.push(cls+"--"+key);
          }
          delete result[key];
      }
      if (result.icon && result.iconMode) {
          classes.push(cls+"--with-"+result.iconMode+"-icon");
      }
      for (let key of ["attrs", "id", "name", "type", "helperText", "rows", "cols", "inputAttrs", "icon", "iconMode", ...mdcAttrs]) {
          delete result[key];
      }
      result["class"] = processClasses(classes, result["class"]);
      return result;
  }
  function oncreate() {
      this.mdcComponent = new MDCTextField(this.refs.textArea);
  }
  function ondestroy() {
      this.mdcComponent.destroy();
  }
  function onstate({current, changed}) {
      for (let key of mdcAttrs) {
          if (changed[key]) {
              this.mdcComponent[key] = current[key];
          }
      }
      for (let key of ["helperText", "outlined", "hintText", "iconMode"]) {
          if (changed[key]) {
              this.mdcComponent.destroy();
              this.mdcComponent = new MDCTextField(this.refs.textArea);
              break;
          }
      }
  }
  function create_main_fragment$1(component, ctx) {
  	var div, text, textarea, text_1, text_2, text_3, text_5, if_block_4_anchor;

  	var if_block = (ctx.icon && ctx.iconMode=="leading") && create_if_block(component, ctx);

  	var textarea_levels = [
  		{ id: ctx.id },
  		{ name: ctx.name },
  		{ class: "mdc-text-field__input" },
  		{ disabled: ctx.disabled },
  		{ rows: ctx.rows },
  		{ cols: ctx.cols },
  		ctx.inputAttrs,
  		{ value: ctx.value }
  	];

  	var textarea_data = {};
  	for (var i = 0; i < textarea_levels.length; i += 1) {
  		textarea_data = shared_js.assign(textarea_data, textarea_levels[i]);
  	}

  	var if_block_1 = (ctx.hintText) && create_if_block_1(component, ctx);

  	var if_block_2 = (ctx.icon && ctx.iconMode=="trailing") && create_if_block_2(component, ctx);

  	var if_block_3 = (ctx.outlined) && create_if_block_3(component, ctx);

  	var div_levels = [
  		ctx.attrs
  	];

  	var div_data = {};
  	for (var i = 0; i < div_levels.length; i += 1) {
  		div_data = shared_js.assign(div_data, div_levels[i]);
  	}

  	var if_block_4 = (ctx.helperText) && create_if_block_4(component, ctx);

  	return {
  		c() {
  			div = shared_js.createElement("div");
  			if (if_block) if_block.c();
  			text = shared_js.createText("\n  ");
  			textarea = shared_js.createElement("textarea");
  			text_1 = shared_js.createText("\n  ");
  			if (if_block_1) if_block_1.c();
  			text_2 = shared_js.createText("\n  ");
  			if (if_block_2) if_block_2.c();
  			text_3 = shared_js.createText("\n  ");
  			if (if_block_3) if_block_3.c();
  			text_5 = shared_js.createText("\n");
  			if (if_block_4) if_block_4.c();
  			if_block_4_anchor = shared_js.createComment();
  			shared_js.setAttributes(textarea, textarea_data);
  			shared_js.setAttributes(div, div_data);
  		},

  		m(target, anchor) {
  			shared_js.insert(target, div, anchor);
  			if (if_block) if_block.m(div, null);
  			shared_js.append(div, text);
  			shared_js.append(div, textarea);
  			shared_js.append(div, text_1);
  			if (if_block_1) if_block_1.m(div, null);
  			shared_js.append(div, text_2);
  			if (if_block_2) if_block_2.m(div, null);
  			shared_js.append(div, text_3);
  			if (if_block_3) if_block_3.m(div, null);
  			component.refs.textArea = div;
  			shared_js.insert(target, text_5, anchor);
  			if (if_block_4) if_block_4.m(target, anchor);
  			shared_js.insert(target, if_block_4_anchor, anchor);
  		},

  		p(changed, ctx) {
  			if (ctx.icon && ctx.iconMode=="leading") {
  				if (if_block) {
  					if_block.p(changed, ctx);
  				} else {
  					if_block = create_if_block(component, ctx);
  					if_block.c();
  					if_block.m(div, text);
  				}
  			} else if (if_block) {
  				if_block.d(1);
  				if_block = null;
  			}

  			shared_js.setAttributes(textarea, shared_js.getSpreadUpdate(textarea_levels, [
  				(changed.id) && { id: ctx.id },
  				(changed.name) && { name: ctx.name },
  				{ class: "mdc-text-field__input" },
  				(changed.disabled) && { disabled: ctx.disabled },
  				(changed.rows) && { rows: ctx.rows },
  				(changed.cols) && { cols: ctx.cols },
  				(changed.inputAttrs) && ctx.inputAttrs,
  				(changed.value) && { value: ctx.value }
  			]));

  			if (ctx.hintText) {
  				if (if_block_1) {
  					if_block_1.p(changed, ctx);
  				} else {
  					if_block_1 = create_if_block_1(component, ctx);
  					if_block_1.c();
  					if_block_1.m(div, text_2);
  				}
  			} else if (if_block_1) {
  				if_block_1.d(1);
  				if_block_1 = null;
  			}

  			if (ctx.icon && ctx.iconMode=="trailing") {
  				if (if_block_2) {
  					if_block_2.p(changed, ctx);
  				} else {
  					if_block_2 = create_if_block_2(component, ctx);
  					if_block_2.c();
  					if_block_2.m(div, text_3);
  				}
  			} else if (if_block_2) {
  				if_block_2.d(1);
  				if_block_2 = null;
  			}

  			if (ctx.outlined) {
  				if (!if_block_3) {
  					if_block_3 = create_if_block_3(component, ctx);
  					if_block_3.c();
  					if_block_3.m(div, null);
  				}
  			} else if (if_block_3) {
  				if_block_3.d(1);
  				if_block_3 = null;
  			}

  			shared_js.setAttributes(div, shared_js.getSpreadUpdate(div_levels, [
  				(changed.attrs) && ctx.attrs
  			]));

  			if (ctx.helperText) {
  				if (if_block_4) {
  					if_block_4.p(changed, ctx);
  				} else {
  					if_block_4 = create_if_block_4(component, ctx);
  					if_block_4.c();
  					if_block_4.m(if_block_4_anchor.parentNode, if_block_4_anchor);
  				}
  			} else if (if_block_4) {
  				if_block_4.d(1);
  				if_block_4 = null;
  			}
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(div);
  			}

  			if (if_block) if_block.d();
  			if (if_block_1) if_block_1.d();
  			if (if_block_2) if_block_2.d();
  			if (if_block_3) if_block_3.d();
  			if (component.refs.textArea === div) component.refs.textArea = null;
  			if (detach) {
  				shared_js.detachNode(text_5);
  			}

  			if (if_block_4) if_block_4.d(detach);
  			if (detach) {
  				shared_js.detachNode(if_block_4_anchor);
  			}
  		}
  	};
  }

  // (2:2) {#if icon && iconMode=="leading"}
  function create_if_block(component, ctx) {
  	var text;

  	var icon_initial_data = {
  	 	class: "mdc-text-field__icon",
  	 	tabindex: "0",
  	 	role: "button"
  	 };
  	var icon = new Icon({
  		root: component.root,
  		store: component.store,
  		slots: { default: shared_js.createFragment() },
  		data: icon_initial_data
  	});

  	return {
  		c() {
  			text = shared_js.createText(ctx.icon);
  			icon._fragment.c();
  		},

  		m(target, anchor) {
  			shared_js.append(icon._slotted.default, text);
  			icon._mount(target, anchor);
  		},

  		p(changed, ctx) {
  			if (changed.icon) {
  				shared_js.setData(text, ctx.icon);
  			}
  		},

  		d(detach) {
  			icon.destroy(detach);
  		}
  	};
  }

  // (6:2) {#if hintText}
  function create_if_block_1(component, ctx) {
  	var label, text;

  	return {
  		c() {
  			label = shared_js.createElement("label");
  			text = shared_js.createText(ctx.hintText);
  			label.className = "mdc-floating-label";
  			label.htmlFor = ctx.id;
  		},

  		m(target, anchor) {
  			shared_js.insert(target, label, anchor);
  			shared_js.append(label, text);
  		},

  		p(changed, ctx) {
  			if (changed.hintText) {
  				shared_js.setData(text, ctx.hintText);
  			}

  			if (changed.id) {
  				label.htmlFor = ctx.id;
  			}
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(label);
  			}
  		}
  	};
  }

  // (9:2) {#if icon && iconMode=="trailing"}
  function create_if_block_2(component, ctx) {
  	var text;

  	var icon_initial_data = {
  	 	class: "mdc-text-field__icon",
  	 	tabindex: "0",
  	 	role: "button"
  	 };
  	var icon = new Icon({
  		root: component.root,
  		store: component.store,
  		slots: { default: shared_js.createFragment() },
  		data: icon_initial_data
  	});

  	return {
  		c() {
  			text = shared_js.createText(ctx.icon);
  			icon._fragment.c();
  		},

  		m(target, anchor) {
  			shared_js.append(icon._slotted.default, text);
  			icon._mount(target, anchor);
  		},

  		p(changed, ctx) {
  			if (changed.icon) {
  				shared_js.setData(text, ctx.icon);
  			}
  		},

  		d(detach) {
  			icon.destroy(detach);
  		}
  	};
  }

  // (12:2) {#if outlined}
  function create_if_block_3(component, ctx) {
  	var div, text_1, div_1;

  	return {
  		c() {
  			div = shared_js.createElement("div");
  			div.innerHTML = "<svg><path class=\"mdc-notched-outline__path\"></path>\n      </svg>";
  			text_1 = shared_js.createText("\n    ");
  			div_1 = shared_js.createElement("div");
  			div.className = "mdc-notched-outline";
  			div_1.className = "mdc-notched-outline__idle";
  		},

  		m(target, anchor) {
  			shared_js.insert(target, div, anchor);
  			shared_js.insert(target, text_1, anchor);
  			shared_js.insert(target, div_1, anchor);
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(div);
  				shared_js.detachNode(text_1);
  				shared_js.detachNode(div_1);
  			}
  		}
  	};
  }

  // (21:0) {#if helperText}
  function create_if_block_4(component, ctx) {
  	var p, text, p_id_value;

  	return {
  		c() {
  			p = shared_js.createElement("p");
  			text = shared_js.createText(ctx.helperText);
  			p.id = p_id_value = "" + ctx.id + "-helper-text";
  			p.className = "mdc-text-field-helper-text";
  			shared_js.setAttribute(p, "aria-hidden", "true");
  		},

  		m(target, anchor) {
  			shared_js.insert(target, p, anchor);
  			shared_js.append(p, text);
  			component.refs.helperText = p;
  		},

  		p(changed, ctx) {
  			if (changed.helperText) {
  				shared_js.setData(text, ctx.helperText);
  			}

  			if ((changed.id) && p_id_value !== (p_id_value = "" + ctx.id + "-helper-text")) {
  				p.id = p_id_value;
  			}
  		},

  		d(detach) {
  			if (detach) {
  				shared_js.detachNode(p);
  			}

  			if (component.refs.helperText === p) component.refs.helperText = null;
  		}
  	};
  }

  function TextArea(options) {
  	shared_js.init(this, options);
  	this.refs = {};
  	this._state = shared_js.assign({}, options.data);
  	this._recompute({ placeholder: 1, helperText: 1, id: 1 }, this._state);
  	this._intro = true;

  	this._handlers.state = [onstate];

  	this._handlers.destroy = [ondestroy];

  	if (!options.root) {
  		this._oncreate = [];
  		this._beforecreate = [];
  		this._aftercreate = [];
  	}

  	this._fragment = create_main_fragment$1(this, this._state);

  	this.root._oncreate.push(() => {
  		onstate.call(this, { changed: shared_js.assignTrue({}, this._state), current: this._state });
  		oncreate.call(this);
  		this.fire("update", { changed: shared_js.assignTrue({}, this._state), current: this._state });
  	});

  	if (options.target) {
  		this._fragment.c();
  		this._mount(options.target, options.anchor);

  		this._lock = true;
  		shared_js.callAll(this._beforecreate);
  		shared_js.callAll(this._oncreate);
  		shared_js.callAll(this._aftercreate);
  		this._lock = false;
  	}
  }

  shared_js.assign(TextArea.prototype, shared_js.proto);

  TextArea.prototype._recompute = function _recompute(changed, state) {
  	if (changed.placeholder || changed.helperText || changed.id) {
  		if (this._differs(state.inputAttrs, (state.inputAttrs = inputAttrs(state)))) changed.inputAttrs = true;
  	}

  	if (this._differs(state.attrs, (state.attrs = attrs$1(state)))) changed.attrs = true;
  };

  return TextArea;

})));
